@@ -18,10 +18,19 @@
   along with OPM.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+// Improve IDE experience
+#ifndef OPM_BLACKOILAQUIFERMODEL_HEADER_INCLUDED
+#include <config.h>
+#define OPM_BLACKOILAQUIFERMODEL_IMPL_HEADER_INCLUDED
+#include <opm/simulators/aquifers/BlackoilAquiferModel.hpp>
+#endif
+
 #include <opm/simulators/aquifers/AquiferConstantFlux.hpp>
 
 #include <opm/common/ErrorMacros.hpp>
 
+#include <fmt/format.h>
+
 #include <algorithm>
 #include <memory>
 #include <stdexcept>
@@ -45,16 +54,22 @@ template <typename TypeTag>
 void
 BlackoilAquiferModel<TypeTag>::initialSolutionApplied()
 {
-    for (auto& aquifer : aquifers)
+    this->computeConnectionAreaFraction();
+
+    for (auto& aquifer : this->aquifers) {
         aquifer->initialSolutionApplied();
+    }
 }
 
 template <typename TypeTag>
 void
 BlackoilAquiferModel<TypeTag>::initFromRestart(const data::Aquifers& aquiferSoln)
 {
-    for (auto& aquifer : this->aquifers)
+    this->computeConnectionAreaFraction();
+
+    for (auto& aquifer : this->aquifers) {
         aquifer->initFromRestart(aquiferSoln);
+    }
 }
 
 template <typename TypeTag>
@@ -67,6 +82,8 @@ BlackoilAquiferModel<TypeTag>::beginEpisode()
     // SCHEDULE setup in this section it is the beginning of a report step
 
     this->createDynamicAquifers(this->simulator_.episodeIndex());
+
+    this->computeConnectionAreaFraction();
 }
 
 template <typename TypeTag>
@@ -78,8 +95,9 @@ template <typename TypeTag>
 void
 BlackoilAquiferModel<TypeTag>::beginTimeStep()
 {
-    for (auto& aquifer : aquifers)
+    for (auto& aquifer : this->aquifers) {
         aquifer->beginTimeStep();
+    }
 }
 
 template <typename TypeTag>
@@ -90,8 +108,9 @@ BlackoilAquiferModel<TypeTag>::addToSource(RateVector& rates,
                                            unsigned spaceIdx,
                                            unsigned timeIdx) const
 {
-    for (auto& aquifer : aquifers)
+    for (auto& aquifer : this->aquifers) {
         aquifer->addToSource(rates, context, spaceIdx, timeIdx);
+    }
 }
 
 template <typename TypeTag>
@@ -100,8 +119,9 @@ BlackoilAquiferModel<TypeTag>::addToSource(RateVector& rates,
                                            unsigned globalSpaceIdx,
                                            unsigned timeIdx) const
 {
-    for (auto& aquifer : aquifers)
+    for (auto& aquifer : this->aquifers) {
         aquifer->addToSource(rates, globalSpaceIdx, timeIdx);
+    }
 }
 
 template <typename TypeTag>
@@ -113,9 +133,10 @@ template <typename TypeTag>
 void
 BlackoilAquiferModel<TypeTag>::endTimeStep()
 {
-    for (auto& aquifer : aquifers) {
+    using NumAq = AquiferNumerical<TypeTag>;
+
+    for (auto& aquifer : this->aquifers) {
         aquifer->endTimeStep();
-        using NumAq = AquiferNumerical<TypeTag>;
         NumAq* num = dynamic_cast<NumAq*>(aquifer.get());
         if (num)
             this->simulator_.vanguard().grid().comm().barrier();
@@ -153,14 +174,18 @@ void BlackoilAquiferModel<TypeTag>::init()
         this->initializeStaticAquifers();
     }
 
+    if (this->needRestartDynamicAquifers()) {
+        this->initializeRestartDynamicAquifers();
+    }
 }
 
 template<typename TypeTag>
 data::Aquifers BlackoilAquiferModel<TypeTag>::aquiferData() const
 {
     data::Aquifers data;
-    for (const auto& aqu : this->aquifers)
+    for (const auto& aqu : this->aquifers) {
         data.insert_or_assign(aqu->aquiferID(), aqu->aquiferData());
+    }
 
     return data;
 }
@@ -170,7 +195,7 @@ template<class Serializer>
 void BlackoilAquiferModel<TypeTag>::
 serializeOp(Serializer& serializer)
 {
-    for (auto& aiPtr : aquifers) {
+    for (auto& aiPtr : this->aquifers) {
         auto* ct = dynamic_cast<AquiferCarterTracy<TypeTag>*>(aiPtr.get());
         auto* fetp = dynamic_cast<AquiferFetkovich<TypeTag>*>(aiPtr.get());
         auto* num = dynamic_cast<AquiferNumerical<TypeTag>*>(aiPtr.get());
@@ -189,6 +214,15 @@ serializeOp(Serializer& serializer)
     }
 }
 
+template <typename TypeTag>
+void BlackoilAquiferModel<TypeTag>::initializeRestartDynamicAquifers()
+{
+    const auto rstStep = this->simulator_.vanguard().eclState()
+        .getInitConfig().getRestartStep() - 1;
+
+    this->createDynamicAquifers(rstStep);
+}
+
 template <typename TypeTag>
 void BlackoilAquiferModel<TypeTag>::initializeStaticAquifers()
 {
@@ -235,6 +269,20 @@ void BlackoilAquiferModel<TypeTag>::initializeStaticAquifers()
     }
 }
 
+template <typename TypeTag>
+bool BlackoilAquiferModel<TypeTag>::needRestartDynamicAquifers() const
+{
+    const auto& initconfig =
+        this->simulator_.vanguard().eclState().getInitConfig();
+
+    if (! initconfig.restartRequested()) {
+        return false;
+    }
+
+    return this->simulator_.vanguard()
+        .schedule()[initconfig.getRestartStep() - 1].hasAnalyticalAquifers();
+}
+
 template <typename TypeTag>
 template <typename AquiferType, typename AquiferData>
 std::unique_ptr<AquiferType>
@@ -304,4 +352,26 @@ void BlackoilAquiferModel<TypeTag>::createDynamicAquifers(const int episode_inde
     }
 }
 
+template <typename TypeTag>
+void BlackoilAquiferModel<TypeTag>::computeConnectionAreaFraction() const
+{
+    auto maxAquID =
+        std::accumulate(this->aquifers.begin(), this->aquifers.end(), 0,
+                        [](const int aquID, const auto& aquifer)
+                        { return std::max(aquID, aquifer->aquiferID()); });
+
+    maxAquID = this->simulator_.vanguard().grid().comm().max(maxAquID);
+
+    auto totalConnArea = std::vector<Scalar>(maxAquID, 0.0);
+    for (const auto& aquifer : this->aquifers) {
+        totalConnArea[aquifer->aquiferID() - 1] += aquifer->totalFaceArea();
+    }
+
+    this->simulator_.vanguard().grid().comm().sum(totalConnArea.data(), maxAquID);
+
+    for (auto& aquifer : this->aquifers) {
+        aquifer->computeFaceAreaFraction(totalConnArea);
+    }
+}
+
 } // namespace Opm
