@@ -142,6 +142,7 @@ struct EclWriteTasklet : public Opm::TaskletInterface
     Opm::UDQState udqState_;
     Opm::EclipseIO& eclIO_;
     int reportStepNum_;
+    std::optional<int> timeStepNum_;
     bool isSubStep_;
     double secondsElapsed_;
     Opm::RestartValue restartValue_;
@@ -153,6 +154,7 @@ struct EclWriteTasklet : public Opm::TaskletInterface
                              const Opm::UDQState& udqState,
                              Opm::EclipseIO& eclIO,
                              int reportStepNum,
+                             std::optional<int> timeStepNum,
                              bool isSubStep,
                              double secondsElapsed,
                              Opm::RestartValue restartValue,
@@ -163,6 +165,7 @@ struct EclWriteTasklet : public Opm::TaskletInterface
         , udqState_(udqState)
         , eclIO_(eclIO)
         , reportStepNum_(reportStepNum)
+        , timeStepNum_(timeStepNum)
         , isSubStep_(isSubStep)
         , secondsElapsed_(secondsElapsed)
         , restartValue_(std::move(restartValue))
@@ -180,7 +183,9 @@ struct EclWriteTasklet : public Opm::TaskletInterface
                                    this->isSubStep_,
                                    this->secondsElapsed_,
                                    std::move(this->restartValue_),
-                                   this->writeDoublePrecision_);
+                                   this->writeDoublePrecision_,
+                                   this->timeStepNum_
+);
     }
 };
 
@@ -241,7 +246,7 @@ eclIO() const
 
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
 void EclGenericWriter<Grid,EquilGrid,GridView,ElementMapper,Scalar>::
-writeInit(const std::function<unsigned int(unsigned int)>& map)
+writeInit()
 {
     if (collectOnIORank_.isIORank()) {
         std::map<std::string, std::vector<int>> integerVectors;
@@ -249,11 +254,21 @@ writeInit(const std::function<unsigned int(unsigned int)>& map)
             integerVectors.emplace("MPI_RANK", collectOnIORank_.globalRanks());
         }
 
-        auto cartMap = cartesianToCompressed(equilGrid_->size(0), UgGridHelpers::globalCell(*equilGrid_));
-
-        eclIO_->writeInitial(computeTrans_(cartMap, map),
+        eclIO_->writeInitial(*this->outputTrans_,
                              integerVectors,
-                             exportNncStructure_(cartMap, map));
+                             this->outputNnc_);
+        this->outputTrans_.reset();
+    }
+}
+template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
+void
+EclGenericWriter<Grid,EquilGrid,GridView,ElementMapper,Scalar>::
+extractOutputTransAndNNC(const std::function<unsigned int(unsigned int)>& map)
+{
+    if (collectOnIORank_.isIORank()) {
+        auto cartMap = cartesianToCompressed(equilGrid_->size(0), UgGridHelpers::globalCell(*equilGrid_));
+        computeTrans_(cartMap, map);
+        exportNncStructure_(cartMap, map);
     }
 
 #if HAVE_MPI
@@ -266,22 +281,34 @@ writeInit(const std::function<unsigned int(unsigned int)>& map)
 }
 
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
-data::Solution
+void
 EclGenericWriter<Grid,EquilGrid,GridView,ElementMapper,Scalar>::
 computeTrans_(const std::unordered_map<int,int>& cartesianToActive,
               const std::function<unsigned int(unsigned int)>& map) const
 {
+    if (!outputTrans_) {
+        outputTrans_ = std::make_unique<data::Solution>();
+    }
+
     const auto& cartMapper = *equilCartMapper_;
     const auto& cartDims = cartMapper.cartesianDimensions();
 
-    auto tranx = data::CellData {
-        UnitSystem::measure::transmissibility,
-        std::vector<double>(cartDims[0] * cartDims[1] * cartDims[2], 0.0),
-        data::TargetType::INIT
+    auto createCellData = [&cartDims]() {
+        return data::CellData{
+                UnitSystem::measure::transmissibility,
+                std::vector<double>(cartDims[0] * cartDims[1] * cartDims[2], 0.0),
+                data::TargetType::INIT
+        };
     };
 
-    auto trany = tranx;
-    auto tranz = tranx;
+    outputTrans_->clear();
+    outputTrans_->emplace("TRANX", createCellData());
+    outputTrans_->emplace("TRANY", createCellData());
+    outputTrans_->emplace("TRANZ", createCellData());
+
+    auto& tranx = this->outputTrans_->at("TRANX");
+    auto& trany = this->outputTrans_->at("TRANY");
+    auto& tranz = this->outputTrans_->at("TRANZ");
 
     using GlobalGridView = typename EquilGrid::LeafGridView;
     using GlobElementMapper = Dune::MultipleCodimMultipleGeomTypeMapper<GlobalGridView>;
@@ -332,26 +359,20 @@ computeTrans_(const std::unordered_map<int,int>& cartesianToActive,
             }
 
             if (gc2 - gc1 == 1 && cartDims[0] > 1 ) {
-                tranx.data<double>()[gc1] = globalTrans().transmissibility(c1, c2);
+                tranx.template data<double>()[gc1] = globalTrans().transmissibility(c1, c2);
                 continue; // skip other if clauses as they are false, last one needs some computation
             }
 
             if (gc2 - gc1 == cartDims[0] && cartDims[1] > 1) {
-                trany.data<double>()[gc1] = globalTrans().transmissibility(c1, c2);
+                trany.template data<double>()[gc1] = globalTrans().transmissibility(c1, c2);
                 continue; // skipt next if clause as it needs some computation
             }
 
             if ( gc2 - gc1 == cartDims[0]*cartDims[1] ||
                  directVerticalNeighbors(cartDims, cartesianToActive, gc1, gc2))
-                tranz.data<double>()[gc1] = globalTrans().transmissibility(c1, c2);
+                tranz.template data<double>()[gc1] = globalTrans().transmissibility(c1, c2);
         }
     }
-
-    return {
-        {"TRANX", tranx},
-        {"TRANY", trany},
-        {"TRANZ", tranz},
-    };
 }
 
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
@@ -508,6 +529,7 @@ exportNncStructure_(const std::unordered_map<int,int>& cartesianToActive,
 template<class Grid, class EquilGrid, class GridView, class ElementMapper, class Scalar>
 void EclGenericWriter<Grid,EquilGrid,GridView,ElementMapper,Scalar>::
 doWriteOutput(const int                          reportStepNum,
+              const std::optional<int>           timeStepNum,
               const bool                         isSubStep,
               data::Solution&&                   localCellData,
               data::Wells&&                      localWellData,
@@ -576,19 +598,22 @@ doWriteOutput(const int                          reportStepNum,
             restartValue.addExtra(flores.name, UnitSystem::measure::rate, flores.values);
         }
     }
+    // make sure that the previous I/O request has been completed
+    // and the number of incomplete tasklets does not increase between
+    // time steps
+    this->taskletRunner_->barrier();
 
-    // first, create a tasklet to write the data for the current time
-    // step to disk
+    // check if there might have been a failure in the TaskletRunner
+    if (this->taskletRunner_->failure()) {
+        throw std::runtime_error("Failure in the TaskletRunner while writing output.");
+    }
+
+    // create a tasklet to write the data for the current time step to disk
     auto eclWriteTasklet = std::make_shared<EclWriteTasklet>(
         actionState,
         isParallel ? this->collectOnIORank_.globalWellTestState() : std::move(localWTestState),
         summaryState, udqState, *this->eclIO_,
-        reportStepNum, isSubStep, curTime, std::move(restartValue), doublePrecision);
-
-    // then, make sure that the previous I/O request has been completed
-    // and the number of incomplete tasklets does not increase between
-    // time steps
-    this->taskletRunner_->barrier();
+        reportStepNum, timeStepNum, isSubStep, curTime, std::move(restartValue), doublePrecision);
 
     // finally, start a new output writing job
     this->taskletRunner_->dispatch(std::move(eclWriteTasklet));
