@@ -3,7 +3,8 @@
   Copyright 2014 Dr. Blatt - HPC-Simulation-Software & Services
   Copyright 2015 IRIS AS
   Copyright 2014 STATOIL ASA.
-
+  Copyright 2023 Inria
+  
   This file is part of the Open Porous Media project (OPM).
 
   OPM is free software: you can redistribute it and/or modify
@@ -22,41 +23,44 @@
 #ifndef OPM_MAIN_HEADER_INCLUDED
 #define OPM_MAIN_HEADER_INCLUDED
 
-#include <flow/flow_ebos_blackoil.hpp>
-#include <flow/flow_ebos_blackoil_legacyassembly.hpp>
+#include <flow/flow_blackoil.hpp>
+#include <flow/flow_blackoil_legacyassembly.hpp>
 
-#include <flow/flow_ebos_gasoil.hpp>
-#include <flow/flow_ebos_gasoildiffuse.hpp>
-#include <flow/flow_ebos_gasoil_energy.hpp>
-#include <flow/flow_ebos_oilwater.hpp>
-#include <flow/flow_ebos_gaswater.hpp>
-#include <flow/flow_ebos_solvent.hpp>
-#include <flow/flow_ebos_polymer.hpp>
-#include <flow/flow_ebos_extbo.hpp>
-#include <flow/flow_ebos_foam.hpp>
-#include <flow/flow_ebos_brine.hpp>
-#include <flow/flow_ebos_brine_saltprecipitation.hpp>
-#include <flow/flow_ebos_gaswater_saltprec_vapwat.hpp>
-#include <flow/flow_ebos_brine_precsalt_vapwat.hpp>
-#include <flow/flow_ebos_onephase.hpp>
-#include <flow/flow_ebos_onephase_energy.hpp>
-#include <flow/flow_ebos_oilwater_brine.hpp>
-#include <flow/flow_ebos_gaswater_brine.hpp>
-#include <flow/flow_ebos_gaswater_energy.hpp>
-#include <flow/flow_ebos_gaswater_dissolution.hpp>
-#include <flow/flow_ebos_gaswater_dissolution_diffuse.hpp>
-#include <flow/flow_ebos_energy.hpp>
-#include <flow/flow_ebos_oilwater_polymer.hpp>
-#include <flow/flow_ebos_oilwater_polymer_injectivity.hpp>
-#include <flow/flow_ebos_micp.hpp>
+#include <flow/flow_gasoil.hpp>
+#include <flow/flow_gasoildiffuse.hpp>
+#include <flow/flow_gasoil_energy.hpp>
+#include <flow/flow_oilwater.hpp>
+#include <flow/flow_gaswater.hpp>
+#include <flow/flow_gaswater_solvent.hpp>
+#include <flow/flow_solvent.hpp>
+#include <flow/flow_solvent_foam.hpp>
+#include <flow/flow_polymer.hpp>
+#include <flow/flow_extbo.hpp>
+#include <flow/flow_foam.hpp>
+#include <flow/flow_brine.hpp>
+#include <flow/flow_brine_saltprecipitation.hpp>
+#include <flow/flow_gaswater_saltprec_vapwat.hpp>
+#include <flow/flow_gaswater_saltprec_energy.hpp>
+#include <flow/flow_brine_precsalt_vapwat.hpp>
+#include <flow/flow_onephase.hpp>
+#include <flow/flow_onephase_energy.hpp>
+#include <flow/flow_oilwater_brine.hpp>
+#include <flow/flow_gaswater_brine.hpp>
+#include <flow/flow_gaswater_energy.hpp>
+#include <flow/flow_gaswater_dissolution.hpp>
+#include <flow/flow_gaswater_dissolution_diffuse.hpp>
+#include <flow/flow_energy.hpp>
+#include <flow/flow_oilwater_polymer.hpp>
+#include <flow/flow_oilwater_polymer_injectivity.hpp>
+#include <flow/flow_micp.hpp>
 
 #include <opm/input/eclipse/EclipseState/EclipseState.hpp>
 
 #include <opm/models/utils/propertysystem.hh>
-#include <opm/models/utils/parametersystem.hh>
+#include <opm/models/utils/parametersystem.hpp>
 
 #include <opm/simulators/flow/Banners.hpp>
-#include <opm/simulators/flow/FlowMainEbos.hpp>
+#include <opm/simulators/flow/FlowMain.hpp>
 
 #if HAVE_DUNE_FEM
 #include <dune/fem/misc/mpimanager.hh>
@@ -68,8 +72,14 @@
 #include <opm/simulators/utils/ParallelEclipseState.hpp>
 #endif
 
+#if HAVE_DAMARIS
+#include <opm/simulators/utils/DamarisKeywords.hpp>
+#endif
+
 #include <cassert>
+#include <charconv>
 #include <cstdlib>
+#include <filesystem>
 #include <iostream>
 #include <memory>
 #include <stdexcept>
@@ -84,7 +94,7 @@ namespace Opm::Properties {
 // simulator.
 namespace TTag {
 struct FlowEarlyBird {
-    using InheritsFrom = std::tuple<EclFlowProblem>;
+    using InheritsFrom = std::tuple<FlowProblem>;
 };
 }
 
@@ -98,13 +108,13 @@ class WellTestState;
 
 // ----------------- Main program -----------------
 template <class TypeTag>
-int flowEbosMain(int argc, char** argv, bool outputCout, bool outputFiles)
+int flowMain(int argc, char** argv, bool outputCout, bool outputFiles)
 {
     // we always want to use the default locale, and thus spare us the trouble
     // with incorrect locale settings.
     resetLocale();
 
-    FlowMainEbos<TypeTag> mainfunc(argc, argv, outputCout, outputFiles);
+    FlowMain<TypeTag> mainfunc(argc, argv, outputCout, outputFiles);
     return mainfunc.execute();
 }
 
@@ -118,17 +128,19 @@ int flowEbosMain(int argc, char** argv, bool outputCout, bool outputFiles)
 class Main
 {
 public:
-    Main(int argc, char** argv);
+    Main(int argc, char** argv, bool ownMPI = true);
 
     // This constructor can be called from Python
-    Main(const std::string& filename);
+    Main(const std::string& filename, bool mpi_init = true, bool mpi_finalize = true);
 
     // This constructor can be called from Python when Python has
     // already parsed a deck
     Main(const std::string& filename,
          std::shared_ptr<EclipseState> eclipseState,
          std::shared_ptr<Schedule> schedule,
-         std::shared_ptr<SummaryConfig> summaryConfig);
+         std::shared_ptr<SummaryConfig> summaryConfig,
+         bool mpi_init = true,
+         bool mpi_finalize = true);
 
     ~Main();
 
@@ -139,8 +151,9 @@ public:
     int runDynamic()
     {
         int exitCode = EXIT_SUCCESS;
-        if (isSimulationRank_) {
-            if (initialize_<Properties::TTag::FlowEarlyBird>(exitCode)) {
+        if (initialize_<Properties::TTag::FlowEarlyBird>(exitCode)) {
+            Parameters::reset();
+            if (isSimulationRank_) {
                 return this->dispatchDynamic_();
             }
         }
@@ -152,8 +165,8 @@ public:
     int runStatic()
     {
         int exitCode = EXIT_SUCCESS;
-        if (isSimulationRank_) {
-            if (initialize_<TypeTag>(exitCode)) {
+        if (initialize_<TypeTag>(exitCode)) {
+            if (isSimulationRank_) {
                 return this->dispatchStatic_<TypeTag>();
             }
         }
@@ -161,24 +174,12 @@ public:
         return exitCode;
     }
 
-    using FlowMainEbosType = FlowMainEbos<Properties::TTag::EclFlowProblemTPFA>;
-    // To be called from the Python interface code. Only do the
-    // initialization and then return a pointer to the FlowEbosMain
-    // object that can later be accessed directly from the Python interface
-    // to e.g. advance the simulator one report step
-    std::unique_ptr<FlowMainEbosType> initFlowEbosBlackoil(int& exitCode)
+    //! \brief Used for test_outputdir.
+    int justInitialize()
     {
-        exitCode = EXIT_SUCCESS;
-        if (initialize_<Properties::TTag::FlowEarlyBird>(exitCode)) {
-            // TODO: check that this deck really represents a blackoil
-            // case. E.g. check that number of phases == 3
-            this->setupVanguard();
-            return flowEbosBlackoilTpfaMainInit(
-                argc_, argv_, outputCout_, outputFiles_);
-        } else {
-            //NOTE: exitCode was set by initialize_() above;
-            return std::unique_ptr<FlowMainEbosType>(); // nullptr
-        }
+        int exitCode = EXIT_SUCCESS;
+        initialize_<Properties::TTag::FlowEarlyBird>(exitCode);
+        return exitCode;
     }
 
 private:
@@ -221,18 +222,18 @@ private:
         }
 
         // Foam case
-        else if (phases.active(Phase::FOAM)) {
+        else if (phases.active(Phase::FOAM) && !phases.active(Phase::SOLVENT)) {
             return this->runFoam();
         }
 
-        // Brine case
-        else if (phases.active(Phase::BRINE)) {
-            return this->runBrine(phases);
-        }
-
         // Solvent case
         else if (phases.active(Phase::SOLVENT)) {
-            return this->runSolvent();
+            return this->runSolvent(phases);
+        }
+
+        // Brine case
+        else if (phases.active(Phase::BRINE) && !thermal) {
+            return this->runBrine(phases);
         }
 
         // Extended BO case
@@ -265,17 +266,19 @@ private:
     int dispatchStatic_()
     {
         this->setupVanguard();
-        return flowEbosMain<TypeTag>(argc_, argv_, outputCout_, outputFiles_);
+        return flowMain<TypeTag>(argc_, argv_, outputCout_, outputFiles_);
     }
 
+protected:
     /// \brief Initialize
     /// \param exitCode The exitCode of the program.
+    /// \param keepKeywords Keep Schedule keywords even if there are no actions
     ///
     /// \return Whether to actually run the simulator. I.e. true if
     /// parsing of command line was successful and no --help,
     /// --print-properties, or --print-parameters have been found.
     template <class TypeTagEarlyBird>
-    bool initialize_(int& exitCode)
+    bool initialize_(int& exitCode, bool keepKeywords = false)
     {
         Dune::Timer externalSetupTimer;
         externalSetupTimer.start();
@@ -296,7 +299,7 @@ private:
         using PreProblem = GetPropType<PreTypeTag, Properties::Problem>;
 
         PreProblem::setBriefDescription("Flow, an advanced reservoir simulator for ECL-decks provided by the Open Porous Media project.");
-        int status = FlowMainEbos<PreTypeTag>::setupParameters_(argc_, argv_, EclGenericVanguard::comm());
+        int status = FlowMain<PreTypeTag>::setupParameters_(argc_, argv_, FlowGenericVanguard::comm());
         if (status != 0) {
             // if setupParameters_ returns a value smaller than 0, there was no error, but
             // the program should abort. This is the case e.g. for the --help and the
@@ -316,23 +319,48 @@ private:
             outputDir = eclipseState_->getIOConfig().getOutputDir();
         }
         else {
-            deckFilename = EWOMS_GET_PARAM(PreTypeTag, std::string, EclDeckFileName);
-            outputDir = EWOMS_GET_PARAM(PreTypeTag, std::string, OutputDir);
+            deckFilename = Parameters::Get<Parameters::EclDeckFileName>();
+            outputDir = Parameters::Get<Parameters::OutputDir>();
         }
 
 #if HAVE_DAMARIS
-        enableDamarisOutput_ = EWOMS_GET_PARAM(PreTypeTag, bool, EnableDamarisOutput);
+        enableDamarisOutput_ = Parameters::Get<Parameters::EnableDamarisOutput>();
+        
+        // Reset to false as we cannot use Damaris if there is only one rank.
+        if ((enableDamarisOutput_ == true) && (FlowGenericVanguard::comm().size() == 1)) {
+            std::string msg ;
+            msg = "\nUse of Damaris (command line argument --enable-damaris-output=true) has been disabled for run with only one rank.\n" ;
+            OpmLog::warning(msg);
+            enableDamarisOutput_ = false ;
+        }
+
         if (enableDamarisOutput_) {
-            this->setupDamaris(outputDir,
-                               EWOMS_GET_PARAM(PreTypeTag, bool, EnableDamarisOutputCollective));
+            // Deal with empty (defaulted) output dir, should be deck dir
+            auto damarisOutputDir = outputDir;
+            if (outputDir.empty()) {
+                auto odir = std::filesystem::path{deckFilename}.parent_path();
+                if (odir.empty()) {
+                    damarisOutputDir = ".";
+                } else {
+                    damarisOutputDir = odir.generic_string();
+                }
+            }
+            // Damaris server ranks will block here until damaris_stop() is called by client ranks
+            this->setupDamaris(damarisOutputDir);
         }
 #endif // HAVE_DAMARIS
 
-        int mpiRank = EclGenericVanguard::comm().rank();
+        // Guard for when the Damaris core(s) return from damaris_start() 
+        // which happens when damaris_stop() is called in main simulation
+        if (!isSimulationRank_) {
+            exitCode = EXIT_SUCCESS;
+            return true;
+        }
+        
+        int mpiRank = FlowGenericVanguard::comm().rank();
         outputCout_ = false;
         if (mpiRank == 0)
-            outputCout_ = EWOMS_GET_PARAM(PreTypeTag, bool, EnableTerminalOutput);
-
+            outputCout_ = Parameters::Get<Parameters::EnableTerminalOutput>();
 
         if (deckFilename.empty()) {
             if (mpiRank == 0) {
@@ -359,11 +387,11 @@ private:
 
         std::string cmdline_params;
         if (outputCout_) {
-            printFlowBanner(EclGenericVanguard::comm().size(),
-                            getNumThreads<PreTypeTag>(),
+            printFlowBanner(FlowGenericVanguard::comm().size(),
+                            getNumThreads(),
                             Opm::moduleVersionName());
             std::ostringstream str;
-            Parameters::printValues<PreTypeTag>(str);
+            Parameters::printValues(str);
             cmdline_params = str.str();
         }
 
@@ -371,12 +399,15 @@ private:
         try {
             this->readDeck(deckFilename,
                            outputDir,
-                           EWOMS_GET_PARAM(PreTypeTag, std::string, OutputMode),
-                           !EWOMS_GET_PARAM(PreTypeTag, bool, SchedRestart),
-                           EWOMS_GET_PARAM(PreTypeTag, bool,  EnableLoggingFalloutWarning),
-                           EWOMS_GET_PARAM(PreTypeTag, std::string, ParsingStrictness),
-                           mpiRank,
-                           EWOMS_GET_PARAM(PreTypeTag, int, EclOutputInterval),
+                           Parameters::Get<Parameters::OutputMode>(),
+                           !Parameters::Get<Parameters::SchedRestart>(),
+                           Parameters::Get<Parameters::EnableLoggingFalloutWarning>(),
+                           Parameters::Get<Parameters::ParsingStrictness>(),
+                           Parameters::Get<Parameters::ActionParsingStrictness>(),
+                           Parameters::Get<Parameters::InputSkipMode>(),
+                           keepKeywords,
+                           getNumThreads(),
+                           Parameters::Get<Parameters::EclOutputInterval>(),
                            cmdline_params,
                            Opm::moduleVersion(),
                            Opm::compileTimestamp());
@@ -399,6 +430,9 @@ private:
         return true;
     }
 
+    void setupVanguard();
+
+private:
     // This function is an extreme special case, if the program has been invoked
     // *exactly* as:
     //
@@ -428,23 +462,24 @@ private:
             return EXIT_FAILURE;
         }
 
-        return flowEbosMICPMain(this->argc_,
-                                this->argv_,
-                                this->outputCout_,
-                                this->outputFiles_);
+        return flowMICPMain(this->argc_,
+                            this->argv_,
+                            this->outputCout_,
+                            this->outputFiles_);
     }
 
     int runTwoPhase(const Phases& phases)
     {
         const bool diffusive = eclipseState_->getSimulationConfig().isDiffusive();
         const bool disgasw = eclipseState_->getSimulationConfig().hasDISGASW();
+        const bool vapwat = eclipseState_->getSimulationConfig().hasVAPWAT();
 
         // oil-gas
         if (phases.active( Phase::OIL ) && phases.active( Phase::GAS )) {
             if (diffusive) {
-                return flowEbosGasOilDiffuseMain(argc_, argv_, outputCout_, outputFiles_);
+                return flowGasOilDiffuseMain(argc_, argv_, outputCout_, outputFiles_);
             } else {
-                return flowEbosGasOilMain(argc_, argv_, outputCout_, outputFiles_);
+                return flowGasOilMain(argc_, argv_, outputCout_, outputFiles_);
             }
         }
 
@@ -456,25 +491,25 @@ private:
                 }
                 return EXIT_FAILURE;
             }
-            return flowEbosOilWaterMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowOilWaterMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
         // gas-water
         else if ( phases.active( Phase::GAS ) && phases.active( Phase::WATER ) ) {
-            if (disgasw) {
+            if (disgasw || vapwat) {
                 if (diffusive) {
-                    return flowEbosGasWaterDissolutionDiffuseMain(argc_, argv_, outputCout_, outputFiles_);
+                    return flowGasWaterDissolutionDiffuseMain(argc_, argv_, outputCout_, outputFiles_);
                 }
-                return flowEbosGasWaterDissolutionMain(argc_, argv_, outputCout_, outputFiles_);
+                return flowGasWaterDissolutionMain(argc_, argv_, outputCout_, outputFiles_);
             }
             if (diffusive) {
                 if (outputCout_) {
-                    std::cerr << "The DIFFUSE option is not available for the two-phase gas/water model without disgasw." << std::endl;
+                    std::cerr << "The DIFFUSE option is not available for the two-phase gas/water model without disgasw or vapwat." << std::endl;
                 }
                 return EXIT_FAILURE;
             }
 
-            return flowEbosGasWaterMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowGasWaterMain(argc_, argv_, outputCout_, outputFiles_);
         }
         else {
             if (outputCout_) {
@@ -500,20 +535,20 @@ private:
         if (phases.active(Phase::POLYMW)) {
             // only oil water two phase for now
             assert (phases.size() == 4);
-            return flowEbosOilWaterPolymerInjectivityMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowOilWaterPolymerInjectivityMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
         if (phases.size() == 3) { // oil water polymer case
-            return flowEbosOilWaterPolymerMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowOilWaterPolymerMain(argc_, argv_, outputCout_, outputFiles_);
         }
         else {
-            return flowEbosPolymerMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowPolymerMain(argc_, argv_, outputCout_, outputFiles_);
         }
     }
 
     int runFoam()
     {
-        return flowEbosFoamMain(argc_, argv_, outputCout_, outputFiles_);
+        return flowFoamMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
     int runWaterOnly(const Phases& phases)
@@ -526,7 +561,7 @@ private:
             return EXIT_FAILURE;
         }
 
-        return flowEbosWaterOnlyMain(argc_, argv_, outputCout_, outputFiles_);
+        return flowWaterOnlyMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
     int runWaterOnlyEnergy(const Phases& phases)
@@ -539,7 +574,7 @@ private:
             return EXIT_FAILURE;
         }
 
-        return flowEbosWaterOnlyEnergyMain(argc_, argv_, outputCout_, outputFiles_);
+        return flowWaterOnlyEnergyMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
     int runBrine(const Phases& phases)
@@ -555,58 +590,79 @@ private:
         if (phases.size() == 3) {
 
             if (phases.active(Phase::OIL)){ // oil water brine case
-                return flowEbosOilWaterBrineMain(argc_, argv_, outputCout_, outputFiles_);
+                return flowOilWaterBrineMain(argc_, argv_, outputCout_, outputFiles_);
             }
             if (phases.active(Phase::GAS)){ // gas water brine case
                 if (eclipseState_->getSimulationConfig().hasPRECSALT() &&
                     eclipseState_->getSimulationConfig().hasVAPWAT()) {
                     //case with water vaporization into gas phase and salt precipitation
-                    return flowEbosGasWaterSaltprecVapwatMain(argc_, argv_, outputCout_, outputFiles_);
+                    return flowGasWaterSaltprecVapwatMain(argc_, argv_, outputCout_, outputFiles_);
                 }
                 else {
-                    return flowEbosGasWaterBrineMain(argc_, argv_, outputCout_, outputFiles_);
+                    return flowGasWaterBrineMain(argc_, argv_, outputCout_, outputFiles_);
                 }
             }
         }
         else if (eclipseState_->getSimulationConfig().hasPRECSALT()) {
             if (eclipseState_->getSimulationConfig().hasVAPWAT()) {
                     //case with water vaporization into gas phase and salt precipitation
-                    return flowEbosBrinePrecsaltVapwatMain(argc_, argv_, outputCout_, outputFiles_);
+                    return flowBrinePrecsaltVapwatMain(argc_, argv_, outputCout_, outputFiles_);
             }
             else {
-                return flowEbosBrineSaltPrecipitationMain(argc_, argv_, outputCout_, outputFiles_);
+                return flowBrineSaltPrecipitationMain(argc_, argv_, outputCout_, outputFiles_);
             }
         }
         else {
-            return flowEbosBrineMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowBrineMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
         return EXIT_FAILURE;
     }
 
-    int runSolvent()
+    int runSolvent(const Phases& phases)
     {
-        return flowEbosSolventMain(argc_, argv_, outputCout_, outputFiles_);
+        if (phases.active(Phase::FOAM)) {
+            return flowSolventFoamMain(argc_, argv_, outputCout_, outputFiles_);
+        }
+        // solvent + gas + water
+        if (!phases.active( Phase::OIL ) && phases.active( Phase::WATER ) && phases.active( Phase::GAS )) {
+            return flowGasWaterSolventMain(argc_, argv_, outputCout_, outputFiles_);
+        }
+
+        // solvent + gas + water + oil
+        if (phases.active( Phase::OIL ) && phases.active( Phase::WATER ) && phases.active( Phase::GAS )) {
+            return flowSolventMain(argc_, argv_, outputCout_, outputFiles_);
+        }
+
+        if (outputCout_)
+            std::cerr << "No valid configuration is found for solvent simulation, valid options include "
+                      << "gas + water + solvent and gas + oil + water + solvent" << std::endl;
+
+        return EXIT_FAILURE;
     }
 
     int runExtendedBlackOil()
     {
-        return flowEbosExtboMain(argc_, argv_, outputCout_, outputFiles_);
+        return flowExtboMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
     int runThermal(const Phases& phases)
     {
         // oil-gas-thermal
         if (!phases.active( Phase::WATER ) && phases.active( Phase::OIL ) && phases.active( Phase::GAS )) {
-            return flowEbosGasOilEnergyMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowGasOilEnergyMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
         // water-gas-thermal
         if (!phases.active( Phase::OIL ) && phases.active( Phase::WATER ) && phases.active( Phase::GAS )) {
-            return flowEbosGasWaterEnergyMain(argc_, argv_, outputCout_, outputFiles_);
+
+            if (phases.active(Phase::BRINE)){
+                return flowGasWaterSaltprecEnergyMain(argc_, argv_, outputCout_, outputFiles_);
+            }
+            return flowGasWaterEnergyMain(argc_, argv_, outputCout_, outputFiles_);
         }
 
-        return flowEbosEnergyMain(argc_, argv_, outputCout_, outputFiles_);
+        return flowEnergyMain(argc_, argv_, outputCout_, outputFiles_);
     }
 
     int runBlackOil()
@@ -615,9 +671,9 @@ private:
         if (diffusive) {
             // Use the traditional linearizer, as the TpfaLinearizer does not
             // support the diffusion module yet.
-            return flowEbosBlackoilMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowBlackoilMain(argc_, argv_, outputCout_, outputFiles_);
         } else {
-            return flowEbosBlackoilTpfaMain(argc_, argv_, outputCout_, outputFiles_);
+            return flowBlackoilTpfaMain(argc_, argv_, outputCout_, outputFiles_);
         }
     }
 
@@ -627,47 +683,67 @@ private:
                   const bool init_from_restart_file,
                   const bool allRanksDbgPrtLog,
                   const std::string& parsingStrictness,
-                  const int mpiRank,
+                  const std::string& actionParsingStrictness,
+                  const std::string& inputSkipMode,
+                  const bool keepKeywords,
+                  const std::size_t numThreads,
                   const int output_param,
                   const std::string& parameters,
                   std::string_view moduleVersion,
                   std::string_view compileTimestamp);
 
-    void setupVanguard();
-
-    template<class TypeTag>
     static int getNumThreads()
     {
 
-        int threads = 1;
+        int threads;
 
 #ifdef _OPENMP
         // This function is called before the parallel OpenMP stuff gets initialized.
-        // That initialization happends after the deck is read and we want this message.
+        // That initialization happens after the deck is read and we want this message.
         // Hence we duplicate the code of setupParallelism to get the number of threads.
-        if (std::getenv("OMP_NUM_THREADS"))
-            threads =  omp_get_max_threads();
-        else
-            threads = std::min(2, omp_get_max_threads());
+        static bool first_time = true;
+        constexpr int default_threads = 2;
+        const int requested_threads = Parameters::Get<Parameters::ThreadsPerProcess>();
+        threads = requested_threads > 0 ? requested_threads : default_threads;
 
-        const int input_threads = EWOMS_GET_PARAM(TypeTag, int, ThreadsPerProcess);
+        const char* env_var = getenv("OMP_NUM_THREADS");
+        if (env_var) {
+            int omp_num_threads = -1;
+            auto result = std::from_chars(env_var, env_var + std::strlen(env_var), omp_num_threads);
+            const bool can_output = first_time && FlowGenericVanguard::comm().rank() == 0;
+            if (result.ec == std::errc() && omp_num_threads > 0) {
+                // Set threads to omp_num_threads if it was successfully parsed and is positive
+                threads = omp_num_threads;
+                if (can_output && requested_threads > 0) {
+                    std::cout << "Warning: Environment variable OMP_NUM_THREADS takes precedence over the --threads-per-process cmdline argument."
+                              << std::endl;
+                }
+            } else {
+                if (can_output) {
+                    std::cout << ("Warning: Invalid value for OMP_NUM_THREADS environment variable.") << std::endl;
+                }
+            }
+        }
 
-        if (input_threads > 0)
-            threads = std::min(input_threads, omp_get_max_threads());
+        first_time = false;
+#else
+        threads = 1;
 #endif
-
         return threads;
     }
 
 #if HAVE_DAMARIS
-    void setupDamaris(const std::string& outputDir,
-                      const bool enableDamarisOutputCollective);
+    void setupDamaris(const std::string& outputDir);
 #endif
 
+protected:
     int argc_{0};
     char** argv_{nullptr};
     bool outputCout_{false};
     bool outputFiles_{false};
+
+private:
+    bool ownMPI_{true}; //!< True if we "own" MPI and should init / finalize
     double setupTime_{0.0};
     std::string deckFilename_{};
     std::string flowProgName_{};
@@ -680,6 +756,8 @@ private:
     std::shared_ptr<EclipseState> eclipseState_{};
     std::shared_ptr<Schedule> schedule_{};
     std::shared_ptr<SummaryConfig> summaryConfig_{};
+    bool mpi_init_{true}; //!< True if MPI_Init should be called
+    bool mpi_finalize_{true}; //!< True if MPI_Finalize should be called
 
     // To demonstrate run with non_world_comm
     bool test_split_comm_ = false;
