@@ -19,7 +19,7 @@
 */
 #ifndef OPM_PARALLELOVERLAPPINGILU0_HEADER_INCLUDED
 #define OPM_PARALLELOVERLAPPINGILU0_HEADER_INCLUDED
-
+#include <opm/common/TimingMacros.hpp>
 #include <opm/simulators/linalg/MILU.hpp>
 #include <opm/simulators/linalg/PreconditionerWithUpdate.hpp>
 #include <dune/istl/paamg/smoother.hh>
@@ -144,6 +144,10 @@ public:
     using block_type = typename matrix_type::block_type;
     using size_type = typename matrix_type::size_type;
 
+    virtual bool hasPerfectUpdate() const override {
+        return true;
+    }
+
 protected:
     struct CRS
     {
