@@ -38,7 +38,7 @@ namespace detail
 
 //! Compute Blocked ILU0 decomposition, when we know junk ghost rows are located at the end of A
 template<class M>
-void ghost_last_bilu0_decomposition (M& A, size_t interiorSize)
+void ghost_last_bilu0_decomposition (M& A, std::size_t interiorSize)
 {
     // iterator types
     using rowiterator = typename M::RowIterator;
@@ -185,6 +185,35 @@ void convertToCRS(const M& A, CRS& lower, CRS& upper, InvVector& inv)
   assert(colcount == numUpper);
 }
 
+template <class PI>
+size_t set_interiorSize( [[maybe_unused]] size_t N, size_t interiorSize, [[maybe_unused]] const PI& comm)
+{
+    return interiorSize;
+}
+
+#if HAVE_MPI
+template<>
+size_t set_interiorSize(size_t N, size_t interiorSize, const Dune::OwnerOverlapCopyCommunication<int,int>& comm)
+{
+    if (interiorSize<N)
+        return interiorSize;
+    auto indexSet = comm.indexSet();
+
+    size_t new_is = 0;
+    for (auto idx = indexSet.begin(); idx!=indexSet.end(); ++idx)
+    {
+        if (idx->local().attribute()==1)
+        {
+            auto loc = idx->local().local();
+            if (loc > new_is) {
+                new_is = loc;
+            }
+        }
+    }
+    return new_is + 1;
+}
+#endif
+
 } // end namespace detail
 
 
@@ -411,6 +440,11 @@ update()
     {
         OPM_TIMEBLOCK(iluDecomposition);
         if (iluIteration_ == 0) {
+
+            if (comm_) {
+                interiorSize_ = detail::set_interiorSize(A_->N(), interiorSize_, *comm_);
+            }
+
             // create ILU-0 decomposition
             if (ordering_.empty())
             {
@@ -419,7 +453,7 @@ update()
                     // The ILU_ matrix is already a copy with the same
                     // sparse structure as A_, but the values of A_ may
                     // have changed, so we must copy all elements.
-                    for (size_t row = 0; row < A_->N(); ++row) {
+                    for (std::size_t row = 0; row < A_->N(); ++row) {
                         const auto& Arow = (*A_)[row];
                         auto& ILUrow = (*ILU_)[row];
                         auto Ait = Arow.begin();
