@@ -27,6 +27,9 @@
 #include <opm/simulators/linalg/twolevelmethodcpr.hh>
 
 #include <dune/istl/paamg/pinfo.hh>
+#include <opm/simulators/linalg/WellOperators.hpp>
+
+#include <cstddef>
 
 namespace Opm
 {
@@ -44,22 +47,22 @@ namespace Opm
         const int max_nw = comm.communicator().max(nw);
         const int rank = comm.communicator().rank();
         int glo_max = 0;
-        size_t loc_max = 0;
+        std::size_t loc_max = 0;
         indset_rw.beginResize();
         for (auto ind = indset.begin(), indend = indset.end(); ind != indend; ++ind) {
             indset_rw.add(ind->global(), LocalIndex(ind->local(), ind->local().attribute(), true));
             const int glo = ind->global();
-            const size_t loc = ind->local().local();
+            const std::size_t loc = ind->local().local();
             glo_max = std::max(glo_max, glo);
             loc_max = std::max(loc_max, loc);
         }
         const int global_max = comm.communicator().max(glo_max);
         // used append the welldofs at the end
         assert(loc_max + 1 == indset.size());
-        size_t local_ind = loc_max + 1;
+        std::size_t local_ind = loc_max + 1;
         for (int i = 0; i < nw; ++i) {
             // need to set unique global number
-            const size_t v = global_max + max_nw * rank + i + 1;
+            const std::size_t v = global_max + max_nw * rank + i + 1;
             // set to true to not have problems with higher levels if growing of domains is used
             indset_rw.add(v, LocalIndex(local_ind, Dune::OwnerOverlapCopyAttributeSet::owner, true));
             ++local_ind;
@@ -74,31 +77,36 @@ namespace Opm
 
     namespace Details
     {
-        using PressureMatrixType = Dune::BCRSMatrix<Opm::MatrixBlock<double, 1, 1>>;
-        using PressureVectorType = Dune::BlockVector<Dune::FieldVector<double, 1>>;
-        using SeqCoarseOperatorType = Dune::MatrixAdapter<PressureMatrixType, PressureVectorType, PressureVectorType>;
-        template <class Comm>
+        template<class Scalar> using PressureMatrixType = Dune::BCRSMatrix<MatrixBlock<Scalar, 1, 1>>;
+        template<class Scalar> using PressureVectorType = Dune::BlockVector<Dune::FieldVector<Scalar, 1>>;
+        template<class Scalar> using SeqCoarseOperatorType = Dune::MatrixAdapter<PressureMatrixType<Scalar>,
+                                                                                 PressureVectorType<Scalar>,
+                                                                                 PressureVectorType<Scalar>>;
+        template<class Scalar, class Comm>
         using ParCoarseOperatorType
-            = Dune::OverlappingSchwarzOperator<PressureMatrixType, PressureVectorType, PressureVectorType, Comm>;
-        template <class Comm>
+            = Opm::GhostLastMatrixAdapter<PressureMatrixType<Scalar>,
+                                           PressureVectorType<Scalar>,
+                                           PressureVectorType<Scalar>,
+                                           Comm>;
+        template<class Scalar, class Comm>
         using CoarseOperatorType = std::conditional_t<std::is_same<Comm, Dune::Amg::SequentialInformation>::value,
-                                                      SeqCoarseOperatorType,
-                                                      ParCoarseOperatorType<Comm>>;
+                                                      SeqCoarseOperatorType<Scalar>,
+                                                      ParCoarseOperatorType<Scalar,Comm>>;
     } // namespace Details
 
-    template <class FineOperator, class Communication, bool transpose = false>
-    class PressureBhpTransferPolicy : public Dune::Amg::LevelTransferPolicyCpr<FineOperator, Details::CoarseOperatorType<Communication>>
+    template<class FineOperator, class Communication, class Scalar, bool transpose = false>
+    class PressureBhpTransferPolicy : public Dune::Amg::LevelTransferPolicyCpr<FineOperator, Details::CoarseOperatorType<Scalar,Communication>>
     {
     public:
-        typedef typename Details::CoarseOperatorType<Communication> CoarseOperator;
-        typedef Dune::Amg::LevelTransferPolicyCpr<FineOperator, CoarseOperator> ParentType;
-        typedef Communication ParallelInformation;
-        typedef typename FineOperator::domain_type FineVectorType;
+        using CoarseOperator = typename Details::CoarseOperatorType<Scalar,Communication>;
+        using ParentType = Dune::Amg::LevelTransferPolicyCpr<FineOperator, CoarseOperator>;
+        using ParallelInformation = Communication;
+        using FineVectorType= typename FineOperator::domain_type;
 
     public:
         PressureBhpTransferPolicy(const Communication& comm,
                                   const FineVectorType& weights,
-                                  const Opm::PropertyTree& prm,
+                                  const PropertyTree& prm,
                                   const std::size_t pressureIndex)
             : communication_(&const_cast<Communication&>(comm))
             , weights_(weights)
@@ -107,15 +115,15 @@ namespace Opm
         {
         }
 
-        virtual void createCoarseLevelSystem(const FineOperator& fineOperator) override
+        void createCoarseLevelSystem(const FineOperator& fineOperator) override
         {
             OPM_TIMEBLOCK(createCoarseLevelSystem);
             using CoarseMatrix = typename CoarseOperator::matrix_type;
             const auto& fineLevelMatrix = fineOperator.getmat();
             const auto& nw = fineOperator.getNumberOfExtraEquations();
             if (prm_.get<bool>("add_wells")) {
-                const size_t average_elements_per_row
-                    = static_cast<size_t>(std::ceil(fineLevelMatrix.nonzeroes() / fineLevelMatrix.N()));
+                const std::size_t average_elements_per_row
+                    = static_cast<std::size_t>(std::ceil(fineLevelMatrix.nonzeroes() / fineLevelMatrix.N()));
                 const double overflow_fraction = 1.2;
                 coarseLevelMatrix_.reset(new CoarseMatrix(fineLevelMatrix.N() + nw,
                                                           fineLevelMatrix.M() + nw,
@@ -162,7 +170,7 @@ namespace Opm
         this->operator_ = Dune::Amg::ConstructionTraits<CoarseOperator>::construct(oargs);
     }
 
-    virtual void calculateCoarseEntries(const FineOperator& fineOperator) override
+    void calculateCoarseEntries(const FineOperator& fineOperator) override
     {
         OPM_TIMEBLOCK(calculateCoarseEntries);
         const auto& fineMatrix = fineOperator.getmat();
@@ -173,15 +181,15 @@ namespace Opm
             auto entryCoarse = rowCoarse->begin();
             for (auto entry = row->begin(), entryEnd = row->end(); entry != entryEnd; ++entry, ++entryCoarse) {
                 assert(entry.index() == entryCoarse.index());
-                double matrix_el = 0;
+                Scalar matrix_el = 0;
                 if (transpose) {
                     const auto& bw = weights_[entry.index()];
-                    for (size_t i = 0; i < bw.size(); ++i) {
+                    for (std::size_t i = 0; i < bw.size(); ++i) {
                         matrix_el += (*entry)[pressure_var_index_][i] * bw[i];
                     }
                 } else {
                     const auto& bw = weights_[row.index()];
-                    for (size_t i = 0; i < bw.size(); ++i) {
+                    for (std::size_t i = 0; i < bw.size(); ++i) {
                         matrix_el += (*entry)[i][pressure_var_index_] * bw[i];
                     }
                 }
@@ -201,7 +209,7 @@ namespace Opm
         }
     }
 
-    virtual void moveToCoarseLevel(const typename ParentType::FineRangeType& fine) override
+    void moveToCoarseLevel(const typename ParentType::FineRangeType& fine) override
     {
         OPM_TIMEBLOCK(moveToCoarseLevel);
         //NB we iterate over fine assumming welldofs is at the end
@@ -212,11 +220,11 @@ namespace Opm
 
         for (auto block = begin; block != end; ++block) {
             const auto& bw = weights_[block.index()];
-            double rhs_el = 0.0;
+            Scalar rhs_el = 0.0;
             if (transpose) {
                 rhs_el = (*block)[pressure_var_index_];
             } else {
-                for (size_t i = 0; i < block->size(); ++i) {
+                for (std::size_t i = 0; i < block->size(); ++i) {
                     rhs_el += (*block)[i] * bw[i];
                 }
             }
@@ -226,7 +234,7 @@ namespace Opm
         this->lhs_ = 0;
     }
 
-    virtual void moveToFineLevel(typename ParentType::FineDomainType& fine) override
+    void moveToFineLevel(typename ParentType::FineDomainType& fine) override
     {
         OPM_TIMEBLOCK(moveToFineLevel);
         //NB we iterate over fine assumming welldofs is at the end
@@ -235,7 +243,7 @@ namespace Opm
         for (auto block = begin; block != end; ++block) {
             if (transpose) {
                 const auto& bw = weights_[block.index()];
-                for (size_t i = 0; i < block->size(); ++i) {
+                for (std::size_t i = 0; i < block->size(); ++i) {
                     (*block)[i] = this->lhs_[block - begin] * bw[i];
                 }
             } else {
@@ -244,7 +252,7 @@ namespace Opm
         }
     }
 
-    virtual PressureBhpTransferPolicy* clone() const override
+    PressureBhpTransferPolicy* clone() const override
     {
         return new PressureBhpTransferPolicy(*this);
     }
