@@ -18,60 +18,78 @@
 */
 
 #include <config.h>
-#include <cmath>
-#include <sstream>
+#include <opm/simulators/linalg/bda/opencl/openclKernels.hpp>
 
 #include <opm/common/OpmLog/OpmLog.hpp>
 #include <opm/common/ErrorMacros.hpp>
 #include <dune/common/timer.hh>
 
-#include <opm/simulators/linalg/bda/opencl/openclKernels.hpp>
 #include <opm/simulators/linalg/bda/opencl/ChowPatelIlu.hpp>  // defines CHOW_PATEL
 
-namespace Opm
-{
-namespace Accelerator
-{
+#include <opm/simulators/linalg/bda/Misc.hpp>
+
+#include <cmath>
+#include <sstream>
+
+namespace Opm::Accelerator {
 
 using Opm::OpmLog;
 using Dune::Timer;
 
 // define static variables and kernels
-int OpenclKernels::verbosity;
-cl::CommandQueue *OpenclKernels::queue;
-std::vector<double> OpenclKernels::tmp;
-bool OpenclKernels::initialized = false;
-std::size_t OpenclKernels::preferred_workgroup_size_multiple = 0;
+template<class Scalar> int OpenclKernels<Scalar>::verbosity;
+template<class Scalar> cl::CommandQueue* OpenclKernels<Scalar>::queue;
+template<class Scalar> std::vector<Scalar> OpenclKernels<Scalar>::tmp;
+template<class Scalar> bool OpenclKernels<Scalar>::initialized = false;
+template<class Scalar> std::size_t OpenclKernels<Scalar>::preferred_workgroup_size_multiple = 0;
 
-std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int, cl::LocalSpaceArg> > OpenclKernels::dot_k;
-std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, const unsigned int, cl::LocalSpaceArg> > OpenclKernels::norm_k;
-std::unique_ptr<cl::KernelFunctor<cl::Buffer&, const double, cl::Buffer&, const unsigned int> > OpenclKernels::axpy_k;
-std::unique_ptr<cl::KernelFunctor<cl::Buffer&, const double, const unsigned int> > OpenclKernels::scale_k;
-std::unique_ptr<cl::KernelFunctor<const double, cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int> > OpenclKernels::vmul_k;
-std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, cl::Buffer&, const double, const double, const unsigned int> > OpenclKernels::custom_k;
-std::unique_ptr<cl::KernelFunctor<const cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int> > OpenclKernels::full_to_pressure_restriction_k;
-std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, const unsigned int, const unsigned int> > OpenclKernels::add_coarse_pressure_correction_k;
-std::unique_ptr<cl::KernelFunctor<const cl::Buffer&, cl::Buffer&, const cl::Buffer&, const unsigned int> > OpenclKernels::prolongate_vector_k;
-std::unique_ptr<spmv_blocked_kernel_type> OpenclKernels::spmv_blocked_k;
-std::unique_ptr<spmv_blocked_kernel_type> OpenclKernels::spmv_blocked_add_k;
-std::unique_ptr<spmv_kernel_type> OpenclKernels::spmv_k;
-std::unique_ptr<spmv_kernel_type> OpenclKernels::spmv_noreset_k;
-std::unique_ptr<residual_blocked_kernel_type> OpenclKernels::residual_blocked_k;
-std::unique_ptr<residual_kernel_type> OpenclKernels::residual_k;
-std::unique_ptr<ilu_apply1_kernel_type> OpenclKernels::ILU_apply1_k;
-std::unique_ptr<ilu_apply2_kernel_type> OpenclKernels::ILU_apply2_k;
-std::unique_ptr<stdwell_apply_kernel_type> OpenclKernels::stdwell_apply_k;
-std::unique_ptr<ilu_decomp_kernel_type> OpenclKernels::ilu_decomp_k;
-std::unique_ptr<isaiL_kernel_type> OpenclKernels::isaiL_k;
-std::unique_ptr<isaiU_kernel_type> OpenclKernels::isaiU_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int, cl::LocalSpaceArg> > OpenclKernels<Scalar>::dot_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, const unsigned int, cl::LocalSpaceArg> > OpenclKernels<Scalar>::norm_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<cl::Buffer&, const Scalar, cl::Buffer&, const unsigned int> > OpenclKernels<Scalar>::axpy_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<cl::Buffer&, const Scalar, const unsigned int> > OpenclKernels<Scalar>::scale_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<const Scalar, cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int> > OpenclKernels<Scalar>::vmul_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, cl::Buffer&, const Scalar, const Scalar, const unsigned int> > OpenclKernels<Scalar>::custom_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<const cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int> > OpenclKernels<Scalar>::full_to_pressure_restriction_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<cl::Buffer&, cl::Buffer&, const unsigned int, const unsigned int> > OpenclKernels<Scalar>::add_coarse_pressure_correction_k;
+template<class Scalar>
+std::unique_ptr<cl::KernelFunctor<const cl::Buffer&, cl::Buffer&, const cl::Buffer&, const unsigned int> > OpenclKernels<Scalar>::prolongate_vector_k;
+template<class Scalar>
+std::unique_ptr<spmv_blocked_kernel_type> OpenclKernels<Scalar>::spmv_blocked_k;
+template<class Scalar>
+std::unique_ptr<spmv_blocked_kernel_type> OpenclKernels<Scalar>::spmv_blocked_add_k;
+template<class Scalar>
+std::unique_ptr<spmv_kernel_type> OpenclKernels<Scalar>::spmv_k;
+template<class Scalar>
+std::unique_ptr<spmv_kernel_type> OpenclKernels<Scalar>::spmv_noreset_k;
+template<class Scalar>
+std::unique_ptr<residual_blocked_kernel_type> OpenclKernels<Scalar>::residual_blocked_k;
+template<class Scalar>
+std::unique_ptr<residual_kernel_type> OpenclKernels<Scalar>::residual_k;
+template<class Scalar>
+std::unique_ptr<ilu_apply1_kernel_type> OpenclKernels<Scalar>::ILU_apply1_k;
+template<class Scalar>
+std::unique_ptr<ilu_apply2_kernel_type> OpenclKernels<Scalar>::ILU_apply2_k;
+template<class Scalar>
+std::unique_ptr<stdwell_apply_kernel_type> OpenclKernels<Scalar>::stdwell_apply_k;
+template<class Scalar>
+std::unique_ptr<ilu_decomp_kernel_type> OpenclKernels<Scalar>::ilu_decomp_k;
+template<class Scalar>
+std::unique_ptr<isaiL_kernel_type> OpenclKernels<Scalar>::isaiL_k;
+template<class Scalar>
+std::unique_ptr<isaiU_kernel_type> OpenclKernels<Scalar>::isaiU_k;
 
-// divide A by B, and round up: return (int)ceil(A/B)
-unsigned int ceilDivision(const unsigned int A, const unsigned int B)
-{
-    return A / B + (A % B > 0);
-}
-
-void OpenclKernels::init(cl::Context *context, cl::CommandQueue *queue_, std::vector<cl::Device>& devices, int verbosity_)
+template<class Scalar>
+void OpenclKernels<Scalar>::init(cl::Context *context,
+                                 cl::CommandQueue *queue_,
+                                 std::vector<cl::Device>& devices, int verbosity_)
 {
     if (initialized) {
         OpmLog::debug("Warning OpenclKernels is already initialized");
@@ -118,10 +136,10 @@ void OpenclKernels::init(cl::Context *context, cl::CommandQueue *queue_, std::ve
     // actually creating the kernels
     dot_k.reset(new cl::KernelFunctor<cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int, cl::LocalSpaceArg>(cl::Kernel(program, "dot_1")));
     norm_k.reset(new cl::KernelFunctor<cl::Buffer&, cl::Buffer&, const unsigned int, cl::LocalSpaceArg>(cl::Kernel(program, "norm")));
-    axpy_k.reset(new cl::KernelFunctor<cl::Buffer&, const double, cl::Buffer&, const unsigned int>(cl::Kernel(program, "axpy")));
-    scale_k.reset(new cl::KernelFunctor<cl::Buffer&, const double, const unsigned int>(cl::Kernel(program, "scale")));
-    vmul_k.reset(new cl::KernelFunctor<const double, cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int>(cl::Kernel(program, "vmul")));
-    custom_k.reset(new cl::KernelFunctor<cl::Buffer&, cl::Buffer&, cl::Buffer&, const double, const double, const unsigned int>(cl::Kernel(program, "custom")));
+    axpy_k.reset(new cl::KernelFunctor<cl::Buffer&, const Scalar, cl::Buffer&, const unsigned int>(cl::Kernel(program, "axpy")));
+    scale_k.reset(new cl::KernelFunctor<cl::Buffer&, const Scalar, const unsigned int>(cl::Kernel(program, "scale")));
+    vmul_k.reset(new cl::KernelFunctor<const Scalar, cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int>(cl::Kernel(program, "vmul")));
+    custom_k.reset(new cl::KernelFunctor<cl::Buffer&, cl::Buffer&, cl::Buffer&, const Scalar, const Scalar, const unsigned int>(cl::Kernel(program, "custom")));
     full_to_pressure_restriction_k.reset(new cl::KernelFunctor<const cl::Buffer&, cl::Buffer&, cl::Buffer&, const unsigned int>(cl::Kernel(program, "full_to_pressure_restriction")));
     add_coarse_pressure_correction_k.reset(new cl::KernelFunctor<cl::Buffer&, cl::Buffer&, const unsigned int, const unsigned int>(cl::Kernel(program, "add_coarse_pressure_correction")));
     prolongate_vector_k.reset(new cl::KernelFunctor<const cl::Buffer&, cl::Buffer&, const cl::Buffer&, const unsigned int>(cl::Kernel(program, "prolongate_vector")));
@@ -146,20 +164,21 @@ void OpenclKernels::init(cl::Context *context, cl::CommandQueue *queue_, std::ve
     initialized = true;
 } // end get_opencl_kernels()
 
-double OpenclKernels::dot(cl::Buffer& in1, cl::Buffer& in2, cl::Buffer& out, int N)
+template<class Scalar>
+Scalar OpenclKernels<Scalar>::dot(cl::Buffer& in1, cl::Buffer& in2, cl::Buffer& out, int N)
 {
     const unsigned int work_group_size = 256;
     const unsigned int num_work_groups = ceilDivision(N, work_group_size);
     const unsigned int total_work_items = num_work_groups * work_group_size;
-    const unsigned int lmem_per_work_group = sizeof(double) * work_group_size;
+    const unsigned int lmem_per_work_group = sizeof(Scalar) * work_group_size;
     Timer t_dot;
     tmp.resize(num_work_groups);
 
     cl::Event event = (*dot_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)), in1, in2, out, N, cl::Local(lmem_per_work_group));
 
-    queue->enqueueReadBuffer(out, CL_TRUE, 0, sizeof(double) * num_work_groups, tmp.data());
+    queue->enqueueReadBuffer(out, CL_TRUE, 0, sizeof(Scalar) * num_work_groups, tmp.data());
 
-    double gpu_sum = 0.0;
+    Scalar gpu_sum = 0.0;
     for (unsigned int i = 0; i < num_work_groups; ++i) {
         gpu_sum += tmp[i];
     }
@@ -174,20 +193,21 @@ double OpenclKernels::dot(cl::Buffer& in1, cl::Buffer& in2, cl::Buffer& out, int
     return gpu_sum;
 }
 
-double OpenclKernels::norm(cl::Buffer& in, cl::Buffer& out, int N)
+template<class Scalar>
+Scalar OpenclKernels<Scalar>::norm(cl::Buffer& in, cl::Buffer& out, int N)
 {
     const unsigned int work_group_size = 256;
     const unsigned int num_work_groups = ceilDivision(N, work_group_size);
     const unsigned int total_work_items = num_work_groups * work_group_size;
-    const unsigned int lmem_per_work_group = sizeof(double) * work_group_size;
+    const unsigned int lmem_per_work_group = sizeof(Scalar) * work_group_size;
     Timer t_norm;
     tmp.resize(num_work_groups);
 
     cl::Event event = (*norm_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)), in, out, N, cl::Local(lmem_per_work_group));
 
-    queue->enqueueReadBuffer(out, CL_TRUE, 0, sizeof(double) * num_work_groups, tmp.data());
+    queue->enqueueReadBuffer(out, CL_TRUE, 0, sizeof(Scalar) * num_work_groups, tmp.data());
 
-    double gpu_norm = 0.0;
+    Scalar gpu_norm = 0.0;
     for (unsigned int i = 0; i < num_work_groups; ++i) {
         gpu_norm += tmp[i];
     }
@@ -203,7 +223,8 @@ double OpenclKernels::norm(cl::Buffer& in, cl::Buffer& out, int N)
     return gpu_norm;
 }
 
-void OpenclKernels::axpy(cl::Buffer& in, const double a, cl::Buffer& out, int N)
+template<class Scalar>
+void OpenclKernels<Scalar>::axpy(cl::Buffer& in, const Scalar a, cl::Buffer& out, int N)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(N, work_group_size);
@@ -220,7 +241,8 @@ void OpenclKernels::axpy(cl::Buffer& in, const double a, cl::Buffer& out, int N)
     }
 }
 
-void OpenclKernels::scale(cl::Buffer& in, const double a, int N)
+template<class Scalar>
+void OpenclKernels<Scalar>::scale(cl::Buffer& in, const Scalar a, int N)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(N, work_group_size);
@@ -237,7 +259,8 @@ void OpenclKernels::scale(cl::Buffer& in, const double a, int N)
     }
 }
 
-void OpenclKernels::vmul(const double alpha, cl::Buffer& in1, cl::Buffer& in2, cl::Buffer& out, int N)
+template<class Scalar>
+void OpenclKernels<Scalar>::vmul(const Scalar alpha, cl::Buffer& in1, cl::Buffer& in2, cl::Buffer& out, int N)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(N, work_group_size);
@@ -254,8 +277,9 @@ void OpenclKernels::vmul(const double alpha, cl::Buffer& in1, cl::Buffer& in2, c
     }
 }
 
-void OpenclKernels::custom(cl::Buffer& p, cl::Buffer& v, cl::Buffer& r,
-                           const double omega, const double beta, int N)
+template<class Scalar>
+void OpenclKernels<Scalar>::custom(cl::Buffer& p, cl::Buffer& v, cl::Buffer& r,
+                                   const Scalar omega, const Scalar beta, int N)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(N, work_group_size);
@@ -272,7 +296,8 @@ void OpenclKernels::custom(cl::Buffer& p, cl::Buffer& v, cl::Buffer& r,
     }
 }
 
-void OpenclKernels::full_to_pressure_restriction(const cl::Buffer& fine_y, cl::Buffer& weights, cl::Buffer& coarse_y, int Nb)
+template<class Scalar>
+void OpenclKernels<Scalar>::full_to_pressure_restriction(const cl::Buffer& fine_y, cl::Buffer& weights, cl::Buffer& coarse_y, int Nb)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(Nb, work_group_size);
@@ -289,7 +314,8 @@ void OpenclKernels::full_to_pressure_restriction(const cl::Buffer& fine_y, cl::B
     }
 }
 
-void OpenclKernels::add_coarse_pressure_correction(cl::Buffer& coarse_x, cl::Buffer& fine_x, int pressure_idx, int Nb)
+template<class Scalar>
+void OpenclKernels<Scalar>::add_coarse_pressure_correction(cl::Buffer& coarse_x, cl::Buffer& fine_x, int pressure_idx, int Nb)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(Nb, work_group_size);
@@ -306,7 +332,8 @@ void OpenclKernels::add_coarse_pressure_correction(cl::Buffer& coarse_x, cl::Buf
     }
 }
 
-void OpenclKernels::prolongate_vector(const cl::Buffer& in, cl::Buffer& out, const cl::Buffer& cols, int N)
+template<class Scalar>
+void OpenclKernels<Scalar>::prolongate_vector(const cl::Buffer& in, cl::Buffer& out, const cl::Buffer& cols, int N)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(N, work_group_size);
@@ -323,32 +350,33 @@ void OpenclKernels::prolongate_vector(const cl::Buffer& in, cl::Buffer& out, con
     }
 }
 
-void OpenclKernels::spmv(cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& rows,
-                         const cl::Buffer& x, cl::Buffer& b, int Nb,
-                         unsigned int block_size, bool reset, bool add)
+template<class Scalar>
+void OpenclKernels<Scalar>::spmv(cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& rows,
+                                 const cl::Buffer& x, cl::Buffer& b, int Nb,
+                                 unsigned int block_size, bool reset, bool add)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(Nb, work_group_size);
     const unsigned int total_work_items = num_work_groups * work_group_size;
-    const unsigned int lmem_per_work_group = sizeof(double) * work_group_size;
+    const unsigned int lmem_per_work_group = sizeof(Scalar) * work_group_size;
     Timer t_spmv;
     cl::Event event;
 
     if (block_size > 1) {
         if (add) {
             event = (*spmv_blocked_add_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                        vals, cols, rows, Nb, x, b, block_size, cl::Local(lmem_per_work_group));
+                                          vals, cols, rows, Nb, x, b, block_size, cl::Local(lmem_per_work_group));
         } else {
             event = (*spmv_blocked_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                        vals, cols, rows, Nb, x, b, block_size, cl::Local(lmem_per_work_group));
+                                      vals, cols, rows, Nb, x, b, block_size, cl::Local(lmem_per_work_group));
         }
     } else {
         if (reset) {
             event = (*spmv_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                        vals, cols, rows, Nb, x, b, cl::Local(lmem_per_work_group));
+                              vals, cols, rows, Nb, x, b, cl::Local(lmem_per_work_group));
         } else {
             event = (*spmv_noreset_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                        vals, cols, rows, Nb, x, b, cl::Local(lmem_per_work_group));
+                                      vals, cols, rows, Nb, x, b, cl::Local(lmem_per_work_group));
         }
     }
 
@@ -360,23 +388,24 @@ void OpenclKernels::spmv(cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& rows,
     }
 }
 
-void OpenclKernels::residual(cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& rows,
-                            cl::Buffer& x, const cl::Buffer& rhs,
-                            cl::Buffer& out, int Nb, unsigned int block_size)
+template<class Scalar>
+void OpenclKernels<Scalar>::residual(cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& rows,
+                                     cl::Buffer& x, const cl::Buffer& rhs,
+                                     cl::Buffer& out, int Nb, unsigned int block_size)
 {
     const unsigned int work_group_size = 32;
     const unsigned int num_work_groups = ceilDivision(Nb, work_group_size);
     const unsigned int total_work_items = num_work_groups * work_group_size;
-    const unsigned int lmem_per_work_group = sizeof(double) * work_group_size;
+    const unsigned int lmem_per_work_group = sizeof(Scalar) * work_group_size;
     Timer t_residual;
     cl::Event event;
 
     if (block_size > 1) {
         event = (*residual_blocked_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                    vals, cols, rows, Nb, x, rhs, out, block_size, cl::Local(lmem_per_work_group));
+                                      vals, cols, rows, Nb, x, rhs, out, block_size, cl::Local(lmem_per_work_group));
     } else {
         event = (*residual_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                    vals, cols, rows, Nb, x, rhs, out, cl::Local(lmem_per_work_group));
+                              vals, cols, rows, Nb, x, rhs, out, cl::Local(lmem_per_work_group));
     }
 
     if (verbosity >= 4) {
@@ -387,22 +416,23 @@ void OpenclKernels::residual(cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& row
     }
 }
 
-void OpenclKernels::ILU_apply1(cl::Buffer& rowIndices, cl::Buffer& vals, cl::Buffer& cols,
-                               cl::Buffer& rows, cl::Buffer& diagIndex,
-                               const cl::Buffer& y, cl::Buffer& x,
-                               cl::Buffer& rowsPerColor, int color,
-                               int rowsThisColor, unsigned int block_size)
+template<class Scalar>
+void OpenclKernels<Scalar>::ILU_apply1(cl::Buffer& rowIndices, cl::Buffer& vals, cl::Buffer& cols,
+                                       cl::Buffer& rows, cl::Buffer& diagIndex,
+                                       const cl::Buffer& y, cl::Buffer& x,
+                                       cl::Buffer& rowsPerColor, int color,
+                                       int rowsThisColor, unsigned int block_size)
 {
     const unsigned int work_group_size = preferred_workgroup_size_multiple;
     const unsigned int num_work_groups = rowsThisColor;
     const unsigned int total_work_items = num_work_groups * work_group_size;
-    const unsigned int lmem_per_work_group = sizeof(double) * work_group_size;
+    const unsigned int lmem_per_work_group = sizeof(Scalar) * work_group_size;
     Timer t_ilu_apply1;
 
     cl::Event event = (*ILU_apply1_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                          rowIndices, vals, cols, rows, diagIndex,
-                          y, x, rowsPerColor, color, block_size,
-                          cl::Local(lmem_per_work_group));
+                                      rowIndices, vals, cols, rows, diagIndex,
+                                      y, x, rowsPerColor, color, block_size,
+                                      cl::Local(lmem_per_work_group));
 
     if (verbosity >= 5) {
         event.wait();
@@ -412,22 +442,23 @@ void OpenclKernels::ILU_apply1(cl::Buffer& rowIndices, cl::Buffer& vals, cl::Buf
     }
 }
 
-void OpenclKernels::ILU_apply2(cl::Buffer& rowIndices, cl::Buffer& vals, cl::Buffer& cols,
-                               cl::Buffer& rows, cl::Buffer& diagIndex,
-                               cl::Buffer& invDiagVals, cl::Buffer& x,
-                               cl::Buffer& rowsPerColor, int color,
-                               int rowsThisColor, unsigned int block_size)
+template<class Scalar>
+void OpenclKernels<Scalar>::ILU_apply2(cl::Buffer& rowIndices, cl::Buffer& vals, cl::Buffer& cols,
+                                       cl::Buffer& rows, cl::Buffer& diagIndex,
+                                       cl::Buffer& invDiagVals, cl::Buffer& x,
+                                       cl::Buffer& rowsPerColor, int color,
+                                       int rowsThisColor, unsigned int block_size)
 {
     const unsigned int work_group_size = preferred_workgroup_size_multiple;
     const unsigned int num_work_groups = rowsThisColor;
     const unsigned int total_work_items = num_work_groups * work_group_size;
-    const unsigned int lmem_per_work_group = sizeof(double) * work_group_size;
+    const unsigned int lmem_per_work_group = sizeof(Scalar) * work_group_size;
     Timer t_ilu_apply2;
 
     cl::Event event = (*ILU_apply2_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                          rowIndices, vals, cols, rows, diagIndex,
-                          invDiagVals, x, rowsPerColor, color, block_size,
-                          cl::Local(lmem_per_work_group));
+                                      rowIndices, vals, cols, rows, diagIndex,
+                                      invDiagVals, x, rowsPerColor, color, block_size,
+                                      cl::Local(lmem_per_work_group));
 
     if (verbosity >= 5) {
         event.wait();
@@ -437,23 +468,24 @@ void OpenclKernels::ILU_apply2(cl::Buffer& rowIndices, cl::Buffer& vals, cl::Buf
     }
 }
 
-void OpenclKernels::ILU_decomp(int firstRow, int lastRow, cl::Buffer& rowIndices,
-                               cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& rows,
-                               cl::Buffer& diagIndex, cl::Buffer& invDiagVals,
-                               int rowsThisColor, unsigned int block_size)
+template<class Scalar>
+void OpenclKernels<Scalar>::ILU_decomp(int firstRow, int lastRow, cl::Buffer& rowIndices,
+                                       cl::Buffer& vals, cl::Buffer& cols, cl::Buffer& rows,
+                                       cl::Buffer& diagIndex, cl::Buffer& invDiagVals,
+                                       int rowsThisColor, unsigned int block_size)
 {
     const unsigned int work_group_size = 128;
     const unsigned int num_work_groups = rowsThisColor;
     const unsigned int total_work_items = num_work_groups * work_group_size;
     const unsigned int num_hwarps_per_group = work_group_size / 16;
-    const unsigned int lmem_per_work_group = num_hwarps_per_group * block_size * block_size * sizeof(double);           // each block needs a pivot
+    const unsigned int lmem_per_work_group = num_hwarps_per_group * block_size * block_size * sizeof(Scalar);           // each block needs a pivot
     Timer t_ilu_decomp;
 
     cl::Event event = (*ilu_decomp_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                          firstRow, lastRow, rowIndices,
-                          vals, cols, rows,
-                          invDiagVals, diagIndex, rowsThisColor,
-                          cl::Local(lmem_per_work_group));
+                                      firstRow, lastRow, rowIndices,
+                                      vals, cols, rows,
+                                      invDiagVals, diagIndex, rowsThisColor,
+                                      cl::Local(lmem_per_work_group));
 
     if (verbosity >= 4) {
         event.wait();
@@ -463,19 +495,20 @@ void OpenclKernels::ILU_decomp(int firstRow, int lastRow, cl::Buffer& rowIndices
     }
 }
 
-void OpenclKernels::apply_stdwells(cl::Buffer& d_Cnnzs_ocl, cl::Buffer &d_Dnnzs_ocl, cl::Buffer &d_Bnnzs_ocl,
-    cl::Buffer &d_Ccols_ocl, cl::Buffer &d_Bcols_ocl, cl::Buffer &d_x, cl::Buffer &d_y,
-    int dim, int dim_wells, cl::Buffer &d_val_pointers_ocl, int num_std_wells)
+template<class Scalar>
+void OpenclKernels<Scalar>::apply_stdwells(cl::Buffer& d_Cnnzs_ocl, cl::Buffer &d_Dnnzs_ocl, cl::Buffer &d_Bnnzs_ocl,
+                                           cl::Buffer &d_Ccols_ocl, cl::Buffer &d_Bcols_ocl, cl::Buffer &d_x, cl::Buffer &d_y,
+                                           int dim, int dim_wells, cl::Buffer &d_val_pointers_ocl, int num_std_wells)
 {
     const unsigned int work_group_size = 32;
     const unsigned int total_work_items = num_std_wells * work_group_size;
-    const unsigned int lmem1 = sizeof(double) * work_group_size;
-    const unsigned int lmem2 = sizeof(double) * dim_wells;
+    const unsigned int lmem1 = sizeof(Scalar) * work_group_size;
+    const unsigned int lmem2 = sizeof(Scalar) * dim_wells;
     Timer t_apply_stdwells;
 
     cl::Event event = (*stdwell_apply_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-                          d_Cnnzs_ocl, d_Dnnzs_ocl, d_Bnnzs_ocl, d_Ccols_ocl, d_Bcols_ocl, d_x, d_y, dim, dim_wells, d_val_pointers_ocl,
-                          cl::Local(lmem1), cl::Local(lmem2), cl::Local(lmem2));
+                                         d_Cnnzs_ocl, d_Dnnzs_ocl, d_Bnnzs_ocl, d_Ccols_ocl, d_Bcols_ocl, d_x, d_y, dim, dim_wells, d_val_pointers_ocl,
+                                         cl::Local(lmem1), cl::Local(lmem2), cl::Local(lmem2));
 
     if (verbosity >= 4) {
         event.wait();
@@ -485,8 +518,9 @@ void OpenclKernels::apply_stdwells(cl::Buffer& d_Cnnzs_ocl, cl::Buffer &d_Dnnzs_
     }
 }
 
-void OpenclKernels::isaiL(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Buffer& mapping, cl::Buffer& nvc,
-    cl::Buffer& luIdxs, cl::Buffer& xxIdxs, cl::Buffer& dxIdxs, cl::Buffer& LUvals, cl::Buffer& invLvals, unsigned int Nb)
+template<class Scalar>
+void OpenclKernels<Scalar>::isaiL(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Buffer& mapping, cl::Buffer& nvc,
+                                  cl::Buffer& luIdxs, cl::Buffer& xxIdxs, cl::Buffer& dxIdxs, cl::Buffer& LUvals, cl::Buffer& invLvals, unsigned int Nb)
 {
     const unsigned int work_group_size = 256;
     const unsigned int num_work_groups = ceilDivision(Nb, work_group_size);
@@ -494,7 +528,7 @@ void OpenclKernels::isaiL(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Bu
 
     Timer t_isaiL;
     cl::Event event = (*isaiL_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-            diagIndex, colPointers, mapping, nvc, luIdxs, xxIdxs, dxIdxs, LUvals, invLvals, Nb);
+                                 diagIndex, colPointers, mapping, nvc, luIdxs, xxIdxs, dxIdxs, LUvals, invLvals, Nb);
 
     if (verbosity >= 4) {
         event.wait();
@@ -504,9 +538,10 @@ void OpenclKernels::isaiL(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Bu
     }
 }
 
-void OpenclKernels::isaiU(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Buffer& rowIndices, cl::Buffer& mapping,
-        cl::Buffer& nvc, cl::Buffer& luIdxs, cl::Buffer& xxIdxs, cl::Buffer& dxIdxs, cl::Buffer& LUvals,
-        cl::Buffer& invDiagVals, cl::Buffer& invUvals, unsigned int Nb)
+template<class Scalar>
+void OpenclKernels<Scalar>::isaiU(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Buffer& rowIndices, cl::Buffer& mapping,
+                                  cl::Buffer& nvc, cl::Buffer& luIdxs, cl::Buffer& xxIdxs, cl::Buffer& dxIdxs, cl::Buffer& LUvals,
+                                  cl::Buffer& invDiagVals, cl::Buffer& invUvals, unsigned int Nb)
 {
     const unsigned int work_group_size = 256;
     const unsigned int num_work_groups = ceilDivision(Nb, work_group_size);
@@ -514,7 +549,7 @@ void OpenclKernels::isaiU(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Bu
 
     Timer t_isaiU;
     cl::Event event = (*isaiU_k)(cl::EnqueueArgs(*queue, cl::NDRange(total_work_items), cl::NDRange(work_group_size)),
-            diagIndex, colPointers, rowIndices, mapping, nvc, luIdxs, xxIdxs, dxIdxs, LUvals, invDiagVals, invUvals, Nb);
+                                 diagIndex, colPointers, rowIndices, mapping, nvc, luIdxs, xxIdxs, dxIdxs, LUvals, invDiagVals, invUvals, Nb);
 
     if (verbosity >= 4) {
         event.wait();
@@ -524,5 +559,10 @@ void OpenclKernels::isaiU(cl::Buffer& diagIndex, cl::Buffer& colPointers, cl::Bu
     }
 }
 
-} // namespace Accelerator
-} // namespace Opm
+template class OpenclKernels<double>;
+
+#if FLOW_INSTANTIATE_FLOAT
+template class OpenclKernels<float>;
+#endif
+
+} // namespace Opm::Accelerator
