@@ -37,41 +37,47 @@
 // otherwise, the nonzeroes of the matrix are assumed to be in a contiguous array, and a single GPU memcpy is enough
 #define COPY_ROW_BY_ROW 0
 
-namespace Opm
-{
-namespace Accelerator
-{
+namespace Opm::Accelerator {
 
-using Opm::OpmLog;
 using Dune::Timer;
 
-template <unsigned int block_size>
-openclSolverBackend<block_size>::openclSolverBackend(int verbosity_, int maxit_, double tolerance_, unsigned int platformID_, unsigned int deviceID_, bool opencl_ilu_parallel_, std::string linsolver) : BdaSolver<block_size>(verbosity_, maxit_, tolerance_, platformID_, deviceID_), opencl_ilu_parallel(opencl_ilu_parallel_) {
-
+template<class Scalar, unsigned int block_size>
+openclSolverBackend<Scalar,block_size>::
+openclSolverBackend(int verbosity_,
+                    int maxit_,
+                    Scalar tolerance_,
+                    unsigned int platformID_,
+                    unsigned int deviceID_,
+                    bool opencl_ilu_parallel_,
+                    std::string linsolver)
+    : Base(verbosity_, maxit_, tolerance_, platformID_, deviceID_)
+    , opencl_ilu_parallel(opencl_ilu_parallel_)
+{
     bool use_cpr, use_isai;
 
-    if (linsolver.compare("ilu0") == 0) {
+    if (linsolver == "ilu0") {
         use_cpr = false;
         use_isai = false;
-    } else if (linsolver.compare("cpr_quasiimpes") == 0) {
+    } else if (linsolver == "cpr_quasiimpes") {
         use_cpr = true;
         use_isai = false;
-    } else if (linsolver.compare("isai") == 0) {
+    } else if (linsolver == "isai") {
         use_cpr = false;
         use_isai = true;
-    } else if (linsolver.compare("cpr_trueimpes") == 0) {
-        OPM_THROW(std::logic_error, "Error openclSolver does not support --linerar-solver=cpr_trueimpes");
+    } else if (linsolver == "cpr_trueimpes") {
+        OPM_THROW(std::logic_error, "Error openclSolver does not support "
+                                    "--linear-solver=cpr_trueimpes");
     } else {
         OPM_THROW(std::logic_error, "Error unknown value for argument --linear-solver, " + linsolver);
     }
 
-    using PreconditionerType = typename Preconditioner<block_size>::PreconditionerType;
+    using PCType = typename Opm::Accelerator::PreconditionerType;
     if (use_cpr) {
-        prec = Preconditioner<block_size>::create(PreconditionerType::CPR, verbosity, opencl_ilu_parallel);
+        prec = openclPreconditioner<Scalar, block_size>::create(PCType::CPR, verbosity, opencl_ilu_parallel);
     } else if (use_isai) {
-        prec = Preconditioner<block_size>::create(PreconditionerType::BISAI, verbosity, opencl_ilu_parallel);
+        prec = openclPreconditioner<Scalar, block_size>::create(PCType::BISAI, verbosity, opencl_ilu_parallel);
     } else {
-        prec = Preconditioner<block_size>::create(PreconditionerType::BILU0, verbosity, opencl_ilu_parallel);
+        prec = openclPreconditioner<Scalar, block_size>::create(PCType::BILU0, verbosity, opencl_ilu_parallel);
     }
 
     std::ostringstream out;
@@ -103,7 +109,7 @@ openclSolverBackend<block_size>::openclSolverBackend(int verbosity_, int maxit_,
         out.clear();
 
         if (platforms.size() <= platformID) {
-            OPM_THROW(std::logic_error, "Error chosen too high OpenCL platform ID");
+            OPM_THROW(std::logic_error, "Error: Invalid OpenCL platform ID selected");
         } else {
             std::string platform_info;
             out << "Chosen:\n";
@@ -119,7 +125,8 @@ openclSolverBackend<block_size>::openclSolverBackend(int verbosity_, int maxit_,
         platforms[platformID].getDevices(CL_DEVICE_TYPE_ALL, &devices);
 
         if (devices.empty()) {
-            OPM_THROW(std::logic_error, "Error openclSolver is selected but no OpenCL devices are found");
+            OPM_THROW(std::logic_error, "Error openclSolver is selected but "
+                                        "no OpenCL devices are found");
         }
         out << "Found " << devices.size() << " OpenCL devices" << "\n";
 
@@ -203,8 +210,7 @@ openclSolverBackend<block_size>::openclSolverBackend(int verbosity_, int maxit_,
         context = std::make_shared<cl::Context>(devices[0]);
         queue.reset(new cl::CommandQueue(*context, devices[0], 0, &err));
 
-        OpenclKernels::init(context.get(), queue.get(), devices, verbosity);
-
+        OpenclKernels<Scalar>::init(context.get(), queue.get(), devices, verbosity);
     } catch (const cl::Error& error) {
         std::ostringstream oss;
         oss << "OpenCL Error: " << error.what() << "(" << error.err() << ")\n";
@@ -217,26 +223,36 @@ openclSolverBackend<block_size>::openclSolverBackend(int verbosity_, int maxit_,
     }
 }
 
-template <unsigned int block_size>
-openclSolverBackend<block_size>::openclSolverBackend(int verbosity_, int maxit_, double tolerance_, bool opencl_ilu_parallel_) :
-    BdaSolver<block_size>(verbosity_, maxit_, tolerance_), opencl_ilu_parallel(opencl_ilu_parallel_)
+template<class Scalar, unsigned int block_size>
+openclSolverBackend<Scalar,block_size>::
+openclSolverBackend(int verbosity_,
+                    int maxit_,
+                    Scalar tolerance_,
+                    bool opencl_ilu_parallel_)
+    : Base(verbosity_, maxit_, tolerance_)
+    , opencl_ilu_parallel(opencl_ilu_parallel_)
 {
     // prec = std::make_unique<BILU0<block_size> >(opencl_ilu_parallel, verbosity_);
     // cpr = std::make_unique<CPR<block_size> >(verbosity_, opencl_ilu_parallel, /*use_amg=*/false);
 }
 
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::setOpencl(std::shared_ptr<cl::Context>& context_, std::shared_ptr<cl::CommandQueue>& queue_) {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+setOpencl(std::shared_ptr<cl::Context>& context_,
+          std::shared_ptr<cl::CommandQueue>& queue_)
+{
     context = context_;
     queue = queue_;
 }
 
-
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContribs, BdaResult& res) {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+gpu_pbicgstab(WellContributions<Scalar>& wellContribs,
+              BdaResult& res)
+{
     float it;
-    double rho, rhop, beta, alpha, omega, tmp1, tmp2;
-    double norm, norm_0;
+    Scalar rho, rhop, beta, alpha, omega, tmp1, tmp2;
+    Scalar norm, norm_0;
 
     Timer t_total, t_prec(false), t_spmv(false), t_well(false), t_rest(false);
 
@@ -246,15 +262,15 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
 
     // set initial values
     events.resize(5);
-    queue->enqueueFillBuffer(d_p, 0, 0, sizeof(double) * N, nullptr, &events[0]);
-    queue->enqueueFillBuffer(d_v, 0, 0, sizeof(double) * N, nullptr, &events[1]);
+    queue->enqueueFillBuffer(d_p, 0, 0, sizeof(Scalar) * N, nullptr, &events[0]);
+    queue->enqueueFillBuffer(d_v, 0, 0, sizeof(Scalar) * N, nullptr, &events[1]);
     rho = 1.0;
     alpha = 1.0;
     omega = 1.0;
 
-    queue->enqueueCopyBuffer(d_b, d_r, 0, 0, sizeof(double) * N, nullptr, &events[2]);
-    queue->enqueueCopyBuffer(d_r, d_rw, 0, 0, sizeof(double) * N, nullptr, &events[3]);
-    queue->enqueueCopyBuffer(d_r, d_p, 0, 0, sizeof(double) * N, nullptr, &events[4]);
+    queue->enqueueCopyBuffer(d_b, d_r, 0, 0, sizeof(Scalar) * N, nullptr, &events[2]);
+    queue->enqueueCopyBuffer(d_r, d_rw, 0, 0, sizeof(Scalar) * N, nullptr, &events[3]);
+    queue->enqueueCopyBuffer(d_r, d_p, 0, 0, sizeof(Scalar) * N, nullptr, &events[4]);
 
     cl::WaitForEvents(events);
     events.clear();
@@ -263,7 +279,7 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
         OPM_THROW(std::logic_error, "openclSolverBackend OpenCL enqueue[Fill|Copy]Buffer error");
     }
 
-    norm = OpenclKernels::norm(d_r, d_tmp, N);
+    norm = OpenclKernels<Scalar>::norm(d_r, d_tmp, N);
     norm_0 = norm;
 
     if (verbosity > 1) {
@@ -277,11 +293,11 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
     }
     for (it = 0.5; it < maxit; it += 0.5) {
         rhop = rho;
-        rho = OpenclKernels::dot(d_rw, d_r, d_tmp, N);
+        rho = OpenclKernels<Scalar>::dot(d_rw, d_r, d_tmp, N);
 
         if (it > 1) {
             beta = (rho / rhop) * (alpha / omega);
-            OpenclKernels::custom(d_p, d_v, d_r, omega, beta, N);
+            OpenclKernels<Scalar>::custom(d_p, d_v, d_r, omega, beta, N);
         }
         if (verbosity >= 3) {
             queue->finish();
@@ -298,7 +314,7 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
         }
 
         // v = A * pw
-        OpenclKernels::spmv(d_Avals, d_Acols, d_Arows, d_pw, d_v, Nb, block_size);
+        OpenclKernels<Scalar>::spmv(d_Avals, d_Acols, d_Arows, d_pw, d_v, Nb, block_size);
         if (verbosity >= 3) {
             queue->finish();
             t_spmv.stop();
@@ -306,20 +322,20 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
         }
 
         // apply wellContributions
-        if(wellContribs.getNumWells() > 0){
-            static_cast<WellContributionsOCL&>(wellContribs).apply(d_pw, d_v);
+        if (wellContribs.getNumWells() > 0) {
+            static_cast<WellContributionsOCL<Scalar>&>(wellContribs).apply(d_pw, d_v);
         }
-        if(verbosity >= 3) {
+        if (verbosity >= 3) {
             queue->finish();
             t_well.stop();
             t_rest.start();
         }
 
-        tmp1 = OpenclKernels::dot(d_rw, d_v, d_tmp, N);
+        tmp1 = OpenclKernels<Scalar>::dot(d_rw, d_v, d_tmp, N);
         alpha = rho / tmp1;
-        OpenclKernels::axpy(d_v, -alpha, d_r, N);      // r = r - alpha * v
-        OpenclKernels::axpy(d_pw, alpha, d_x, N);      // x = x + alpha * pw
-        norm = OpenclKernels::norm(d_r, d_tmp, N);
+        OpenclKernels<Scalar>::axpy(d_v, -alpha, d_r, N);      // r = r - alpha * v
+        OpenclKernels<Scalar>::axpy(d_pw, alpha, d_x, N);      // x = x + alpha * pw
+        norm = OpenclKernels<Scalar>::norm(d_r, d_tmp, N);
         if (verbosity >= 3) {
             queue->finish();
             t_rest.stop();
@@ -343,8 +359,8 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
         }
 
         // t = A * s
-        OpenclKernels::spmv(d_Avals, d_Acols, d_Arows, d_s, d_t, Nb, block_size);
-        if(verbosity >= 3){
+        OpenclKernels<Scalar>::spmv(d_Avals, d_Acols, d_Arows, d_s, d_t, Nb, block_size);
+        if (verbosity >= 3) {
             queue->finish();
             t_spmv.stop();
             t_well.start();
@@ -352,7 +368,7 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
 
         // apply wellContributions
         if(wellContribs.getNumWells() > 0){
-            static_cast<WellContributionsOCL&>(wellContribs).apply(d_s, d_t);
+            static_cast<WellContributionsOCL<Scalar>&>(wellContribs).apply(d_s, d_t);
         }
         if (verbosity >= 3) {
             queue->finish();
@@ -360,12 +376,12 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
             t_rest.start();
         }
 
-        tmp1 = OpenclKernels::dot(d_t, d_r, d_tmp, N);
-        tmp2 = OpenclKernels::dot(d_t, d_t, d_tmp, N);
+        tmp1 = OpenclKernels<Scalar>::dot(d_t, d_r, d_tmp, N);
+        tmp2 = OpenclKernels<Scalar>::dot(d_t, d_t, d_tmp, N);
         omega = tmp1 / tmp2;
-        OpenclKernels::axpy(d_s, omega, d_x, N);     // x = x + omega * s
-        OpenclKernels::axpy(d_t, -omega, d_r, N);    // r = r - omega * t
-        norm = OpenclKernels::norm(d_r, d_tmp, N);
+        OpenclKernels<Scalar>::axpy(d_s, omega, d_x, N);     // x = x + omega * s
+        OpenclKernels<Scalar>::axpy(d_t, -omega, d_r, N);    // r = r - omega * t
+        norm = OpenclKernels<Scalar>::norm(d_r, d_tmp, N);
         if (verbosity >= 3) {
             queue->finish();
             t_rest.stop();
@@ -382,7 +398,7 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
         }
     }
 
-    res.iterations = std::min(it, (float)maxit);
+    res.iterations = std::min(it, static_cast<float>(maxit));
     res.reduction = norm / norm_0;
     res.conv_rate  = static_cast<double>(pow(res.reduction, 1.0 / it));
     res.elapsed = t_total.stop();
@@ -390,7 +406,8 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
 
     if (verbosity > 0) {
         std::ostringstream out;
-        out << "=== converged: " << res.converged << ", conv_rate: " << res.conv_rate << ", time: " << res.elapsed << \
+        out << "=== converged: " << res.converged << ", conv_rate: "
+            << res.conv_rate << ", time: " << res.elapsed <<
             ", time per iteration: " << res.elapsed / it << ", iterations: " << it;
         OpmLog::info(out.str());
     }
@@ -405,9 +422,11 @@ void openclSolverBackend<block_size>::gpu_pbicgstab(WellContributions& wellContr
     }
 }
 
-
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::initialize(std::shared_ptr<BlockedMatrix> matrix, std::shared_ptr<BlockedMatrix> jacMatrix) {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+initialize(std::shared_ptr<BlockedMatrix<Scalar>> matrix,
+           std::shared_ptr<BlockedMatrix<Scalar>> jacMatrix)
+{
     this->Nb = matrix->Nb;
     this->N = Nb * block_size;
     this->nnzb = matrix->nnzbs;
@@ -437,22 +456,21 @@ void openclSolverBackend<block_size>::initialize(std::shared_ptr<BlockedMatrix>
         mat = matrix;
         jacMat = jacMatrix;
 
-        d_x = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_b = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_rb = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_r = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_rw = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_p = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_pw = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_s = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_t = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_v = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
-        d_tmp = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * N);
+        d_x = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_b = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_rb = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_r = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_rw = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_p = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_pw = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_s = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_t = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_v = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
+        d_tmp = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * N);
 
-        d_Avals = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(double) * nnz);
+        d_Avals = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(Scalar) * nnz);
         d_Acols = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(int) * nnzb);
         d_Arows = cl::Buffer(*context, CL_MEM_READ_WRITE, sizeof(int) * (Nb + 1));
-
     } catch (const cl::Error& error) {
         std::ostringstream oss;
         oss << "OpenCL Error: " << error.what() << "(" << error.err() << ")\n";
@@ -467,8 +485,10 @@ void openclSolverBackend<block_size>::initialize(std::shared_ptr<BlockedMatrix>
     initialized = true;
 } // end initialize()
 
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::copy_system_to_gpu() {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+copy_system_to_gpu()
+{
     Timer t;
     events.resize(5);
 
@@ -476,18 +496,25 @@ void openclSolverBackend<block_size>::copy_system_to_gpu() {
     int sum = 0;
     for (int i = 0; i < Nb; ++i) {
         int size_row = mat->rowPointers[i + 1] - mat->rowPointers[i];
-        memcpy(vals_contiguous.data() + sum, mat->nnzValues + sum, size_row * sizeof(double) * block_size * block_size);
+        memcpy(vals_contiguous.data() + sum, mat->nnzValues + sum,
+               size_row * sizeof(Scalar) * block_size * block_size);
         sum += size_row * block_size * block_size;
     }
-    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0, sizeof(double) * nnz, vals_contiguous.data(), nullptr, &events[0]);
+    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0,
+                                    sizeof(Scalar) * nnz, vals_contiguous.data(),
+                                    nullptr, &events[0]);
 #else
-    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0, sizeof(double) * nnz, mat->nnzValues, nullptr, &events[0]);
+    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0,
+                                    sizeof(Scalar) * nnz, mat->nnzValues, nullptr, &events[0]);
 #endif
 
-    err |= queue->enqueueWriteBuffer(d_Acols, CL_TRUE, 0, sizeof(int) * nnzb, mat->colIndices, nullptr, &events[1]);
-    err |= queue->enqueueWriteBuffer(d_Arows, CL_TRUE, 0, sizeof(int) * (Nb + 1), mat->rowPointers, nullptr, &events[2]);
-    err |= queue->enqueueWriteBuffer(d_b, CL_TRUE, 0, sizeof(double) * N, h_b, nullptr, &events[3]);
-    err |= queue->enqueueFillBuffer(d_x, 0, 0, sizeof(double) * N, nullptr, &events[4]);
+    err |= queue->enqueueWriteBuffer(d_Acols, CL_TRUE, 0,
+                                     sizeof(int) * nnzb, mat->colIndices, nullptr, &events[1]);
+    err |= queue->enqueueWriteBuffer(d_Arows, CL_TRUE, 0,
+                                     sizeof(int) * (Nb + 1), mat->rowPointers, nullptr, &events[2]);
+    err |= queue->enqueueWriteBuffer(d_b, CL_TRUE, 0,
+                                     sizeof(Scalar) * N, h_b, nullptr, &events[3]);
+    err |= queue->enqueueFillBuffer(d_x, 0, 0, sizeof(Scalar) * N, nullptr, &events[4]);
 
     cl::WaitForEvents(events);
     events.clear();
@@ -504,8 +531,10 @@ void openclSolverBackend<block_size>::copy_system_to_gpu() {
 } // end copy_system_to_gpu()
 
 // don't copy rowpointers and colindices, they stay the same
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::update_system_on_gpu() {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+update_system_on_gpu()
+{
     Timer t;
     events.resize(3);
 
@@ -513,16 +542,21 @@ void openclSolverBackend<block_size>::update_system_on_gpu() {
     int sum = 0;
     for (int i = 0; i < Nb; ++i) {
         int size_row = mat->rowPointers[i + 1] - mat->rowPointers[i];
-        memcpy(vals_contiguous.data() + sum, mat->nnzValues + sum, size_row * sizeof(double) * block_size * block_size);
+        memcpy(vals_contiguous.data() + sum, mat->nnzValues + sum,
+               size_row * sizeof(Scalar) * block_size * block_size);
         sum += size_row * block_size * block_size;
     }
-    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0, sizeof(double) * nnz, vals_contiguous.data(), nullptr, &events[0]);
+    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0,
+                                    sizeof(Scalar) * nnz, vals_contiguous.data(),
+                                    nullptr, &events[0]);
 #else
-    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0, sizeof(double) * nnz, mat->nnzValues, nullptr, &events[0]);
+    err = queue->enqueueWriteBuffer(d_Avals, CL_TRUE, 0,
+                                    sizeof(Scalar) * nnz, mat->nnzValues, nullptr, &events[0]);
 #endif
 
-    err |= queue->enqueueWriteBuffer(d_b, CL_TRUE, 0, sizeof(double) * N, h_b, nullptr, &events[1]);
-    err |= queue->enqueueFillBuffer(d_x, 0, 0, sizeof(double) * N, nullptr, &events[2]);
+    err |= queue->enqueueWriteBuffer(d_b, CL_TRUE, 0,
+                                     sizeof(Scalar) * N, h_b, nullptr, &events[1]);
+    err |= queue->enqueueFillBuffer(d_x, 0, 0, sizeof(Scalar) * N, nullptr, &events[2]);
 
     cl::WaitForEvents(events);
     events.clear();
@@ -538,9 +572,10 @@ void openclSolverBackend<block_size>::update_system_on_gpu() {
     }
 } // end update_system_on_gpu()
 
-
-template <unsigned int block_size>
-bool openclSolverBackend<block_size>::analyze_matrix() {
+template<class Scalar, unsigned int block_size>
+bool openclSolverBackend<Scalar,block_size>::
+analyze_matrix()
+{
     Timer t;
 
     bool success;
@@ -560,9 +595,10 @@ bool openclSolverBackend<block_size>::analyze_matrix() {
     return success;
 } // end analyze_matrix()
 
-
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::update_system(double *vals, double *b) {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+update_system(Scalar* vals, Scalar* b)
+{
     Timer t;
 
     mat->nnzValues = vals;
@@ -575,9 +611,10 @@ void openclSolverBackend<block_size>::update_system(double *vals, double *b) {
     }
 } // end update_system()
 
-
-template <unsigned int block_size>
-bool openclSolverBackend<block_size>::create_preconditioner() {
+template<class Scalar, unsigned int block_size>
+bool openclSolverBackend<Scalar,block_size>::
+create_preconditioner()
+{
     Timer t;
 
     bool result;
@@ -594,9 +631,10 @@ bool openclSolverBackend<block_size>::create_preconditioner() {
     return result;
 } // end create_preconditioner()
 
-
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::solve_system(WellContributions &wellContribs, BdaResult &res) {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+solve_system(WellContributions<Scalar>& wellContribs, BdaResult& res)
+{
     Timer t;
 
     // actually solve
@@ -604,7 +642,8 @@ void openclSolverBackend<block_size>::solve_system(WellContributions &wellContri
         gpu_pbicgstab(wellContribs, res);
     } catch (const cl::Error& error) {
         std::ostringstream oss;
-        oss << "openclSolverBackend::solve_system error: " << error.what() << "(" << error.err() << ")\n";
+        oss << "openclSolverBackend::solve_system error: " << error.what()
+            << "(" << error.err() << ")\n";
         oss << getErrorString(error.err());
         // rethrow exception
         OPM_THROW(std::logic_error, oss.str());
@@ -618,17 +657,17 @@ void openclSolverBackend<block_size>::solve_system(WellContributions &wellContri
         out << "openclSolver::solve_system(): " << t.stop() << " s";
         OpmLog::info(out.str());
     }
-
 } // end solve_system()
 
-
 // copy result to host memory
 // caller must be sure that x is a valid array
-template <unsigned int block_size>
-void openclSolverBackend<block_size>::get_result(double *x) {
+template<class Scalar, unsigned int block_size>
+void openclSolverBackend<Scalar,block_size>::
+get_result(Scalar* x)
+{
     Timer t;
 
-    queue->enqueueReadBuffer(d_x, CL_TRUE, 0, sizeof(double) * N, x);
+    queue->enqueueReadBuffer(d_x, CL_TRUE, 0, sizeof(Scalar) * N, x);
 
     if (verbosity > 2) {
         std::ostringstream out;
@@ -637,13 +676,13 @@ void openclSolverBackend<block_size>::get_result(double *x) {
     }
 } // end get_result()
 
-
-template <unsigned int block_size>
-SolverStatus openclSolverBackend<block_size>::solve_system(std::shared_ptr<BlockedMatrix> matrix,
-                                                           double *b,
-                                                           std::shared_ptr<BlockedMatrix> jacMatrix,
-                                                           WellContributions& wellContribs,
-                                                           BdaResult &res)
+template<class Scalar, unsigned int block_size>
+SolverStatus openclSolverBackend<Scalar,block_size>::
+solve_system(std::shared_ptr<BlockedMatrix<Scalar>> matrix,
+             Scalar* b,
+             std::shared_ptr<BlockedMatrix<Scalar>> jacMatrix,
+             WellContributions<Scalar>& wellContribs,
+             BdaResult& res)
 {
     if (initialized == false) {
         initialize(matrix, jacMatrix);
@@ -668,21 +707,18 @@ SolverStatus openclSolverBackend<block_size>::solve_system(std::shared_ptr<Block
     return SolverStatus::BDA_SOLVER_SUCCESS;
 }
 
+#define INSTANTIATE_TYPE(T)                  \
+    template class openclSolverBackend<T,1>; \
+    template class openclSolverBackend<T,2>; \
+    template class openclSolverBackend<T,3>; \
+    template class openclSolverBackend<T,4>; \
+    template class openclSolverBackend<T,5>; \
+    template class openclSolverBackend<T,6>;
 
-#define INSTANTIATE_BDA_FUNCTIONS(n)                                        \
-template openclSolverBackend<n>::openclSolverBackend(                       \
-    int, int, double, unsigned int, unsigned int, bool, std::string); \
-template openclSolverBackend<n>::openclSolverBackend(int, int, double, bool); \
-template void openclSolverBackend<n>::setOpencl(std::shared_ptr<cl::Context>&, std::shared_ptr<cl::CommandQueue>&);
+INSTANTIATE_TYPE(double)
 
-INSTANTIATE_BDA_FUNCTIONS(1);
-INSTANTIATE_BDA_FUNCTIONS(2);
-INSTANTIATE_BDA_FUNCTIONS(3);
-INSTANTIATE_BDA_FUNCTIONS(4);
-INSTANTIATE_BDA_FUNCTIONS(5);
-INSTANTIATE_BDA_FUNCTIONS(6);
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
-#undef INSTANTIATE_BDA_FUNCTIONS
-
-} // namespace Accelerator
-} // namespace Opm
+} // namespace Opm::Accelerator
