@@ -21,12 +21,12 @@
 #include "config.h"
 #endif // HAVE_CONFIG_H
 
-#include <cassert>
-#include <iostream>
-#include <vector>
-
 #include <algorithm>
+#include <cassert>
+#include <cstddef>
+#include <ostream>
 #include <numeric>
+#include <vector>
 
 #include <opm/input/eclipse/Units/Units.hpp>
 #include <opm/simulators/timestepping/AdaptiveSimulatorTimer.hpp>
@@ -62,7 +62,7 @@ namespace Opm
         return ( report_step_ == 0 ) && ( current_step_ == 0 );
     }
 
-AdaptiveSimulatorTimer& AdaptiveSimulatorTimer::operator++ ()
+    AdaptiveSimulatorTimer& AdaptiveSimulatorTimer::operator++ ()
     {
         ++current_step_;
         current_time_ += dt_;
@@ -115,6 +115,12 @@ AdaptiveSimulatorTimer& AdaptiveSimulatorTimer::operator++ ()
         return dt_;
     }
 
+    void AdaptiveSimulatorTimer::setCurrentStepLength(double dt)
+    {
+        assert(dt > 0);
+        dt_ = dt;
+    }
+
     double AdaptiveSimulatorTimer::stepLengthTaken() const
     {
         assert( ! steps_.empty() );
@@ -157,7 +163,7 @@ AdaptiveSimulatorTimer& AdaptiveSimulatorTimer::operator++ ()
     report(std::ostream& os) const
     {
         os << "Sub steps started at time = " <<  unit::convert::to( start_time_, unit::day ) << " (days)" << std::endl;
-        for( size_t i=0; i<steps_.size(); ++i )
+        for (std::size_t i = 0; i < steps_.size(); ++i)
         {
             os << " step[ " << i << " ] = " << unit::convert::to( steps_[ i ], unit::day ) << " (days)" << std::endl;
         }
