@@ -97,19 +97,17 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"COMPRIV", {true, std::nullopt}},
         {"COMPRP", {true, std::nullopt}},
         {"COMPRPL", {true, std::nullopt}},
+        {"COMPS", {false,  std::nullopt}},
         {"COMPSEGL", {true, std::nullopt}},
         {"COMPVE", {true, std::nullopt}},
         {"COMPVEL", {true, std::nullopt}},
         {"CPIFACT", {true, std::nullopt}},
         {"CPIFACTL", {true, std::nullopt}},
-        {"CSKIN", {true, std::nullopt}},
         {"CONNECTION", {true, std::nullopt}},
         {"CONNECTION_PROBE", {true, std::nullopt}},
-        {"COORDSYS", {true, std::nullopt}},
+        {"COORDSYS", {false, std::string{"Multiple grid systems not supported, COORDSYS is ignored."}}},
         {"COPYBOX", {true, std::nullopt}},
         {"CRITPERM", {true, std::nullopt}},
-        {"DATUMR", {true, std::nullopt}},
-        {"DATUMRX", {true, std::nullopt}},
         {"DCQDEFN", {true, std::nullopt}},
         {"DEBUG", {false, std::nullopt}},
         {"DELAYACT", {true, std::nullopt}},
@@ -163,7 +161,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"ENDDYN", {true, std::nullopt}},
         {"ENDFIN", {true, std::nullopt}},
         {"ENDNUM", {true, std::nullopt}},
-        {"ENDSKIP", {true, std::nullopt}},
         {"ENKRVD", {true, std::nullopt}},
         {"ENKSRVD", {true, std::nullopt}},
         {"ENPCVD", {true, std::nullopt}},
@@ -172,7 +169,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"EPSDBGS", {false, std::nullopt}},
         {"EPSDEBUG", {false, std::nullopt}},
         {"EQLZCORN", {true, std::nullopt}},
-        {"EQUALREG", {true, std::nullopt}},
         {"ESSNODE", {true, std::nullopt}},
         {"EXCAVATE", {true, std::nullopt}},
         {"EXCEL", {false, std::nullopt}},
@@ -180,9 +176,9 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"EXTHOST", {true, std::nullopt}},
         {"EXTRAPMS", {false, std::nullopt}},
         {"EXTREPGL", {true, std::nullopt}},
-        {"FBHPDEF", {true, std::nullopt}},
         {"FHERCHBL", {true, std::nullopt}},
         {"FRICTION", {true, std::nullopt}},
+        {"FIELDSEP", {false, std::nullopt}},
         {"FIPOWG", {false, std::string{"Report not available"}}},
         {"FIPSEP", {true, std::nullopt}},
         {"FLUXREG", {true, std::nullopt}},
@@ -213,6 +209,7 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"GASPERIO", {true, std::nullopt}},
         {"GASSATC", {true, std::nullopt}},
         {"GASYEAR", {true, std::nullopt}},
+        {"GASWAT", {false, std::nullopt}},
         {"GCALECON", {true, std::nullopt}},
         {"GCONCAL", {true, std::nullopt}},
         {"GCONENG", {true, std::nullopt}},
@@ -226,7 +223,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"GDIMS", {true, std::nullopt}},
         {"GDORIENT", {false, std::nullopt}},
         {"GDRILPOT", {true, std::nullopt}},
-        {"GECON", {true, std::nullopt}},
         {"GECONT", {true, std::nullopt}},
         {"GETGLOB", {true, std::nullopt}},
         {"GI", {true, std::nullopt}},
@@ -247,7 +243,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"GRAVDRM", {true, std::nullopt}},
         {"GRDREACH", {true, std::nullopt}},
         {"GRUPMAST", {true, std::nullopt}},
-        {"GRUPNET", {true, std::nullopt}},
         {"GRUPRIG", {true, std::nullopt}},
         {"GRUPSLAV", {true, std::nullopt}},
         {"GRUPTARG", {true, std::nullopt}},
@@ -285,6 +280,7 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"HMWPIMLT", {true, std::nullopt}},
         {"HMxxxxxx", {true, std::nullopt}},
         {"HRFIN", {true, std::nullopt}},
+        {"HWELLS", {false, std::nullopt}},
         {"HWKRO", {true, std::nullopt}},
         {"HWKRORG", {true, std::nullopt}},
         {"HWKRORW", {true, std::nullopt}},
@@ -401,6 +397,7 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"MULT_XYZ", {true, std::nullopt}},
         {"NARROW", {true, std::nullopt}},
         {"NCONSUMP", {true, std::nullopt}},
+        {"NCOMPS", {false, std::nullopt}},
         {"NEFAC", {true, std::nullopt}},
         {"NETCOMPA", {true, std::nullopt}},
         {"NEXT", {false, std::nullopt}},
@@ -421,7 +418,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"NOGGF", {true, std::nullopt}},
         {"NOINSPEC", {false, std::nullopt}},
         {"NOMONITO", {true, std::nullopt}},
-        {"NONNC", {true, std::nullopt}},
         {"NORSSPEC", {false, std::nullopt}},
         {"NOWARN", {false, std::nullopt}},
         {"NSTACK", {false, std::nullopt}},
@@ -436,7 +432,7 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"OPTIONS", {true, std::nullopt}},
         {"OUTRAD", {true, std::string{"Use the DRV keyword instead"}}},
         {"OUTSOL", {false, std::nullopt}},
-        {"PARAOPTS", {true, std::nullopt}},
+        {"PARAOPTS", {false, std::nullopt}},
         {"PCG32D", {true, std::nullopt}},
         {"PCW32D", {true, std::nullopt}},
         {"PERMJFUN", {true, std::nullopt}},
@@ -479,7 +475,7 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"PERMXY", {true, std::nullopt}},
         {"PERMYZ", {true, std::nullopt}},
         {"PERMZX", {true, std::nullopt}},
-        {"PETGRID", {true, std::nullopt}},
+        {"PETGRID", {true, std::string{"OPM Flow cannot read GSG format grid input, re-export case from Petrel using GRDECL instead."}}},
         {"PICOND", {true, std::nullopt}},
         {"PIMULTAB", {true, std::nullopt}},
         {"PINCHNUM", {true, std::nullopt}},
@@ -490,7 +486,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"PLYATEMP", {true, std::nullopt}},
         {"PLYCAMAX", {true, std::nullopt}},
         {"PLYDHFLF", {true, std::nullopt}},
-        {"PPCWMAX", {true, std::nullopt}},
         {"PRORDER", {true, std::nullopt}},
         {"PRVD", {true, std::nullopt}},
         {"PVTGWO", {true, std::nullopt}},
@@ -550,9 +545,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"RSSPEC", {true, std::nullopt}},
         {"SAMG", {true, std::nullopt}},
         {"SAVE", {false, std::nullopt}},
-        {"SKIP", {true, std::nullopt}},
-        {"SKIP100", {true, std::nullopt}},
-        {"SKIP300", {true, std::nullopt}},
         {"SALTNODE", {true, std::nullopt}},
         {"SALTREST", {true, std::nullopt}},
         {"SCALELIM", {true, std::nullopt}},
@@ -583,9 +575,10 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"SMULTZ", {true, std::nullopt}},
         {"SOCRS", {true, std::nullopt}},
         {"SOF32D", {true, std::nullopt}},
+        {"SOLID", {false, std::nullopt}},
         {"SOLVCONC", {true, std::nullopt}},
         {"SOLVDIMS", {true, std::nullopt}},
-        {"SOLVDIRS", {true, std::nullopt}},
+        {"SOLVDIRS", {false, std::nullopt}},
         {"SOLVFRAC", {true, std::nullopt}},
         {"SOLVNUM", {true, std::nullopt}},
         {"SOLWNUM", {true, std::nullopt}},
@@ -646,8 +639,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"TUNINGL", {false, std::nullopt}},
         {"TUNINGS", {false, std::nullopt}},
         {"TZONE", {true, std::nullopt}},
-        {"UDT", {true, std::nullopt}},
-        {"UDTDIMS", {true, std::nullopt}},
         {"UNCODHMD", {true, std::nullopt}},
         {"UNIFOUTS", {false, std::nullopt}},
         {"UNIFSAVE", {false, std::nullopt}},
@@ -669,7 +660,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"VISCD", {true, std::nullopt}},
         {"VISDATES", {true, std::nullopt}},
         {"VISOPTS", {true, std::nullopt}},
-        {"WAGHYSTR", {true, std::nullopt}},
         {"WAITBAL", {true, std::nullopt}},
         {"WALKALIN", {true, std::nullopt}},
         {"WALQCALC", {true, std::nullopt}},
@@ -678,17 +668,11 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"WBHGLR", {true, std::nullopt}},
         {"WBOREVOL", {true, std::nullopt}},
         {"WCALCVAL", {true, std::nullopt}},
-        {"WBP", {false, std::nullopt}},
-        {"WBP4", {false, std::nullopt}},
-        {"WBP5", {false, std::nullopt}},
-        {"WBP9", {false, std::nullopt}},
         {"WCONINJ", {true, std::nullopt}},
         {"WCONINJP", {true, std::nullopt}},
         {"WCUTBACK", {true, std::nullopt}},
         {"WCUTBACT", {true, std::nullopt}},
         {"WCYCLE", {true, std::nullopt}},
-        {"WDFACCOR", {true, std::nullopt}},
-        {"WDFAC", {true, std::nullopt}},
         {"WDRILTIM", {true, std::nullopt}},
         {"WDRILPRI", {true, std::nullopt}},
         {"WDRILRES", {true, std::nullopt}},
@@ -698,6 +682,7 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"WELDEBUG", {true, std::nullopt}},
         {"WELDRAW", {true, std::nullopt}},
         {"WELEVNT", {true, std::nullopt}},
+        {"WELLSTRE", {false, std::nullopt}},
         {"WELMOVEL", {true, std::nullopt}},
         {"WELOPENL", {true, std::nullopt}},
         {"WELPRI", {true, std::nullopt}},
@@ -709,11 +694,11 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"WFRICTNL", {true, std::nullopt}},
         {"WGASPROD", {true, std::nullopt}},
         {"WGORPEN", {true, std::nullopt}},
+        {"WINJGAS", {false, std::nullopt}},
         {"WH2NUM", {true, std::nullopt}},
         {"WH3NUM", {true, std::nullopt}},
         {"WHEDREFD", {true, std::nullopt}},
         {"WHTEMP", {true, std::nullopt}},
-        {"WINJMULT", {true, std::nullopt}},
         {"WLIMTOL", {true, std::nullopt}},
         {"WLIFT", {true, std::nullopt}},
         {"WLISTARG", {true, std::nullopt}},
@@ -723,7 +708,6 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"WNETDP", {true, std::nullopt}},
         {"WORKLIM", {true, std::nullopt}},
         {"WORKTHP", {true, std::nullopt}},
-        {"WPAVE", {false, std::nullopt}},
         {"WPIMULTL", {true, std::nullopt}},
         {"WPITAB", {true, std::nullopt}},
         {"WPLUG", {true, std::nullopt}},
@@ -753,11 +737,9 @@ const KeywordValidation::UnsupportedKeywords& unsupportedKeywords()
         {"WTADD", {true, std::nullopt}},
         {"WTEMPQ", {true, std::nullopt}},
         {"WTHPMAX", {true, std::nullopt}},
-        {"WTMULT", {true, std::nullopt}},
-        {"WVFPDP", {true, std::nullopt}},
-        {"WWPAVE", {true, std::nullopt}},
         {"ZIPPY2", {false, std::nullopt}},
         {"ZIPP2OFF", {false, std::nullopt}},
+        {"ZMFVD", {false, std::nullopt}},        
     };
 
     return unsupported_keywords;
