@@ -33,6 +33,8 @@
 #include <string>
 #include <vector>
 
+#include <opm/grid/utility/SparseTable.hpp>
+
 #include <opm/input/eclipse/Schedule/Group/Group.hpp>
 #include <opm/input/eclipse/Schedule/Group/GuideRate.hpp>
 #include <opm/input/eclipse/Schedule/Schedule.hpp>
@@ -78,26 +80,26 @@
 
 #include <opm/simulators/utils/DeferredLogger.hpp>
 
-namespace Opm::Properties {
+namespace Opm::Parameters {
 
-template<class TypeTag, class MyTypeTag>
-struct EnableTerminalOutput {
-    using type = UndefinedProperty;
-};
+struct EnableTerminalOutput { static constexpr bool value = true; };
 
-} // namespace Opm::Properties
+} // namespace Opm::Parameters
 
 namespace Opm {
 
+#if COMPILE_BDA_BRIDGE
+template<class Scalar> class WellContributions;
+#endif
+
         /// Class for handling the blackoil well model.
         template<typename TypeTag>
         class BlackoilWellModel : public BaseAuxiliaryModule<TypeTag>
-                                , public BlackoilWellModelGeneric
+                                , public BlackoilWellModelGeneric<GetPropType<TypeTag,
+                                                                              Properties::Scalar>>
         {
         public:
             // ---------      Types      ---------
-            using ModelParameters = BlackoilModelParameters<TypeTag>;
-
             using Grid = GetPropType<TypeTag, Properties::Grid>;
             using EquilGrid = GetPropType<TypeTag, Properties::EquilGrid>;
             using FluidSystem = GetPropType<TypeTag, Properties::FluidSystem>;
@@ -109,12 +111,15 @@ namespace Opm {
             using GlobalEqVector = GetPropType<TypeTag, Properties::GlobalEqVector>;
             using SparseMatrixAdapter = GetPropType<TypeTag, Properties::SparseMatrixAdapter>;
             using GasLiftSingleWell = typename WellInterface<TypeTag>::GasLiftSingleWell;
-            using GLiftOptWells = typename BlackoilWellModelGeneric::GLiftOptWells;
-            using GLiftProdWells = typename BlackoilWellModelGeneric::GLiftProdWells;
+            using GLiftOptWells = typename BlackoilWellModelGeneric<Scalar>::GLiftOptWells;
+            using GLiftProdWells = typename BlackoilWellModelGeneric<Scalar>::GLiftProdWells;
             using GLiftWellStateMap =
-                typename BlackoilWellModelGeneric::GLiftWellStateMap;
-            using GLiftEclWells = typename GasLiftGroupInfo::GLiftEclWells;
-            using GLiftSyncGroups = typename GasLiftSingleWellGeneric::GLiftSyncGroups;
+                typename BlackoilWellModelGeneric<Scalar>::GLiftWellStateMap;
+            using GLiftEclWells = typename GasLiftGroupInfo<Scalar>::GLiftEclWells;
+            using GLiftSyncGroups = typename GasLiftSingleWellGeneric<Scalar>::GLiftSyncGroups;
+
+            using ModelParameters = BlackoilModelParameters<Scalar>;
+
             constexpr static std::size_t pressureVarIndex = GetPropType<TypeTag, Properties::Indices>::pressureSwitchIdx;
             typedef typename BaseAuxiliaryModule<TypeTag>::NeighborSet NeighborSet;
 
@@ -234,7 +239,7 @@ namespace Opm {
 
             using WellInterfacePtr = std::shared_ptr<WellInterface<TypeTag> >;
 
-            using BlackoilWellModelGeneric::initFromRestartFile;
+            using BlackoilWellModelGeneric<Scalar>::initFromRestartFile;
             void initFromRestartFile(const RestartValue& restartValues)
             {
                 initFromRestartFile(restartValues,
@@ -243,7 +248,7 @@ namespace Opm {
                                     param_.use_multisegment_well_);
             }
 
-            using BlackoilWellModelGeneric::prepareDeserialize;
+            using BlackoilWellModelGeneric<Scalar>::prepareDeserialize;
             void prepareDeserialize(const int report_step)
             {
                 prepareDeserialize(report_step, grid().size(0),
@@ -259,13 +264,33 @@ namespace Opm {
                     return this->wasDynamicallyShutThisTimeStep(well_index);
                 });
 
-                const auto& tracerRates = simulator_.problem().tracerModel().getWellTracerRates();
-                this->assignWellTracerRates(wsrpt, tracerRates);
+                {
+                    const auto& tracerRates = this->simulator_.problem()
+                        .tracerModel().getWellTracerRates();
+                    const auto& freeTracerRates = simulator_.problem()
+                        .tracerModel().getWellFreeTracerRates();
+                    const auto& solTracerRates = simulator_.problem()
+                        .tracerModel().getWellSolTracerRates();
+                    const auto& mswTracerRates = simulator_.problem()
+                        .tracerModel().getMswTracerRates();
 
+                    this->assignWellTracerRates(wsrpt, tracerRates);
+                    this->assignWellTracerRates(wsrpt, freeTracerRates);
+                    this->assignWellTracerRates(wsrpt, solTracerRates);
+                    this->assignMswTracerRates(wsrpt, mswTracerRates);
+                }
 
-                BlackoilWellModelGuideRates(*this).assignWellGuideRates(wsrpt, this->reportStepIndex());
+                BlackoilWellModelGuideRates(*this)
+                    .assignWellGuideRates(wsrpt, this->reportStepIndex());
+
+                this->assignWellTargets(wsrpt);
                 this->assignShutConnections(wsrpt, this->reportStepIndex());
-
+                // only used to compute gas injection mass rates for CO2STORE runs
+                // The gas reference density is thus the same for all pvt regions
+                // We therefore for simplicity use 0 here 
+                if (eclState().runspec().co2Storage()) {
+                    this->assignMassGasRate(wsrpt, FluidSystem::referenceDensity(FluidSystem::gasPhaseIdx, 0));
+                }
                 return wsrpt;
             }
 
@@ -274,18 +299,21 @@ namespace Opm {
                 return this->computeWellBlockAveragePressures();
             }
 
-            // subtract Binv(D)rw from r;
-            void apply( BVector& r) const;
-
             // subtract B*inv(D)*C * x from A*x
             void apply(const BVector& x, BVector& Ax) const;
 
+            void applyDomain(const BVector& x, BVector& Ax, const int domainIndex) const;
+
+#if COMPILE_BDA_BRIDGE
             // accumulate the contributions of all Wells in the WellContributions object
-            void getWellContributions(WellContributions& x) const;
+            void getWellContributions(WellContributions<Scalar>& x) const;
+#endif
 
             // apply well model with scaling of alpha
             void applyScaleAdd(const Scalar alpha, const BVector& x, BVector& Ax) const;
 
+            void applyScaleAddDomain(const Scalar alpha, const BVector& x, BVector& Ax, const int domainIndex) const;
+
             // Check if well equations is converged.
             ConvergenceReport getWellConvergence(const std::vector<Scalar>& B_avg, const bool checkWellGroupControls = false) const;
 
@@ -332,10 +360,16 @@ namespace Opm {
             WellInterfacePtr getWell(const std::string& well_name) const;
             bool hasWell(const std::string& well_name) const;
 
-            using PressureMatrix = Dune::BCRSMatrix<Opm::MatrixBlock<double, 1, 1>>;
+            using PressureMatrix = Dune::BCRSMatrix<Opm::MatrixBlock<Scalar, 1, 1>>;
 
             void addWellPressureEquations(PressureMatrix& jacobian, const BVector& weights,const bool use_well_weights) const;
 
+            void addWellPressureEquationsDomain([[maybe_unused]] PressureMatrix& jacobian,
+                                                [[maybe_unused]] const BVector& weights,
+                                                [[maybe_unused]] const bool use_well_weights,
+                                                [[maybe_unused]] const int domainIndex) const;
+
+
             void addWellPressureEquationsStruct(PressureMatrix& jacobian) const;
 
             void initGliftEclWellMap(GLiftEclWells &ecl_well_map);
@@ -354,8 +388,8 @@ namespace Opm {
             void updateWellControlsDomain(DeferredLogger& deferred_logger, const Domain& domain);
 
             void logPrimaryVars() const;
-            std::vector<double> getPrimaryVarsDomain(const Domain& domain) const;
-            void setPrimaryVarsDomain(const Domain& domain, const std::vector<double>& vars);
+            std::vector<Scalar> getPrimaryVarsDomain(const Domain& domain) const;
+            void setPrimaryVarsDomain(const Domain& domain, const std::vector<Scalar>& vars);
 
             void setupDomains(const std::vector<Domain>& domains);
 
@@ -388,10 +422,10 @@ namespace Opm {
             std::size_t global_num_cells_{};
             // the number of the cells in the local grid
             std::size_t local_num_cells_{};
-            double gravity_{};
-            std::vector<double> depth_{};
+            Scalar gravity_{};
+            std::vector<Scalar> depth_{};
             bool alternative_well_rate_init_{};
-
+            std::map<std::string, Scalar> well_group_thp_calc_;
             std::unique_ptr<RateConverterType> rateConverter_{};
             std::map<std::string, std::unique_ptr<AverageRegionalPressureType>> regionalAveragePressureCalculator_{};
 
@@ -416,6 +450,9 @@ namespace Opm {
             // Keep track of the domain of each well, if using subdomains.
             std::map<std::string, int> well_domain_;
 
+            // Store the local index of the wells perforated cells in the domain, if using sumdomains
+            SparseTable<int> well_local_cells_;
+
             const Grid& grid() const
             { return simulator_.vanguard().grid(); }
 
@@ -443,6 +480,8 @@ namespace Opm {
                                               const double dt,
                                               DeferredLogger& local_deferredLogger);
 
+            void computeWellGroupThp(const double dt, DeferredLogger& local_deferredLogger);
+
             /// Update rank's notion of intersecting wells and their
             /// associate solution variables.
             ///
@@ -481,7 +520,7 @@ namespace Opm {
             data::WellBlockAveragePressures
             computeWellBlockAveragePressures() const;
 
-            ParallelWBPCalculation::EvaluatorFactory
+            typename ParallelWBPCalculation<Scalar>::EvaluatorFactory
             makeWellSourceEvaluatorFactory(const std::vector<Well>::size_type wellIdx) const;
 
             void registerOpenWellsForWBPCalculation();
@@ -489,12 +528,12 @@ namespace Opm {
             void updateAverageFormationFactor();
 
             void computePotentials(const std::size_t widx,
-                                   const WellState& well_state_copy,
+                                   const WellState<Scalar>& well_state_copy,
                                    std::string& exc_msg,
                                    ExceptionType::ExcEnum& exc_type,
                                    DeferredLogger& deferred_logger) override;
 
-            const std::vector<double>& wellPerfEfficiencyFactors() const;
+            const std::vector<Scalar>& wellPerfEfficiencyFactors() const;
 
             void calculateProductivityIndexValuesShutWells(const int reportStepIdx, DeferredLogger& deferred_logger) override;
             void calculateProductivityIndexValues(DeferredLogger& deferred_logger) override;
@@ -517,15 +556,19 @@ namespace Opm {
             bool maybeDoGasLiftOptimize(DeferredLogger& deferred_logger);
 
             void gasLiftOptimizationStage1(DeferredLogger& deferred_logger,
-                GLiftProdWells &prod_wells, GLiftOptWells &glift_wells,
-                GasLiftGroupInfo &group_info, GLiftWellStateMap &state_map);
+                                           GLiftProdWells& prod_wells,
+                                           GLiftOptWells& glift_wells,
+                                           GasLiftGroupInfo<Scalar>& group_info,
+                                           GLiftWellStateMap& state_map);
 
             // cannot be const since it accesses the non-const WellState
-            void gasLiftOptimizationStage1SingleWell(WellInterface<TypeTag> *well,
-                DeferredLogger& deferred_logger,
-                GLiftProdWells &prod_wells, GLiftOptWells &glift_wells,
-                GasLiftGroupInfo &group_info, GLiftWellStateMap &state_map,
-                GLiftSyncGroups& groups_to_sync);
+            void gasLiftOptimizationStage1SingleWell(WellInterface<TypeTag>* well,
+                                                     DeferredLogger& deferred_logger,
+                                                     GLiftProdWells& prod_wells,
+                                                     GLiftOptWells& glift_wells,
+                                                     GasLiftGroupInfo<Scalar>& group_info,
+                                                     GLiftWellStateMap& state_map,
+                                                     GLiftSyncGroups& groups_to_sync);
 
             void extractLegacyCellPvtRegionIndex_();
 
@@ -536,14 +579,14 @@ namespace Opm {
 
             void wellTesting(const int timeStepIdx, const double simulationTime, DeferredLogger& deferred_logger);
 
-            void calcRates(const int fipnum,
-                           const int pvtreg,
-                           const std::vector<double>& production_rates,
-                           std::vector<double>& resv_coeff) override;
+            void calcResvCoeff(const int fipnum,
+                               const int pvtreg,
+                               const std::vector<Scalar>& production_rates,
+                               std::vector<Scalar>& resv_coeff) override;
 
-            void calcInjRates(const int fipnum,
-                           const int pvtreg,
-                           std::vector<double>& resv_coeff) override;
+            void calcInjResvCoeff(const int fipnum,
+                                  const int pvtreg,
+                                  std::vector<Scalar>& resv_coeff) override;
 
             void computeWellTemperature();
 
@@ -553,10 +596,22 @@ namespace Opm {
 
         private:
             BlackoilWellModel(Simulator& simulator, const PhaseUsage& pu);
+
+            // These members are used to avoid reallocation in specific functions
+            // (e.g., apply, applyDomain) instead of using local variables.
+            // Their state is not relevant between function calls, so they can
+            // (and must) be mutable, as the functions using them are const.
+            mutable BVector x_local_;
+            mutable BVector Ax_local_;
+            mutable BVector res_local_;
+            mutable GlobalEqVector linearize_res_local_;
         };
 
 
 } // namespace Opm
 
+#ifndef OPM_BLACKOILWELLMODEL_IMPL_HEADER_INCLUDED
 #include "BlackoilWellModel_impl.hpp"
 #endif
+
+#endif
