@@ -37,7 +37,8 @@
 
 namespace Opm {
 
-bool BlackoilWellModelConstraints::
+template<class Scalar>
+bool BlackoilWellModelConstraints<Scalar>::
 hasTHPConstraints() const
 {
     int local_result = false;
@@ -49,8 +50,9 @@ hasTHPConstraints() const
     return wellModel_.comm().max(local_result);
 }
 
-std::pair<Group::InjectionCMode, double>
-BlackoilWellModelConstraints::
+template<class Scalar>
+std::pair<Group::InjectionCMode, Scalar>
+BlackoilWellModelConstraints<Scalar>::
 checkGroupInjectionConstraints(const Group& group,
                                const int reportStepIdx,
                                const Phase& phase) const
@@ -73,20 +75,25 @@ checkGroupInjectionConstraints(const Group& group,
     {
         if (currentControl != Group::InjectionCMode::RATE)
         {
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(), well_state, reportStepIdx, phasePos, /*isInjector*/true);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                          wellModel_.schedule(),
+                                                                          well_state,
+                                                                          reportStepIdx,
+                                                                          phasePos,
+                                                                          /*isInjector*/true);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
 
             const auto& controls = group.injectionControls(phase, wellModel_.summaryState());
-            double target = controls.surface_max_rate;
+            Scalar target = controls.surface_max_rate;
 
             if (group.has_gpmaint_control(phase, Group::InjectionCMode::RATE))
                 target = wellModel_.groupState().gpmaint_target(group.name());
 
             if (target < current_rate) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = target / current_rate;
                 return std::make_pair(Group::InjectionCMode::RATE, scale);
@@ -97,19 +104,24 @@ checkGroupInjectionConstraints(const Group& group,
     {
         if (currentControl != Group::InjectionCMode::RESV)
         {
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(), well_state, reportStepIdx, phasePos, /*isInjector*/true);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellResRates(group,
+                                                                      wellModel_.schedule(),
+                                                                      well_state,
+                                                                      reportStepIdx,
+                                                                      phasePos,
+                                                                      /*isInjector*/true);
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
 
             const auto& controls = group.injectionControls(phase, wellModel_.summaryState());
-            double target = controls.resv_max_rate;
+            Scalar target = controls.resv_max_rate;
 
             if (group.has_gpmaint_control(phase, Group::InjectionCMode::RESV))
                 target = wellModel_.groupState().gpmaint_target(group.name());
 
             if (target < current_rate) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = target / current_rate;
                 return std::make_pair(Group::InjectionCMode::RESV, scale);
@@ -120,26 +132,32 @@ checkGroupInjectionConstraints(const Group& group,
     {
         if (currentControl != Group::InjectionCMode::REIN)
         {
-            double production_Rate = 0.0;
+            Scalar production_Rate = 0.0;
             const auto& controls = group.injectionControls(phase, wellModel_.summaryState());
             const Group& groupRein = wellModel_.schedule().getGroup(controls.reinj_group, reportStepIdx);
-            production_Rate += WellGroupHelpers::sumWellSurfaceRates(groupRein, wellModel_.schedule(),
-                                                                     well_state, reportStepIdx,
-                                                                     phasePos, /*isInjector*/false);
+            production_Rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(groupRein,
+                                                                             wellModel_.schedule(),
+                                                                             well_state,
+                                                                             reportStepIdx,
+                                                                             phasePos,
+                                                                             /*isInjector*/false);
 
             // sum over all nodes
             production_Rate = wellModel_.comm().sum(production_Rate);
 
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(),
-                                                                  well_state, reportStepIdx,
-                                                                  phasePos, /*isInjector*/true);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                          wellModel_.schedule(),
+                                                                          well_state,
+                                                                          reportStepIdx,
+                                                                          phasePos,
+                                                                          /*isInjector*/true);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
 
             if (controls.target_reinj_fraction*production_Rate < current_rate) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = controls.target_reinj_fraction*production_Rate / current_rate;
                 return std::make_pair(Group::InjectionCMode::REIN, scale);
@@ -150,38 +168,55 @@ checkGroupInjectionConstraints(const Group& group,
     {
         if (currentControl != Group::InjectionCMode::VREP)
         {
-            double voidage_rate = 0.0;
+            Scalar voidage_rate = 0.0;
             const auto& controls = group.injectionControls(phase, wellModel_.summaryState());
             const Group& groupVoidage = wellModel_.schedule().getGroup(controls.voidage_group, reportStepIdx);
-            voidage_rate += WellGroupHelpers::sumWellResRates(groupVoidage, wellModel_.schedule(),
-                                                              well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Aqua], false);
-            voidage_rate += WellGroupHelpers::sumWellResRates(groupVoidage, wellModel_.schedule(),
-                                                              well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Liquid], false);
-            voidage_rate += WellGroupHelpers::sumWellResRates(groupVoidage, wellModel_.schedule(),
-                                                              well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Vapour], false);
+            voidage_rate += WellGroupHelpers<Scalar>::sumWellResRates(groupVoidage,
+                                                                      wellModel_.schedule(),
+                                                                      well_state,
+                                                                      reportStepIdx,
+                                                                      pu.phase_pos[BlackoilPhases::Aqua],
+                                                                      false);
+            voidage_rate += WellGroupHelpers<Scalar>::sumWellResRates(groupVoidage,
+                                                                      wellModel_.schedule(),
+                                                                      well_state,
+                                                                      reportStepIdx,
+                                                                      pu.phase_pos[BlackoilPhases::Liquid],
+                                                                      false);
+            voidage_rate += WellGroupHelpers<Scalar>::sumWellResRates(groupVoidage,
+                                                                      wellModel_.schedule(),
+                                                                      well_state, reportStepIdx,
+                                                                      pu.phase_pos[BlackoilPhases::Vapour],
+                                                                      false);
 
             // sum over all nodes
             voidage_rate = wellModel_.comm().sum(voidage_rate);
 
-            double total_rate = 0.0;
-            total_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
-                                                            well_state, reportStepIdx,
-                                                            pu.phase_pos[BlackoilPhases::Aqua], true);
-            total_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
-                                                            well_state, reportStepIdx,
-                                                            pu.phase_pos[BlackoilPhases::Liquid], true);
-            total_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
-                                                            well_state, reportStepIdx,
-                                                            pu.phase_pos[BlackoilPhases::Vapour], true);
+            Scalar total_rate = 0.0;
+            total_rate += WellGroupHelpers<Scalar>::sumWellResRates(group,
+                                                                    wellModel_.schedule(),
+                                                                    well_state,
+                                                                    reportStepIdx,
+                                                                    pu.phase_pos[BlackoilPhases::Aqua],
+                                                                    true);
+            total_rate += WellGroupHelpers<Scalar>::sumWellResRates(group,
+                                                                    wellModel_.schedule(),
+                                                                    well_state,
+                                                                    reportStepIdx,
+                                                                    pu.phase_pos[BlackoilPhases::Liquid],
+                                                                    true);
+            total_rate += WellGroupHelpers<Scalar>::sumWellResRates(group,
+                                                                    wellModel_.schedule(),
+                                                                    well_state,
+                                                                    reportStepIdx,
+                                                                    pu.phase_pos[BlackoilPhases::Vapour],
+                                                                    true);
 
             // sum over all nodes
             total_rate = wellModel_.comm().sum(total_rate);
 
             if (controls.target_void_fraction*voidage_rate < total_rate) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (total_rate > 1e-12)
                     scale = controls.target_void_fraction*voidage_rate / total_rate;
                 return std::make_pair(Group::InjectionCMode::VREP, scale);
@@ -191,8 +226,9 @@ checkGroupInjectionConstraints(const Group& group,
     return std::make_pair(Group::InjectionCMode::NONE, 1.0);
 }
 
-std::pair<Group::ProductionCMode, double>
-BlackoilWellModelConstraints::
+template<class Scalar>
+std::pair<Group::ProductionCMode, Scalar>
+BlackoilWellModelConstraints<Scalar>::
 checkGroupProductionConstraints(const Group& group,
                                 const int reportStepIdx,
                                 DeferredLogger& deferred_logger) const
@@ -207,16 +243,19 @@ checkGroupProductionConstraints(const Group& group,
     {
         if (currentControl != Group::ProductionCMode::ORAT)
         {
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(),
-                                                                  well_state, reportStepIdx,
-                                                                  pu.phase_pos[BlackoilPhases::Liquid], false);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                          wellModel_.schedule(),
+                                                                          well_state,
+                                                                          reportStepIdx,
+                                                                          pu.phase_pos[BlackoilPhases::Liquid],
+                                                                          false);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
 
             if (controls.oil_target < current_rate  ) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = controls.oil_target / current_rate;
                 return std::make_pair(Group::ProductionCMode::ORAT, scale);
@@ -228,17 +267,19 @@ checkGroupProductionConstraints(const Group& group,
     {
         if (currentControl != Group::ProductionCMode::WRAT)
         {
-
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(),
-                                                                  well_state, reportStepIdx,
-                                                                  pu.phase_pos[BlackoilPhases::Aqua], false);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                          wellModel_.schedule(),
+                                                                          well_state,
+                                                                          reportStepIdx,
+                                                                          pu.phase_pos[BlackoilPhases::Aqua],
+                                                                          false);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
 
             if (controls.water_target < current_rate  ) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = controls.water_target / current_rate;
                 return std::make_pair(Group::ProductionCMode::WRAT, scale);
@@ -249,15 +290,18 @@ checkGroupProductionConstraints(const Group& group,
     {
         if (currentControl != Group::ProductionCMode::GRAT)
         {
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(),
-                                                                  well_state, reportStepIdx,
-                                                                  pu.phase_pos[BlackoilPhases::Vapour], false);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                          wellModel_.schedule(),
+                                                                          well_state,
+                                                                          reportStepIdx,
+                                                                          pu.phase_pos[BlackoilPhases::Vapour],
+                                                                          false);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
             if (controls.gas_target < current_rate  ) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = controls.gas_target / current_rate;
                 return std::make_pair(Group::ProductionCMode::GRAT, scale);
@@ -268,22 +312,31 @@ checkGroupProductionConstraints(const Group& group,
     {
         if (currentControl != Group::ProductionCMode::LRAT)
         {
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(),
-                                                                  well_state, reportStepIdx,
-                                                                  pu.phase_pos[BlackoilPhases::Liquid], false);
-            current_rate += WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(),
-                                                                  well_state, reportStepIdx,
-                                                                  pu.phase_pos[BlackoilPhases::Aqua], false);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                          wellModel_.schedule(),
+                                                                          well_state,
+                                                                          reportStepIdx,
+                                                                          pu.phase_pos[BlackoilPhases::Liquid],
+                                                                          false);
+            current_rate += WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                          wellModel_.schedule(),
+                                                                          well_state,
+                                                                          reportStepIdx,
+                                                                          pu.phase_pos[BlackoilPhases::Aqua],
+                                                                          false);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
 
             bool skip = false;
             if (controls.liquid_target == controls.oil_target) {
-                double current_water_rate = WellGroupHelpers::sumWellSurfaceRates(group, wellModel_.schedule(),
-                                                                                  well_state, reportStepIdx,
-                                                                                  pu.phase_pos[BlackoilPhases::Aqua], false);
+                Scalar current_water_rate = WellGroupHelpers<Scalar>::sumWellSurfaceRates(group,
+                                                                                          wellModel_.schedule(),
+                                                                                          well_state,
+                                                                                          reportStepIdx,
+                                                                                          pu.phase_pos[BlackoilPhases::Aqua],
+                                                                                          false);
                 current_water_rate = wellModel_.comm().sum(current_water_rate);
                 if (std::abs(current_water_rate) < 1e-12) {
                     skip = true;
@@ -292,7 +345,7 @@ checkGroupProductionConstraints(const Group& group,
             }
 
             if (!skip && controls.liquid_target < current_rate ) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = controls.liquid_target / current_rate;
                 return std::make_pair(Group::ProductionCMode::LRAT, scale);
@@ -308,26 +361,35 @@ checkGroupProductionConstraints(const Group& group,
     {
         if (currentControl != Group::ProductionCMode::RESV)
         {
-            double current_rate = 0.0;
-            current_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
-                                                              well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Aqua], false);
-            current_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
-                                                              well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Liquid], false);
-            current_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
-                                                              well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Vapour], false);
+            Scalar current_rate = 0.0;
+            current_rate += WellGroupHelpers<Scalar>::sumWellResRates(group,
+                                                                      wellModel_.schedule(),
+                                                                      well_state,
+                                                                      reportStepIdx,
+                                                                      pu.phase_pos[BlackoilPhases::Aqua],
+                                                                      false);
+            current_rate += WellGroupHelpers<Scalar>::sumWellResRates(group,
+                                                                      wellModel_.schedule(),
+                                                                      well_state,
+                                                                      reportStepIdx,
+                                                                      pu.phase_pos[BlackoilPhases::Liquid],
+                                                                      false);
+            current_rate += WellGroupHelpers<Scalar>::sumWellResRates(group,
+                                                                      wellModel_.schedule(),
+                                                                      well_state,
+                                                                      reportStepIdx,
+                                                                      pu.phase_pos[BlackoilPhases::Vapour],
+                                                                      false);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
 
-            double target = controls.resv_target;
+            Scalar target = controls.resv_target;
             if (group.has_gpmaint_control(Group::ProductionCMode::RESV))
                 target = wellModel_.groupState().gpmaint_target(group.name());
 
             if ( target < current_rate ) {
-                double scale = 1.0;
+                Scalar scale = 1.0;
                 if (current_rate > 1e-12)
                     scale = target / current_rate;
                 return std::make_pair(Group::ProductionCMode::RESV, scale);
@@ -338,10 +400,11 @@ checkGroupProductionConstraints(const Group& group,
     {
         OPM_DEFLOG_THROW(std::runtime_error, "Group " + group.name() + "PRBL control for production groups not implemented", deferred_logger);
     }
-    return std::make_pair(Group::ProductionCMode::NONE, 1.0);
+    return std::make_pair(Group::ProductionCMode::NONE, Scalar(1.0));
 }
 
-bool BlackoilWellModelConstraints::
+template<class Scalar>
+bool BlackoilWellModelConstraints<Scalar>::
 checkGroupConstraints(const Group& group,
                       const int reportStepIdx,
                       DeferredLogger& deferred_logger) const
@@ -378,11 +441,12 @@ checkGroupConstraints(const Group& group,
     return violated;
 }
 
-void BlackoilWellModelConstraints::
+template<class Scalar>
+void BlackoilWellModelConstraints<Scalar>::
 actionOnBrokenConstraints(const Group& group,
                           const Group::InjectionCMode& newControl,
                           const Phase& controlPhase,
-                          GroupState& group_state,
+                          GroupState<Scalar>& group_state,
                           DeferredLogger& deferred_logger) const
 {
     auto oldControl = wellModel_.groupState().injection_control(group.name(), controlPhase);
@@ -400,14 +464,15 @@ actionOnBrokenConstraints(const Group& group,
     }
 }
 
-bool BlackoilWellModelConstraints::
+template<class Scalar>
+bool BlackoilWellModelConstraints<Scalar>::
 actionOnBrokenConstraints(const Group& group,
                           const int reportStepIdx,
                           const Group::GroupLimitAction group_limit_action,
                           const Group::ProductionCMode& newControl,
-                          const WellState& well_state,
+                          const WellState<Scalar>& well_state,
                           std::optional<std::string>& worst_offending_well,
-                          GroupState& group_state,
+                          GroupState<Scalar>& group_state,
                           DeferredLogger& deferred_logger) const
 {
 
@@ -416,9 +481,9 @@ actionOnBrokenConstraints(const Group& group,
         wellModel_.groupState().production_control(group.name());
 
     std::string ss;
-    switch(group_limit_action.allRates) {
+    switch (group_limit_action.allRates) {
     case Group::ExceedAction::NONE: {
-        if (oldControl != newControl && oldControl != Group::ProductionCMode::NONE) {
+        if (oldControl != newControl) {
             if ((group_limit_action.water == Group::ExceedAction::RATE &&
                  newControl == Group::ProductionCMode::WRAT) ||
                 (group_limit_action.gas == Group::ExceedAction::RATE &&
@@ -457,11 +522,10 @@ actionOnBrokenConstraints(const Group& group,
         break;
     }
     case Group::ExceedAction::WELL: {
-
         std::tie(worst_offending_well, std::ignore) =
-            WellGroupHelpers::worstOffendingWell(group, wellModel_.schedule(), reportStepIdx,
-                                                 newControl, wellModel_.phaseUsage(),
-                                                 wellModel_.comm(), well_state, deferred_logger);
+            WellGroupHelpers<Scalar>::worstOffendingWell(group, wellModel_.schedule(), reportStepIdx,
+                                                         newControl, wellModel_.phaseUsage(),
+                                                         wellModel_.comm(), well_state, deferred_logger);
         break;
     }
     case Group::ExceedAction::PLUG: {
@@ -493,14 +557,15 @@ actionOnBrokenConstraints(const Group& group,
     return changed;
 }
 
-bool BlackoilWellModelConstraints::
+template<class Scalar>
+bool BlackoilWellModelConstraints<Scalar>::
 updateGroupIndividualControl(const Group& group,
                              const int reportStepIdx,
-                             std::map<std::pair<std::string,Opm::Phase>,std::string>& switched_inj,
+                             std::map<std::pair<std::string,Phase>,std::string>& switched_inj,
                              std::map<std::string, std::string>& switched_prod,
                              std::map<std::string, std::pair<std::string, std::string>>& closed_offending_wells,
-                             GroupState& group_state,
-                             WellState& well_state,
+                             GroupState<Scalar>& group_state,
+                             WellState<Scalar>& well_state,
                              DeferredLogger& deferred_logger) const
 {
     bool changed = false;
@@ -520,13 +585,13 @@ updateGroupIndividualControl(const Group& group,
                                      Group::InjectionCMode2String(changed_this.first));
                 this->actionOnBrokenConstraints(group, changed_this.first, phase,
                                                 group_state, deferred_logger);
-                WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second,
-                                                                      group,
-                                                                      wellModel_.schedule(),
-                                                                      reportStepIdx,
-                                                                      /* isInjector */ false,
-                                                                      wellModel_.groupState(),
-                                                                      well_state);
+                WellGroupHelpers<Scalar>::updateWellRatesFromGroupTargetScale(changed_this.second,
+                                                                              group,
+                                                                              wellModel_.schedule(),
+                                                                              reportStepIdx,
+                                                                              /* isInjector */ false,
+                                                                              wellModel_.groupState(),
+                                                                              well_state);
                 changed = true;
             }
         }
@@ -549,13 +614,13 @@ updateGroupIndividualControl(const Group& group,
             if(changed) {
                 switched_prod.insert_or_assign(group.name(),
                                     Group::ProductionCMode2String(changed_this.first));
-                WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second,
-                                                                    group,
-                                                                    wellModel_.schedule(),
-                                                                    reportStepIdx,
-                                                                    /* isInjector */ false,
-                                                                    wellModel_.groupState(),
-                                                                    well_state);
+                WellGroupHelpers<Scalar>::updateWellRatesFromGroupTargetScale(changed_this.second,
+                                                                              group,
+                                                                              wellModel_.schedule(),
+                                                                              reportStepIdx,
+                                                                              /* isInjector */ false,
+                                                                              wellModel_.groupState(),
+                                                                              well_state);
             } else if (worst_offending_well) {
                 closed_offending_wells.insert_or_assign(group.name(), 
                             std::make_pair(Group::ProductionCMode2String(changed_this.first), *worst_offending_well));
@@ -566,4 +631,10 @@ updateGroupIndividualControl(const Group& group,
     return changed;
 }
 
+template class BlackoilWellModelConstraints<double>;
+
+#if FLOW_INSTANTIATE_FLOAT
+template class BlackoilWellModelConstraints<float>;
+#endif
+
 }
