@@ -46,7 +46,6 @@
 #include <memory>
 #include <optional>
 #include <string>
-#include <type_traits>
 #include <unordered_map>
 #include <unordered_set>
 #include <vector>
@@ -54,19 +53,19 @@
 namespace Opm {
     class DeferredLogger;
     class EclipseState;
-    class GasLiftGroupInfo;
-    class GasLiftSingleWellGeneric;
-    class GasLiftWellState;
+    template<class Scalar> class GasLiftGroupInfo;
+    template<class Scalar> class GasLiftSingleWellGeneric;
+    template<class Scalar> class GasLiftWellState;
     class Group;
     class GuideRateConfig;
-    class ParallelWellInfo;
+    template<class Scalar> class ParallelWellInfo;
     class RestartValue;
     class Schedule;
     struct SimulatorUpdate;
     class SummaryConfig;
-    class VFPProperties;
-    class WellInterfaceGeneric;
-    class WellState;
+    template<class Scalar> class VFPProperties;
+    template<class Scalar> class WellInterfaceGeneric;
+    template<class Scalar> class WellState;
 } // namespace Opm
 
 namespace Opm { namespace data {
@@ -79,13 +78,14 @@ namespace Opm { namespace data {
 namespace Opm {
 
 /// Class for handling the blackoil well model.
+template<class Scalar>
 class BlackoilWellModelGeneric
 {
 public:
     // ---------      Types      ---------
-    using GLiftOptWells = std::map<std::string, std::unique_ptr<GasLiftSingleWellGeneric>>;
-    using GLiftProdWells = std::map<std::string, const WellInterfaceGeneric*>;
-    using GLiftWellStateMap = std::map<std::string, std::unique_ptr<GasLiftWellState>>;
+    using GLiftOptWells = std::map<std::string, std::unique_ptr<GasLiftSingleWellGeneric<Scalar>>>;
+    using GLiftProdWells = std::map<std::string, const WellInterfaceGeneric<Scalar>*>;
+    using GLiftWellStateMap = std::map<std::string, std::unique_ptr<GasLiftWellState<Scalar>>>;
 
     BlackoilWellModelGeneric(Schedule& schedule,
                              const SummaryState& summaryState,
@@ -114,14 +114,14 @@ public:
     std::vector<Well> getLocalWells(const int timeStepIdx) const;
     const Schedule& schedule() const { return schedule_; }
     const PhaseUsage& phaseUsage() const { return phase_usage_; }
-    const GroupState& groupState() const { return this->active_wgstate_.group_state; }
-    std::vector<const WellInterfaceGeneric*> genericWells() const
+    const GroupState<Scalar>& groupState() const { return this->active_wgstate_.group_state; }
+    std::vector<const WellInterfaceGeneric<Scalar>*> genericWells() const
     { return {well_container_generic_.begin(), well_container_generic_.end()}; }
 
     /*
       Immutable version of the currently active wellstate.
     */
-    const WellState& wellState() const
+    const WellState<Scalar>& wellState() const
     {
         return this->active_wgstate_.well_state;
     }
@@ -129,7 +129,7 @@ public:
     /*
       Mutable version of the currently active wellstate.
     */
-    WellState& wellState()
+    WellState<Scalar>& wellState()
     {
         return this->active_wgstate_.well_state;
     }
@@ -138,26 +138,25 @@ public:
       Will return the currently active nupcolWellState; must initialize
       the internal nupcol wellstate with initNupcolWellState() first.
     */
-    const WellState& nupcolWellState() const
+    const WellState<Scalar>& nupcolWellState() const
     {
         return this->nupcol_wgstate_.well_state;
     }
-    GroupState& groupState() { return this->active_wgstate_.group_state; }
+    GroupState<Scalar>& groupState() { return this->active_wgstate_.group_state; }
 
     WellTestState& wellTestState() { return this->active_wgstate_.well_test_state; }
 
     const WellTestState& wellTestState() const { return this->active_wgstate_.well_test_state; }
 
-
-    double wellPI(const int well_index) const;
-    double wellPI(const std::string& well_name) const;
+    Scalar wellPI(const int well_index) const;
+    Scalar wellPI(const std::string& well_name) const;
 
     void updateEclWells(const int timeStepIdx,
                         const SimulatorUpdate& sim_update,
                         const SummaryState& st);
 
     void initFromRestartFile(const RestartValue& restartValues,
-                             WellTestState wtestState,
+                             std::unique_ptr<WellTestState> wtestState,
                              const std::size_t numCells,
                              bool handle_ms_well);
 
@@ -194,7 +193,7 @@ public:
     bool forceShutWellByName(const std::string& wellname,
                              const double simulation_time);
 
-    const std::vector<PerforationData>& perfData(const int well_idx) const
+    const std::vector<PerforationData<Scalar>>& perfData(const int well_idx) const
     { return well_perf_data_[well_idx]; }
 
     const Parallel::Communication& comm() const { return comm_; }
@@ -210,7 +209,8 @@ public:
     bool shouldBalanceNetwork(const int reportStepIndex,
                               const int iterationIdx) const;
 
-    void updateClosedWellsThisStep(const std::string& well_name) const {
+    void updateClosedWellsThisStep(const std::string& well_name) const
+    {
         this->closed_this_step_.insert(well_name);
     }
     bool wasDynamicallyShutThisTimeStep(const std::string& well_name) const;
@@ -254,7 +254,6 @@ public:
     }
 
 protected:
-
     /*
       The dynamic state of the well model is maintained with an instance
       of the WellState class. Currently we have
@@ -275,7 +274,6 @@ protected:
            nupcol_well_state_ and the initNupcolWellState() function.
     */
 
-
     /*
       Will return the last good wellstate. This is typcially used when
       initializing a new report step where the Schedule object might
@@ -283,25 +281,22 @@ protected:
       prevWellState() must have been stored with the commitWellState()
       function first.
     */
-    const WellState& prevWellState() const
+    const WellState<Scalar>& prevWellState() const
     {
         return this->last_valid_wgstate_.well_state;
     }
 
-
-    const WGState& prevWGState() const
+    const WGState<Scalar>& prevWGState() const
     {
         return this->last_valid_wgstate_;
     }
 
-
-
     /*
       Will store a copy of the input argument well_state in the
       last_valid_well_state_ member, that state can then be recovered
       with a subsequent call to resetWellState().
     */
-    void commitWGState(WGState wgstate)
+    void commitWGState(WGState<Scalar> wgstate)
     {
         this->last_valid_wgstate_ = std::move(wgstate);
     }
@@ -328,41 +323,46 @@ protected:
 
     /// \brief Create the parallel well information
     /// \param localWells The local wells from ECL schedule
-    std::vector<std::reference_wrapper<ParallelWellInfo>> createLocalParallelWellInfo(const std::vector<Well>& wells);
+    std::vector<std::reference_wrapper<ParallelWellInfo<Scalar>>>
+    createLocalParallelWellInfo(const std::vector<Well>& wells);
 
     void initializeWellProdIndCalculators();
     void initializeWellPerfData();
 
     bool wasDynamicallyShutThisTimeStep(const int well_index) const;
 
-    double updateNetworkPressures(const int reportStepIdx);
+    Scalar updateNetworkPressures(const int reportStepIdx);
 
     void updateWsolvent(const Group& group,
                         const int reportStepIdx,
-                        const WellState& wellState);
+                        const WellState<Scalar>& wellState);
     void setWsolvent(const Group& group,
                      const int reportStepIdx,
-                     double wsolvent);
-    virtual void calcRates(const int fipnum,
-                           const int pvtreg,
-                           const std::vector<double>& production_rates,
-                           std::vector<double>& resv_coeff) = 0;
-    virtual void calcInjRates(const int fipnum,
-                           const int pvtreg,
-                           std::vector<double>& resv_coeff) = 0;
+                     Scalar wsolvent);
+    virtual void calcResvCoeff(const int fipnum,
+                               const int pvtreg,
+                               const std::vector<Scalar>& production_rates,
+                               std::vector<Scalar>& resv_coeff) = 0;
+    virtual void calcInjResvCoeff(const int fipnum,
+                                  const int pvtreg,
+                                  std::vector<Scalar>& resv_coeff) = 0;
 
     void assignShutConnections(data::Wells& wsrpt,
                                const int reportStepIndex) const;
+    void assignWellTargets(data::Wells& wsrpt) const;
+    void assignProductionWellTargets(const Well& well, data::WellControlLimits& limits) const;
+    void assignInjectionWellTargets(const Well& well, data::WellControlLimits& limits) const;
     void assignGroupControl(const Group& group,
                             data::GroupData& gdata) const;
     void assignGroupValues(const int reportStepIdx,
                            std::map<std::string, data::GroupData>& gvalues) const;
-    void assignNodeValues(std::map<std::string, data::NodeData>& nodevalues, const int reportStepIdx) const;
+    void assignNodeValues(std::map<std::string, data::NodeData>& nodevalues,
+                          const int reportStepIdx) const;
 
     void calculateEfficiencyFactors(const int reportStepIdx);
 
     void checkGconsaleLimits(const Group& group,
-                             WellState& well_state,
+                             WellState<Scalar>& well_state,
                              const int reportStepIdx,
                              DeferredLogger& deferred_logger);
 
@@ -390,12 +390,12 @@ protected:
     void gasLiftOptimizationStage2(DeferredLogger& deferred_logger,
                                    GLiftProdWells& prod_wells,
                                    GLiftOptWells& glift_wells,
-                                   GasLiftGroupInfo& group_info,
+                                   GasLiftGroupInfo<Scalar>& group_info,
                                    GLiftWellStateMap& map,
                                    const int episodeIndex);
 
     virtual void computePotentials(const std::size_t widx,
-                                   const WellState& well_state_copy,
+                                   const WellState<Scalar>& well_state_copy,
                                    std::string& exc_msg,
                                    ExceptionType::ExcEnum& exc_type,
                                    DeferredLogger& deferred_logger) = 0;
@@ -408,11 +408,14 @@ protected:
 
     void initInjMult();
 
-
     void updateInjMult(DeferredLogger& deferred_logger);
     void updateInjFCMult(DeferredLogger& deferred_logger);
 
-    void updateFiltrationParticleVolume(const double dt, const std::size_t water_index);
+    void updateFiltrationModelsPostStep(const double dt,
+                                        const std::size_t water_index,
+                                        DeferredLogger& deferred_logger);
+
+    void updateFiltrationModelsPreStep(DeferredLogger& deferred_logger);
 
     // create the well container
     virtual void createWellContainer(const int time_step) = 0;
@@ -434,9 +437,14 @@ protected:
     std::vector<std::string> getWellsForTesting(const int timeStepIdx,
                                                 const double simulationTime);
 
-    using WellTracerRates = std::map<std::pair<std::string, std::string>, double>;
+    using WellTracerRates = std::map<std::pair<std::string, std::string>, Scalar>;
     void assignWellTracerRates(data::Wells& wsrpt,
                                const WellTracerRates& wellTracerRates) const;
+    using MswTracerRates = std::map<std::tuple<std::string, std::string, std::size_t>, Scalar>;
+    void assignMswTracerRates(data::Wells& wsrpt,
+                              const MswTracerRates& mswTracerRates) const;
+    void assignMassGasRate(data::Wells& wsrpt,
+                           const Scalar& gasDensity) const;
 
     Schedule& schedule_;
     const SummaryState& summaryState_;
@@ -453,7 +461,7 @@ protected:
     std::optional<int> last_run_wellpi_{};
 
     std::vector<Well> wells_ecl_;
-    std::vector<std::vector<PerforationData>> well_perf_data_;
+    std::vector<std::vector<PerforationData<Scalar>>> well_perf_data_;
 
     /// Connection index mappings
     class ConnectionIndexMap
@@ -545,29 +553,29 @@ protected:
     std::function<bool(const Well&)> not_on_process_{};
 
     // a vector of all the wells.
-    std::vector<WellInterfaceGeneric*> well_container_generic_{};
+    std::vector<WellInterfaceGeneric<Scalar>*> well_container_generic_{};
 
     std::vector<int> local_shut_wells_{};
 
-    std::vector<ParallelWellInfo> parallel_well_info_;
-    std::vector<std::reference_wrapper<ParallelWellInfo>> local_parallel_well_info_;
+    std::vector<ParallelWellInfo<Scalar>> parallel_well_info_;
+    std::vector<std::reference_wrapper<ParallelWellInfo<Scalar>>> local_parallel_well_info_;
 
-    std::vector<WellProdIndexCalculator> prod_index_calc_;
-    mutable ParallelWBPCalculation wbpCalculationService_;
+    std::vector<WellProdIndexCalculator<Scalar>> prod_index_calc_;
+    mutable ParallelWBPCalculation<Scalar> wbpCalculationService_;
 
     std::vector<int> pvt_region_idx_;
 
     mutable std::unordered_set<std::string> closed_this_step_;
 
     GuideRate guideRate_;
-    std::unique_ptr<VFPProperties> vfp_properties_{};
-    std::map<std::string, double> node_pressures_; // Storing network pressures for output.
+    std::unique_ptr<VFPProperties<Scalar>> vfp_properties_{};
+    std::map<std::string, Scalar> node_pressures_; // Storing network pressures for output.
 
     // previous injection multiplier, it is used in the injection multiplier calculation for WINJMULT keyword
-    std::unordered_map<std::string, std::vector<double>> prev_inj_multipliers_;
+    std::unordered_map<std::string, std::vector<Scalar>> prev_inj_multipliers_;
 
     // Handling for filter cake injection multipliers
-    std::unordered_map<std::string, WellFilterCake> filter_cake_;
+    std::unordered_map<std::string, WellFilterCake<Scalar>> filter_cake_;
 
     /*
       The various wellState members should be accessed and modified
@@ -575,9 +583,9 @@ protected:
       commitWellState(), resetWellState(), nupcolWellState() and
       updateNupcolWellState().
     */
-    WGState active_wgstate_;
-    WGState last_valid_wgstate_;
-    WGState nupcol_wgstate_;
+    WGState<Scalar> active_wgstate_;
+    WGState<Scalar> last_valid_wgstate_;
+    WGState<Scalar> nupcol_wgstate_;
 
     bool glift_debug = false;
 
@@ -587,13 +595,22 @@ protected:
 
     // Store maps of group name and new group controls for output
     std::map<std::string, std::string> switched_prod_groups_;
-    std::map<std::pair<std::string, Opm::Phase>, std::string> switched_inj_groups_;
+    std::map<std::pair<std::string, Phase>, std::string> switched_inj_groups_;
     // Store map of group name and close offending well for output
     std::map<std::string, std::pair<std::string, std::string>> closed_offending_wells_;
 
-
 private:
-    WellInterfaceGeneric* getGenWell(const std::string& well_name);
+    WellInterfaceGeneric<Scalar>* getGenWell(const std::string& well_name);
+
+    template <typename Iter, typename Body>
+    void wellUpdateLoop(Iter first, Iter last, const int timeStepIdx, Body&& body);
+
+    void updateEclWellsConstraints(const int timeStepIdx,
+                                   const SimulatorUpdate& sim_update,
+                                   const SummaryState& st);
+
+    void updateEclWellsCTFFromAction(const int timeStepIdx,
+                                     const SimulatorUpdate& sim_update);
 };
 
 
