@@ -29,8 +29,6 @@ namespace Dune {
 template<class Matrix> class UMFPack;
 }
 
-#include <memory>
-
 namespace Opm {
 
 class DeferredLogger;
@@ -42,17 +40,17 @@ namespace mswellhelpers
     /// Applies umfpack and checks for singularity
     template <typename MatrixType, typename VectorType>
     VectorType
-    applyUMFPack(const MatrixType& D,
-                 std::shared_ptr<Dune::UMFPack<MatrixType>>& linsolver,
+    applyUMFPack(Dune::UMFPack<MatrixType>& linsolver,
                  VectorType x);
 
 
 
     /// Applies umfpack and checks for singularity
-    template <typename MatrixType, typename VectorType>
+    template <typename VectorType, typename MatrixType>
     Dune::Matrix<typename MatrixType::block_type>
-    invertWithUMFPack(const MatrixType& D,
-                      std::shared_ptr<Dune::UMFPack<MatrixType> >& linsolver);
+    invertWithUMFPack(const int size,
+                      const int bsize,
+                      Dune::UMFPack<MatrixType>& linsolver);
 
 
 
@@ -61,13 +59,6 @@ namespace mswellhelpers
     VectorType
     invDX(const MatrixType& D, VectorType x, DeferredLogger& deferred_logger);
 
-
-    template <typename ValueType>
-    ValueType calculateFrictionFactor(const double area, const double diameter,
-                                      const ValueType& w, const double roughness,
-                                      const ValueType& mu);
-
-
     // calculating the friction pressure loss
     // l is the segment length
     // area is the segment cross area
@@ -76,26 +67,26 @@ namespace mswellhelpers
     // density is density
     // roughness is the absolute roughness
     // mu is the average phase viscosity
-    template <typename ValueType>
-    ValueType frictionPressureLoss(const double l, const double diameter,
-                                   const double area, const double roughness,
+    template <typename ValueType, typename Scalar>
+    ValueType frictionPressureLoss(const Scalar l, const Scalar diameter,
+                                   const Scalar area, const Scalar roughness,
                                    const ValueType& density,
                                    const ValueType& w, const ValueType& mu);
 
 
-    template <typename ValueType>
+    template <typename ValueType, typename Scalar>
     ValueType valveContrictionPressureLoss(const ValueType& mass_rate,
                                            const ValueType& density,
-                                           const double area_con, const double cv);
+                                           const Scalar area_con, const Scalar cv);
 
 
-    template <typename ValueType>
-    ValueType velocityHead(const double area, const ValueType& mass_rate,
+    template <typename ValueType, typename Scalar>
+    ValueType velocityHead(const Scalar area, const ValueType& mass_rate,
                            const ValueType& density);
 
 
     // calculating the viscosity of oil-water emulsion at local conditons
-    template <typename ValueType>
+    template <typename ValueType, typename Scalar>
     ValueType emulsionViscosity(const ValueType& water_fraction,
                                 const ValueType& water_viscosity,
                                 const ValueType& oil_fraction,
