@@ -20,9 +20,8 @@
 */
 
 #include <config.h>
-#include <opm/simulators/wells/MultisegmentWellAssemble.hpp>
 
-#include <opm/core/props/BlackoilPhases.hpp>
+#include <opm/simulators/wells/MultisegmentWellAssemble.hpp>
 
 #include <opm/material/fluidsystems/BlackOilFluidSystem.hpp>
 
@@ -30,6 +29,8 @@
 #include <opm/models/blackoil/blackoilonephaseindices.hh>
 #include <opm/models/blackoil/blackoiltwophaseindices.hh>
 
+#include <opm/simulators/utils/BlackoilPhases.hpp>
+
 #include <opm/simulators/wells/MultisegmentWellEquations.hpp>
 #include <opm/simulators/wells/MultisegmentWellPrimaryVariables.hpp>
 #include <opm/simulators/wells/WellAssemble.hpp>
@@ -81,17 +82,18 @@ private:
     MultisegmentWellEquations<Scalar,numWellEq,numEq>& eqns_; //!< Reference to equation system
 };
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
-assembleControlEq(const WellState& well_state,
-                  const GroupState& group_state,
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
+assembleControlEq(const WellState<Scalar>& well_state,
+                  const GroupState<Scalar>& group_state,
                   const Schedule& schedule,
                   const SummaryState& summaryState,
                   const Well::InjectionControls& inj_controls,
                   const Well::ProductionControls& prod_controls,
-                  const double rho,
+                  const Scalar rho,
                   const PrimaryVariables& primary_variables,
                   Equations& eqns1,
+                  const bool stopped_or_zero_target,
                   DeferredLogger& deferred_logger) const
 {
     static constexpr int Gas = BlackoilPhases::Vapour;
@@ -116,12 +118,12 @@ assembleControlEq(const WellState& well_state,
         return rates;
     };
 
-    if (well_.stopppedOrZeroRateTarget(summaryState, well_state)) {
+    if (stopped_or_zero_target) {
         control_eq = primary_variables.getWQTotal();
     } else if (well_.isInjector() ) {
         // Find scaling factor to get injection rate,
         const InjectorType injectorType = inj_controls.injector_type;
-        double scaling = 1.0;
+        Scalar scaling;
         const auto& pu = well_.phaseUsage();
         switch (injectorType) {
         case InjectorType::WATER:
@@ -197,27 +199,63 @@ assembleControlEq(const WellState& well_state,
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
-assemblePressureLoss(const int seg,
-                     const int seg_upwind,
-                     const EvalWell& accelerationPressureLoss,
-                     Equations& eqns1) const
-{
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
+assembleAccelerationTerm(const int seg_target,
+                         const int seg,
+                         const int seg_upwind,
+                         const EvalWell& accelerationTerm,
+                         Equations& eqns1) const
+{   // seg_target:  segment for which we are assembling the acc term
+    // seg:         segment for wich we have computed the term
+    // seg_upwind:  upwind segment to seg
+    // acceleration term shold be
+    //  * velocity head for seg_target if seg = seg_target
+    //  * negative velocity head for seg if seg != seg_target   
     MultisegmentWellEquationAccess<Scalar,numWellEq,Indices::numEq> eqns(eqns1);
-    eqns.residual()[seg][SPres] -= accelerationPressureLoss.value();
-    eqns.D()[seg][seg][SPres][SPres] -= accelerationPressureLoss.derivative(SPres + Indices::numEq);
-    eqns.D()[seg][seg][SPres][WQTotal] -= accelerationPressureLoss.derivative(WQTotal + Indices::numEq);
+    eqns.residual()[seg_target][SPres] -= accelerationTerm.value();
+    eqns.D()[seg_target][seg][SPres][SPres] -= accelerationTerm.derivative(SPres + Indices::numEq);
+    eqns.D()[seg_target][seg][SPres][WQTotal] -= accelerationTerm.derivative(WQTotal + Indices::numEq);
     if constexpr (has_wfrac_variable) {
-        eqns.D()[seg][seg_upwind][SPres][WFrac] -= accelerationPressureLoss.derivative(WFrac + Indices::numEq);
+        eqns.D()[seg_target][seg_upwind][SPres][WFrac] -= accelerationTerm.derivative(WFrac + Indices::numEq);
     }
     if constexpr (has_gfrac_variable) {
-        eqns.D()[seg][seg_upwind][SPres][GFrac] -= accelerationPressureLoss.derivative(GFrac + Indices::numEq);
+        eqns.D()[seg_target][seg_upwind][SPres][GFrac] -= accelerationTerm.derivative(GFrac + Indices::numEq);
     }
+}                     
+
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
+assembleHydroPressureLoss(const int seg,
+                          const int seg_density,
+                          const EvalWell& hydro_pressure_drop_seg,
+                          Equations& eqns1) const
+{
+    MultisegmentWellEquationAccess<Scalar,numWellEq,Indices::numEq> eqns(eqns1);
+    eqns.residual()[seg][SPres] -= hydro_pressure_drop_seg.value();
+    for (int pv_idx = 0; pv_idx < numWellEq; ++pv_idx) {
+        eqns.D()[seg][seg_density][SPres][pv_idx] -= hydro_pressure_drop_seg.derivative(pv_idx + Indices::numEq);
+    }
+
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
+assemblePressureEqExtraDerivatives(const int seg,
+                                   const int seg_upwind,
+                                   const EvalWell& extra_derivatives,
+                                   Equations& eqns1) const
+{
+    MultisegmentWellEquationAccess<Scalar,numWellEq,Indices::numEq> eqns(eqns1);
+    // disregard residual
+    // Frac - derivatives are zero (they belong to upwind^2)
+    eqns.D()[seg][seg_upwind][SPres][SPres] += extra_derivatives.derivative(SPres + Indices::numEq);
+    eqns.D()[seg][seg_upwind][SPres][WQTotal] += extra_derivatives.derivative(WQTotal + Indices::numEq);
+}
+
+
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
 assemblePressureEq(const int seg,
                    const int seg_upwind,
                    const int outlet_segment_index,
@@ -228,7 +266,7 @@ assemblePressureEq(const int seg,
                    bool gfrac) const
 {
     MultisegmentWellEquationAccess<Scalar,numWellEq,Indices::numEq> eqns(eqns1);
-    eqns.residual()[seg][SPres] = pressure_equation.value();
+    eqns.residual()[seg][SPres] += pressure_equation.value();
     eqns.D()[seg][seg][SPres][SPres] += pressure_equation.derivative(SPres + Indices::numEq);
     eqns.D()[seg][seg][SPres][WQTotal] += pressure_equation.derivative(WQTotal + Indices::numEq);
     if (wfrac) {
@@ -245,8 +283,8 @@ assemblePressureEq(const int seg,
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
 assembleTrivialEq(const int seg,
                   const Scalar value,
                   Equations& eqns1) const
@@ -256,8 +294,8 @@ assembleTrivialEq(const int seg,
     eqns.D()[seg][seg][SPres][WQTotal] = 1.;
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
 assembleAccumulationTerm(const int seg,
                          const int comp_idx,
                          const EvalWell& accumulation_term,
@@ -270,8 +308,8 @@ assembleAccumulationTerm(const int seg,
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
 assembleOutflowTerm(const int seg,
                     const int seg_upwind,
                     const int comp_idx,
@@ -290,8 +328,8 @@ assembleOutflowTerm(const int seg,
     // pressure derivative should be zero
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
 assembleInflowTerm(const int seg,
                    const int inlet,
                    const int inlet_upwind,
@@ -311,8 +349,8 @@ assembleInflowTerm(const int seg,
     // pressure derivative should be zero
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellAssemble<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellAssemble<FluidSystem,Indices>::
 assemblePerforationEq(const int seg,
                       const int cell_idx,
                       const int comp_idx,
@@ -338,34 +376,43 @@ assemblePerforationEq(const int seg,
     }
 }
 
-#define INSTANCE(...) \
-template class MultisegmentWellAssemble<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,__VA_ARGS__,double>;
+template<class Scalar>
+using FS = BlackOilFluidSystem<Scalar,BlackOilDefaultIndexTraits>;
 
-// One phase
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>)
+#define INSTANTIATE(T,...) \
+    template class MultisegmentWellAssemble<FS<T>,__VA_ARGS__>;
 
-// Two phase
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>)
-// Blackoil
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)
-INSTANCE(BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)
+#define INSTANTIATE_TYPE(T)                                                  \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<1u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,true,false,0u,0u>)
+
+INSTANTIATE_TYPE(double)
+
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
 }
