@@ -20,11 +20,13 @@
 */
 
 #include <config.h>
-#include <opm/common/TimingMacros.hpp>
-
 #include <opm/simulators/wells/MultisegmentWellEquations.hpp>
+
 #include <dune/istl/umfpack.hh>
 
+#include <opm/common/ErrorMacros.hpp>
+#include <opm/common/TimingMacros.hpp>
+
 #include <opm/input/eclipse/Schedule/MSW/WellSegments.hpp>
 
 #if COMPILE_BDA_BRIDGE
@@ -53,8 +55,7 @@ MultisegmentWellEquations(const MultisegmentWellGeneric<Scalar>& well)
 
 template<class Scalar, int numWellEq, int numEq>
 void MultisegmentWellEquations<Scalar,numWellEq,numEq>::
-init(const int num_cells,
-     const int numPerfs,
+init(const int numPerfs,
      const std::vector<int>& cells,
      const std::vector<std::vector<int>>& segment_inlets,
      const std::vector<std::vector<int>>& perforations)
@@ -76,8 +77,8 @@ init(const int num_cells,
         }
         duneD_.setSize(well_.numberOfSegments(), well_.numberOfSegments(), nnz_d);
     }
-    duneB_.setSize(well_.numberOfSegments(), num_cells, numPerfs);
-    duneC_.setSize(well_.numberOfSegments(), num_cells, numPerfs);
+    duneB_.setSize(well_.numberOfSegments(), numPerfs, numPerfs);
+    duneC_.setSize(well_.numberOfSegments(), numPerfs, numPerfs);
 
     // we need to add the off diagonal ones
     for (auto row = duneD_.createbegin(),
@@ -106,8 +107,7 @@ init(const int num_cells,
               end = duneC_.createend(); row != end; ++row) {
         // the number of the row corresponds to the segment number now.
         for (const int& perf : perforations[row.index()]) {
-            const int cell_idx = cells[perf];
-            row.insert(cell_idx);
+            row.insert(perf);
         }
     }
 
@@ -116,12 +116,14 @@ init(const int num_cells,
               end = duneB_.createend(); row != end; ++row) {
         // the number of the row corresponds to the segment number now.
         for (const int& perf : perforations[row.index()]) {
-            const int cell_idx = cells[perf];
-            row.insert(cell_idx);
+            row.insert(perf);
         }
     }
 
     resWell_.resize(well_.numberOfSegments());
+
+    // Store the global index of well perforated cells
+    cells_ = cells;
 }
 
 template<class Scalar, int numWellEq, int numEq>
@@ -167,7 +169,12 @@ void MultisegmentWellEquations<Scalar,numWellEq,numEq>::createSolver()
         return;
     }
 
-    duneDSolver_ = std::make_shared<Dune::UMFPack<DiagMatWell>>(duneD_, 0);
+    if constexpr (std::is_same_v<Scalar,float>) {
+        OPM_THROW(std::runtime_error, "MultisegmentWell support requires UMFPACK, "
+                                      "and UMFPACK does not support float");
+    } else {
+        duneDSolver_ = std::make_shared<Dune::UMFPack<DiagMatWell>>(duneD_, 0);
+    }
 #else
     OPM_THROW(std::runtime_error, "MultisegmentWell support requires UMFPACK. "
               "Reconfigure opm-simulators with SuiteSparse/UMFPACK support and recompile.");
@@ -202,7 +209,7 @@ recoverSolutionWell(const BVector& x, BVectorWell& xw) const
 #if COMPILE_BDA_BRIDGE
 template<class Scalar, int numWellEq, int numEq>
 void MultisegmentWellEquations<Scalar,numWellEq,numEq>::
-extract(WellContributions& wellContribs) const
+extract(WellContributions<Scalar>& wellContribs) const
 {
     unsigned int Mb = duneB_.N();       // number of blockrows in duneB_, duneC_ and duneD_
     unsigned int BnumBlocks = duneB_.nonzeroes();
@@ -225,46 +232,50 @@ extract(WellContributions& wellContribs) const
     }
 
     // duneD
-    Dune::UMFPack<DiagMatWell> umfpackMatrix(duneD_, 0);
-    double* Dvals = umfpackMatrix.getInternalMatrix().getValues();
-    auto* Dcols = umfpackMatrix.getInternalMatrix().getColStart();
-    auto* Drows = umfpackMatrix.getInternalMatrix().getRowIndex();
+    if constexpr (std::is_same_v<Scalar,float>) {
+        OPM_THROW(std::runtime_error, "Cannot use UMFPack with floats");
+    } else {
+        Dune::UMFPack<DiagMatWell> umfpackMatrix(duneD_, 0);
+        double* Dvals = umfpackMatrix.getInternalMatrix().getValues();
+        auto* Dcols = umfpackMatrix.getInternalMatrix().getColStart();
+        auto* Drows = umfpackMatrix.getInternalMatrix().getRowIndex();
 
-    // duneB
-    std::vector<unsigned int> Bcols;
-    std::vector<unsigned int> Brows;
-    std::vector<double> Bvals;
-    Bcols.reserve(BnumBlocks);
-    Brows.reserve(Mb+1);
-    Bvals.reserve(BnumBlocks * numEq * numWellEq);
-    Brows.emplace_back(0);
-    unsigned int sumBlocks = 0;
-    for (auto rowB = duneB_.begin(); rowB != duneB_.end(); ++rowB) {
-        int sizeRow = 0;
-        for (auto colB = rowB->begin(), endB = rowB->end(); colB != endB; ++colB) {
-            Bcols.emplace_back(colB.index());
-            for (int i = 0; i < numWellEq; ++i) {
-                for (int j = 0; j < numEq; ++j) {
-                    Bvals.emplace_back((*colB)[i][j]);
+        // duneB
+        std::vector<unsigned int> Bcols;
+        std::vector<unsigned int> Brows;
+        std::vector<double> Bvals;
+        Bcols.reserve(BnumBlocks);
+        Brows.reserve(Mb+1);
+        Bvals.reserve(BnumBlocks * numEq * numWellEq);
+        Brows.emplace_back(0);
+        unsigned int sumBlocks = 0;
+        for (auto rowB = duneB_.begin(); rowB != duneB_.end(); ++rowB) {
+            int sizeRow = 0;
+            for (auto colB = rowB->begin(), endB = rowB->end(); colB != endB; ++colB) {
+                Bcols.emplace_back(colB.index());
+                for (int i = 0; i < numWellEq; ++i) {
+                    for (int j = 0; j < numEq; ++j) {
+                        Bvals.emplace_back((*colB)[i][j]);
+                    }
                 }
+                sizeRow++;
             }
-            sizeRow++;
+            sumBlocks += sizeRow;
+            Brows.emplace_back(sumBlocks);
         }
-        sumBlocks += sizeRow;
-        Brows.emplace_back(sumBlocks);
-    }
 
-    wellContribs.addMultisegmentWellContribution(numEq,
-                                                 numWellEq,
-                                                 Mb,
-                                                 Bvals,
-                                                 Bcols,
-                                                 Brows,
-                                                 DnumBlocks,
-                                                 Dvals,
-                                                 Dcols,
-                                                 Drows,
-                                                 Cvals);
+        wellContribs.addMultisegmentWellContribution(numEq,
+                                                     numWellEq,
+                                                     Mb,
+                                                     Bvals,
+                                                     Bcols,
+                                                     Brows,
+                                                     DnumBlocks,
+                                                     Dvals,
+                                                     Dcols,
+                                                     Drows,
+                                                     Cvals);
+    }
 }
 #endif
 
@@ -288,11 +299,12 @@ extract(SparseMatrixAdapter& jacobian) const
     for (std::size_t rowC = 0; rowC < duneC_.N(); ++rowC) {
         for (auto colC = duneC_[rowC].begin(),
                   endC = duneC_[rowC].end(); colC != endC; ++colC) {
-            const auto row_index = colC.index();
+            // map the well perforated cell index to global cell index
+            const auto row_index = cells_[colC.index()];
             for (std::size_t rowB = 0; rowB < duneB_.N(); ++rowB) {
                 for (auto colB = duneB_[rowB].begin(),
                           endB = duneB_[rowB].end(); colB != endB; ++colB) {
-                    const auto col_index = colB.index();
+                    const auto col_index = cells_[colB.index()];
                     OffDiagMatrixBlockWellType tmp1;
                     detail::multMatrixImpl(invDuneD[rowC][rowB], (*colB), tmp1, std::true_type());
                     typename SparseMatrixAdapter::MatrixBlock tmp2;
@@ -311,19 +323,21 @@ extractCPRPressureMatrix(PressureMatrix& jacobian,
                          const BVector& weights,
                          const int pressureVarIndex,
                          const bool /*use_well_weights*/,
-                         const WellInterfaceGeneric& well,
+                         const WellInterfaceGeneric<Scalar>& well,
                          const int seg_pressure_var_ind,
-                         const WellState& well_state) const
+                         const WellState<Scalar>& well_state) const
 {
     // Add the pressure contribution to the cpr system for the well
 
     // Add for coupling from well to reservoir
-    const int welldof_ind = duneC_.M() + well.indexOfWell();
+    const int number_cells = weights.size();
+    const int welldof_ind = number_cells + well.indexOfWell();
     if (!well.isPressureControlled(well_state)) {
         for (std::size_t rowC = 0; rowC < duneC_.N(); ++rowC) {
             for (auto colC = duneC_[rowC].begin(),
                       endC = duneC_[rowC].end(); colC != endC; ++colC) {
-                const auto row_index = colC.index();
+                // map the well perforated cell index to global cell index
+                const auto row_index = cells_[colC.index()];
                 const auto& bw = weights[row_index];
                 double matel = 0.0;
 
@@ -343,7 +357,8 @@ extractCPRPressureMatrix(PressureMatrix& jacobian,
         for (std::size_t rowB = 0; rowB < duneB_.N(); ++rowB) {
             for (auto colB = duneB_[rowB].begin(),
                       endB = duneB_[rowB].end(); colB != endB; ++colB) {
-                const auto col_index = colB.index();
+                // map the well perforated cell index to global cell index
+                const auto col_index = cells_[colB.index()];
                 const auto& bw = weights[col_index];
                 well_weight += bw;
                 num_perfs += 1;
@@ -359,7 +374,8 @@ extractCPRPressureMatrix(PressureMatrix& jacobian,
             const auto& bw = well_weight;
             for (auto colB = duneB_[rowB].begin(),
                       endB = duneB_[rowB].end(); colB != endB; ++colB) {
-                const auto col_index = colB.index();
+                // map the well perforated cell index to global cell index
+                const auto col_index = cells_[colB.index()];
                 double matel = 0.0;
                 for (std::size_t i = 0; i< bw.size(); ++i) {
                     matel += bw[i] *(*colB)[i][pressureVarIndex];
@@ -386,27 +402,34 @@ extractCPRPressureMatrix(PressureMatrix& jacobian,
     }
 }
 
-#define INSTANCE(numWellEq, numEq) \
-template class MultisegmentWellEquations<double,numWellEq,numEq>; \
-template void MultisegmentWellEquations<double,numWellEq,numEq>:: \
-    extract(Linear::IstlSparseMatrixAdapter<MatrixBlock<double,numEq,numEq>>&) const; \
-template void MultisegmentWellEquations<double,numWellEq,numEq>:: \
-    extractCPRPressureMatrix(Dune::BCRSMatrix<MatrixBlock<double,1,1>>&, \
-                             const MultisegmentWellEquations<double,numWellEq,numEq>::BVector&, \
-                             const int, \
-                             const bool, \
-                             const WellInterfaceGeneric&, \
-                             const int, \
-                             const WellState&) const;
+#define INSTANTIATE(T, numWellEq, numEq)                                                       \
+    template class MultisegmentWellEquations<T,numWellEq,numEq>;                               \
+    template void MultisegmentWellEquations<T,numWellEq,numEq>::                               \
+        extract(Linear::IstlSparseMatrixAdapter<MatrixBlock<T,numEq,numEq>>&) const;           \
+    template void MultisegmentWellEquations<T,numWellEq,numEq>::                               \
+        extractCPRPressureMatrix(Dune::BCRSMatrix<MatrixBlock<T,1,1>>&,                        \
+                                 const MultisegmentWellEquations<T,numWellEq,numEq>::BVector&, \
+                                 const int,                                                    \
+                                 const bool,                                                   \
+                                 const WellInterfaceGeneric<T>&,                               \
+                                 const int,                                                    \
+                                 const WellState<T>&) const;
 
-INSTANCE(2,1)
-INSTANCE(2,2)
-INSTANCE(2,6)
-INSTANCE(3,2)
-INSTANCE(3,3)
-INSTANCE(3,4)
-INSTANCE(4,3)
-INSTANCE(4,4)
-INSTANCE(4,5)
+#define INSTANTIATE_TYPE(T) \
+    INSTANTIATE(T,2,1)      \
+    INSTANTIATE(T,2,2)      \
+    INSTANTIATE(T,2,6)      \
+    INSTANTIATE(T,3,2)      \
+    INSTANTIATE(T,3,3)      \
+    INSTANTIATE(T,3,4)      \
+    INSTANTIATE(T,4,3)      \
+    INSTANTIATE(T,4,4)      \
+    INSTANTIATE(T,4,5)
+
+INSTANTIATE_TYPE(double)
+
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
 }
