@@ -39,10 +39,10 @@ namespace Opm
 template<class Scalar, int numWellEq, int numEq> class MultisegmentWellEquationAccess;
 template<class Scalar> class MultisegmentWellGeneric;
 #if COMPILE_BDA_BRIDGE
-class WellContributions;
+template<class Scalar> class WellContributions;
 #endif
-class WellInterfaceGeneric;
-class WellState;
+template<class Scalar> class WellInterfaceGeneric;
+template<class Scalar> class WellState;
 
 template<class Scalar, int numWellEq, int numEq>
 class MultisegmentWellEquations
@@ -70,13 +70,11 @@ public:
     MultisegmentWellEquations(const MultisegmentWellGeneric<Scalar>& well);
 
     //! \brief Setup sparsity pattern for the matrices.
-    //! \param num_cells Total number of cells
     //! \param numPerfs Number of perforations
     //! \param cells Cell indices for perforations
     //! \param segment_inlets Cell indices for segment inlets
     //! \param segment_perforations Cell indices for segment perforations
-    void init(const int num_cells,
-              const int numPerfs,
+    void init(const int numPerfs,
               const std::vector<int>& cells,
               const std::vector<std::vector<int>>& segment_inlets,
               const std::vector<std::vector<int>>& segment_perforations);
@@ -105,7 +103,7 @@ public:
 
 #if COMPILE_BDA_BRIDGE
     //! \brief Add the matrices of this well to the WellContributions object.
-    void extract(WellContributions& wellContribs) const;
+    void extract(WellContributions<Scalar>& wellContribs) const;
 #endif
 
     //! \brief Add the matrices of this well to the sparse matrix adapter.
@@ -118,9 +116,9 @@ public:
                                   const BVector& weights,
                                   const int pressureVarIndex,
                                   const bool /*use_well_weights*/,
-                                  const WellInterfaceGeneric& well,
+                                  const WellInterfaceGeneric<Scalar>& well,
                                   const int seg_pressure_var_ind,
-                                  const WellState& well_state) const;
+                                  const WellState<Scalar>& well_state) const;
 
     //! \brief Returns a const reference to the residual.
     const BVectorWell& residual() const
@@ -145,6 +143,9 @@ public:
     BVectorWell resWell_;
 
     const MultisegmentWellGeneric<Scalar>& well_; //!< Reference to well
+
+    // Store the global index of well perforated cells
+    std::vector<int> cells_;
 };
 
 }
