@@ -67,9 +67,9 @@ MultisegmentWellEval(WellInterfaceIndices<FluidSystem,Indices>& baseif)
 template<typename FluidSystem, typename Indices>
 void
 MultisegmentWellEval<FluidSystem,Indices>::
-initMatrixAndVectors(const int num_cells)
+initMatrixAndVectors()
 {
-    linSys_.init(num_cells, baseif_.numPerfs(),
+    linSys_.init(baseif_.numPerfs(),
                  baseif_.cells(), segments_.inlets(),
                  segments_.perforations());
     primary_variables_.resize(this->numberOfSegments());
@@ -78,36 +78,36 @@ initMatrixAndVectors(const int num_cells)
 template<typename FluidSystem, typename Indices>
 ConvergenceReport
 MultisegmentWellEval<FluidSystem,Indices>::
-getWellConvergence(const WellState& well_state,
-                   const std::vector<double>& B_avg,
+getWellConvergence(const WellState<Scalar>& well_state,
+                   const std::vector<Scalar>& B_avg,
                    DeferredLogger& deferred_logger,
-                   const double max_residual_allowed,
-                   const double tolerance_wells,
-                   const double relaxed_inner_tolerance_flow_ms_well,
-                   const double tolerance_pressure_ms_wells,
-                   const double relaxed_inner_tolerance_pressure_ms_well,
+                   const Scalar max_residual_allowed,
+                   const Scalar tolerance_wells,
+                   const Scalar relaxed_inner_tolerance_flow_ms_well,
+                   const Scalar tolerance_pressure_ms_wells,
+                   const Scalar relaxed_inner_tolerance_pressure_ms_well,
                    const bool relax_tolerance, 
                    const bool well_is_stopped) const
 {
     assert(int(B_avg.size()) == baseif_.numComponents());
 
     // checking if any residual is NaN or too large. The two large one is only handled for the well flux
-    std::vector<std::vector<double>> abs_residual(this->numberOfSegments(),
-                                                  std::vector<double>(numWellEq, 0.0));
+    std::vector<std::vector<Scalar>> abs_residual(this->numberOfSegments(),
+                                                  std::vector<Scalar>(numWellEq, 0.0));
     for (int seg = 0; seg < this->numberOfSegments(); ++seg) {
         for (int eq_idx = 0; eq_idx < numWellEq; ++eq_idx) {
             abs_residual[seg][eq_idx] = std::abs(linSys_.residual()[seg][eq_idx]);
         }
     }
 
-    std::vector<double> maximum_residual(numWellEq, 0.0);
+    std::vector<Scalar> maximum_residual(numWellEq, 0.0);
 
     ConvergenceReport report;
     // TODO: the following is a little complicated, maybe can be simplified in some way?
     for (int eq_idx = 0; eq_idx < numWellEq; ++eq_idx) {
         for (int seg = 0; seg < this->numberOfSegments(); ++seg) {
             if (eq_idx < baseif_.numComponents()) { // phase or component mass equations
-                const double flux_residual = B_avg[eq_idx] * abs_residual[seg][eq_idx];
+                const Scalar flux_residual = B_avg[eq_idx] * abs_residual[seg][eq_idx];
                 if (flux_residual > maximum_residual[eq_idx]) {
                     maximum_residual[eq_idx] = flux_residual;
                 }
@@ -115,7 +115,7 @@ getWellConvergence(const WellState& well_state,
                 // for the top segment (seg == 0), it is control equation, will be checked later separately
                 if (seg > 0) {
                     // Pressure equation
-                    const double pressure_residual = abs_residual[seg][eq_idx];
+                    const Scalar pressure_residual = abs_residual[seg][eq_idx];
                     if (pressure_residual > maximum_residual[eq_idx]) {
                         maximum_residual[eq_idx] = pressure_residual;
                     }
@@ -127,7 +127,7 @@ getWellConvergence(const WellState& well_state,
     using CR = ConvergenceReport;
     for (int eq_idx = 0; eq_idx < numWellEq; ++eq_idx) {
         if (eq_idx < baseif_.numComponents()) { // phase or component mass equations
-            const double flux_residual = maximum_residual[eq_idx];
+            const Scalar flux_residual = maximum_residual[eq_idx];
             // TODO: the report can not handle the segment number yet.
 
             if (std::isnan(flux_residual)) {
@@ -140,7 +140,7 @@ getWellConvergence(const WellState& well_state,
                 report.setWellFailed({CR::WellFailure::Type::MassBalance, CR::Severity::Normal, eq_idx, baseif_.name()});
             }
         } else { // pressure equation
-            const double pressure_residual = maximum_residual[eq_idx];
+            const Scalar pressure_residual = maximum_residual[eq_idx];
             const int dummy_component = -1;
             if (std::isnan(pressure_residual)) {
                 report.setWellFailed({CR::WellFailure::Type::Pressure, CR::Severity::NotANumber, dummy_component, baseif_.name()});
@@ -170,7 +170,7 @@ getWellConvergence(const WellState& well_state,
     // for BHP or THP controlled wells, we need to make sure the flow direction is correct
     if (!well_is_stopped && baseif_.isPressureControlled(well_state)) {
         // checking the flow direction
-        const double sign = baseif_.isProducer() ? -1. : 1.;
+        const Scalar sign = baseif_.isProducer() ? -1. : 1.;
         const auto weight_total_flux = this->primary_variables_.getWQTotal() * sign;
         constexpr int dummy_phase = -1;
         if (weight_total_flux < 0.) {
@@ -199,7 +199,7 @@ template<typename FluidSystem, typename Indices>
 void
 MultisegmentWellEval<FluidSystem,Indices>::
 assembleAccelerationPressureLoss(const int seg,
-                                 WellState& well_state)
+                                 WellState<Scalar>& well_state)
 {
     // Computes and assembles p-drop due to acceleration
     assert(seg != 0); // top segment can not enter here
@@ -207,7 +207,7 @@ assembleAccelerationPressureLoss(const int seg,
     const auto& segment_set = this->segmentSet();
 
     // Add segment head
-    const double seg_area = segment_set[seg].crossArea();
+    const Scalar seg_area = segment_set[seg].crossArea();
     const EvalWell signed_velocity_head = segments_.accelerationPressureLossContribution(seg, seg_area);
     segments.pressure_drop_accel[seg] = signed_velocity_head.value();
     
@@ -225,7 +225,8 @@ assembleAccelerationPressureLoss(const int seg,
     // Subtract inlet head(s), opposite signs from above
     for (const int inlet : segments_.inlets(seg)) {
         // area used in formula is max of areas
-        const double inlet_area = std::max(seg_area, segment_set[inlet].crossArea());
+        const Scalar inlet_area = std::max(seg_area,
+                                           static_cast<Scalar>(segment_set[inlet].crossArea()));
         const EvalWell signed_velocity_head_inlet = segments_.accelerationPressureLossContribution(inlet, inlet_area);
         segments.pressure_drop_accel[seg] -= signed_velocity_head_inlet.value();
 
@@ -246,7 +247,7 @@ template<typename FluidSystem, typename Indices>
 void
 MultisegmentWellEval<FluidSystem,Indices>::
 assembleDefaultPressureEq(const int seg,
-                          WellState& well_state,
+                          WellState<Scalar>& well_state,
                           const bool use_average_density)
 {
     assert(seg != 0); // not top segment
@@ -293,7 +294,7 @@ void
 MultisegmentWellEval<FluidSystem,Indices>::
 assembleICDPressureEq(const int seg,
                       const UnitSystem& unit_system,
-                      WellState& well_state,
+                      WellState<Scalar>& well_state,
                       const SummaryState& summary_state,
                       const bool use_average_density,
                       DeferredLogger& deferred_logger)
@@ -377,7 +378,7 @@ template<typename FluidSystem, typename Indices>
 void
 MultisegmentWellEval<FluidSystem,Indices>::
 assembleAccelerationAndHydroPressureLosses(const int seg,
-                                           WellState& well_state,
+                                           WellState<Scalar>& well_state,
                                            const bool use_average_density)
 {
     if (this->accelerationalPressureLossConsidered()) {
@@ -409,7 +410,7 @@ void
 MultisegmentWellEval<FluidSystem,Indices>::
 assemblePressureEq(const int seg,
                    const UnitSystem& unit_system,
-                   WellState& well_state,
+                   WellState<Scalar>& well_state,
                    const SummaryState& summary_state,
                    const bool use_average_density,
                    DeferredLogger& deferred_logger)
@@ -437,7 +438,7 @@ getFiniteWellResiduals(const std::vector<Scalar>& B_avg,
 
     for (int seg = 0; seg < this->numberOfSegments(); ++seg) {
         for (int eq_idx = 0; eq_idx < numWellEq; ++eq_idx) {
-            double residual = 0.;
+            Scalar residual = 0.;
             if (eq_idx < baseif_.numComponents()) {
                 residual = std::abs(linSys_.residual()[seg][eq_idx]) * B_avg[eq_idx];
             } else {
@@ -459,7 +460,7 @@ getFiniteWellResiduals(const std::vector<Scalar>& B_avg,
 
     // handling the control equation residual
     {
-        const double control_residual = std::abs(linSys_.residual()[0][numWellEq - 1]);
+        const Scalar control_residual = std::abs(linSys_.residual()[0][numWellEq - 1]);
         if (std::isnan(control_residual) || std::isinf(control_residual)) {
            deferred_logger.debug(fmt::format("nan or inf value for control residual for well {}", baseif_.name()));
            return {false, residuals};
@@ -471,14 +472,14 @@ getFiniteWellResiduals(const std::vector<Scalar>& B_avg,
 }
 
 template<typename FluidSystem, typename Indices>
-double
+typename MultisegmentWellEval<FluidSystem,Indices>::Scalar
 MultisegmentWellEval<FluidSystem,Indices>::
-getControlTolerance(const WellState& well_state,
-                    const double tolerance_wells,
-                    const double tolerance_pressure_ms_wells,
+getControlTolerance(const WellState<Scalar>& well_state,
+                    const Scalar tolerance_wells,
+                    const Scalar tolerance_pressure_ms_wells,
                     DeferredLogger& deferred_logger) const
 {
-    double control_tolerance = 0.;
+    Scalar control_tolerance = 0.;
 
     const int well_index = baseif_.indexOfWell();
     const auto& ws = well_state.well(well_index);
@@ -538,19 +539,19 @@ getControlTolerance(const WellState& well_state,
 }
 
 template<typename FluidSystem, typename Indices>
-double
+typename MultisegmentWellEval<FluidSystem,Indices>::Scalar
 MultisegmentWellEval<FluidSystem,Indices>::
-getResidualMeasureValue(const WellState& well_state,
-                        const std::vector<double>& residuals,
-                        const double tolerance_wells,
-                        const double tolerance_pressure_ms_wells,
+getResidualMeasureValue(const WellState<Scalar>& well_state,
+                        const std::vector<Scalar>& residuals,
+                        const Scalar tolerance_wells,
+                        const Scalar tolerance_pressure_ms_wells,
                         DeferredLogger& deferred_logger) const
 {
     assert(int(residuals.size()) == numWellEq + 1);
 
-    const double rate_tolerance = tolerance_wells;
+    const Scalar rate_tolerance = tolerance_wells;
     [[maybe_unused]] int count = 0;
-    double sum = 0;
+    Scalar sum = 0;
     for (int eq_idx = 0; eq_idx < numWellEq - 1; ++eq_idx) {
         if (residuals[eq_idx] > rate_tolerance) {
             sum += residuals[eq_idx] / rate_tolerance;
@@ -558,13 +559,13 @@ getResidualMeasureValue(const WellState& well_state,
         }
     }
 
-    const double pressure_tolerance = tolerance_pressure_ms_wells;
+    const Scalar pressure_tolerance = tolerance_pressure_ms_wells;
     if (residuals[SPres] > pressure_tolerance) {
         sum += residuals[SPres] / pressure_tolerance;
         ++count;
     }
 
-    const double control_tolerance = getControlTolerance(well_state,
+    const Scalar control_tolerance = getControlTolerance(well_state,
                                                          tolerance_wells,
                                                          tolerance_pressure_ms_wells,
                                                          deferred_logger);
@@ -576,38 +577,43 @@ getResidualMeasureValue(const WellState& well_state,
     return sum;
 }
 
-#define INSTANCE(...) \
-template class MultisegmentWellEval<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,__VA_ARGS__>;
+template<class Scalar>
+using FS = BlackOilFluidSystem<Scalar,BlackOilDefaultIndexTraits>;
 
-// One phase
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>)
+#define INSTANTIATE(T,...) \
+    template class MultisegmentWellEval<FS<T>,__VA_ARGS__>;
 
-// Two phase
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,true,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<1u,0u,0u,0u,false,false,0u,0u,0u>)
+#define INSTANTIATE_TYPE(T)                                                  \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<1u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,true,false,0u,0u>)
 
-// Blackoil
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)
-INSTANCE(BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)
-INSTANCE(BlackOilIndices<1u,0u,0u,0u,true,false,0u,0u>)
+INSTANTIATE_TYPE(double)
+
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
 } // namespace Opm
