@@ -37,23 +37,26 @@
 #include <opm/simulators/wells/WellInterfaceIndices.hpp>
 #include <opm/simulators/wells/WellState.hpp>
 
+#include <fmt/format.h>
+
 #include <algorithm>
+#include <limits>
 
 namespace Opm {
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 resize(const int numSegments)
 {
     value_.resize(numSegments);
     evaluation_.resize(numSegments);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 init()
 {
-    for (size_t seg = 0; seg < value_.size(); ++seg) {
+    for (std::size_t seg = 0; seg < value_.size(); ++seg) {
         for (int eq_idx = 0; eq_idx < numWellEq; ++eq_idx) {
             evaluation_[seg][eq_idx] = 0.0;
             evaluation_[seg][eq_idx].setValue(value_[seg][eq_idx]);
@@ -62,9 +65,10 @@ init()
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
-update(const WellState& well_state, const bool stop_or_zero_rate_target)
+template<class FluidSystem, class Indices>
+void MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
+update(const WellState<Scalar>& well_state,
+       const bool stop_or_zero_rate_target)
 {
     static constexpr int Water = BlackoilPhases::Aqua;
     static constexpr int Gas = BlackoilPhases::Vapour;
@@ -85,9 +89,9 @@ update(const WellState& well_state, const bool stop_or_zero_rate_target)
     const auto& segment_pressure = segments.pressure;
     const PhaseUsage& pu = well_.phaseUsage();
 
-    for (size_t seg = 0; seg < value_.size(); ++seg) {
+    for (std::size_t seg = 0; seg < value_.size(); ++seg) {
         // calculate the total rate for each segment
-        double total_seg_rate = 0.0;
+        Scalar total_seg_rate = 0.0;
         // the segment pressure
         value_[seg][SPres] = segment_pressure[seg];
         // TODO: under what kind of circustances, the following will be wrong?
@@ -98,9 +102,9 @@ update(const WellState& well_state, const bool stop_or_zero_rate_target)
 
         if (seg == 0) {
             if (well_.isInjector()) {
-                total_seg_rate = std::max(total_seg_rate, 0.);
+                total_seg_rate = std::max(total_seg_rate, Scalar{0.});
             } else {
-                total_seg_rate = std::min(total_seg_rate, 0.);
+                total_seg_rate = std::min(total_seg_rate, Scalar{0.});
             }
         }
         value_[seg][WQTotal] = total_seg_rate;
@@ -150,26 +154,26 @@ update(const WellState& well_state, const bool stop_or_zero_rate_target)
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 updateNewton(const BVectorWell& dwells,
-             const double relaxation_factor,
-             const double dFLimit,
+             const Scalar relaxation_factor,
+             const Scalar dFLimit,
              const bool stop_or_zero_rate_target,
-             const double max_pressure_change)
+             const Scalar max_pressure_change)
 {
-    const std::vector<std::array<double, numWellEq>> old_primary_variables = value_;
+    const std::vector<std::array<Scalar, numWellEq>> old_primary_variables = value_;
 
-    for (size_t seg = 0; seg < value_.size(); ++seg) {
+    for (std::size_t seg = 0; seg < value_.size(); ++seg) {
         if (has_wfrac_variable) {
             const int sign = dwells[seg][WFrac] > 0. ? 1 : -1;
-            const double dx_limited = sign * std::min(std::abs(dwells[seg][WFrac]) * relaxation_factor, dFLimit);
+            const Scalar dx_limited = sign * std::min(std::abs(dwells[seg][WFrac]) * relaxation_factor, dFLimit);
             value_[seg][WFrac] = old_primary_variables[seg][WFrac] - dx_limited;
         }
 
         if (has_gfrac_variable) {
             const int sign = dwells[seg][GFrac] > 0. ? 1 : -1;
-            const double dx_limited = sign * std::min(std::abs(dwells[seg][GFrac]) * relaxation_factor, dFLimit);
+            const Scalar dx_limited = sign * std::min(std::abs(dwells[seg][GFrac]) * relaxation_factor, dFLimit);
             value_[seg][GFrac] = old_primary_variables[seg][GFrac] - dx_limited;
         }
 
@@ -179,8 +183,11 @@ updateNewton(const BVectorWell& dwells,
         // update the segment pressure
         {
             const int sign = dwells[seg][SPres] > 0.? 1 : -1;
-            const double dx_limited = sign * std::min(std::abs(dwells[seg][SPres]) * relaxation_factor, max_pressure_change);
-            value_[seg][SPres] = std::max(old_primary_variables[seg][SPres] - dx_limited, 1e5);
+            const Scalar dx_limited = sign * std::min(std::abs(dwells[seg][SPres]) * relaxation_factor, max_pressure_change);
+            // some cases might have defaulted bhp constraint of 1 bar, we use a slightly smaller value as the bhp lower limit for Newton update
+            // so that bhp constaint can be an active control when needed.
+            const Scalar lower_limit = (seg == 0) ? bhp_lower_limit : seg_pres_lower_limit;
+            value_[seg][SPres] = std::max(old_primary_variables[seg][SPres] - dx_limited, lower_limit);
         }
 
         // update the total rate // TODO: should we have a limitation of the total rate change?
@@ -190,9 +197,9 @@ updateNewton(const BVectorWell& dwells,
             // make sure that no injector produce and no producer inject
             if (seg == 0) {
                 if (well_.isInjector()) {
-                    value_[seg][WQTotal] = std::max(value_[seg][WQTotal], 0.0);
+                    value_[seg][WQTotal] = std::max(value_[seg][WQTotal], Scalar{0.0});
                 } else {
-                    value_[seg][WQTotal] = std::min(value_[seg][WQTotal], 0.0);
+                    value_[seg][WQTotal] = std::min(value_[seg][WQTotal], Scalar{0.0});
                 }
             }
         }
@@ -203,12 +210,12 @@ updateNewton(const BVectorWell& dwells,
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
-                const double rho,
-                const bool stop_or_zero_rate_target,
-                WellState& well_state,
+                const Scalar rho,
+                WellState<Scalar>& well_state,
+                const SummaryState& summary_state,
                 DeferredLogger& deferred_logger) const
 {
     static constexpr int Gas = BlackoilPhases::Vapour;
@@ -227,8 +234,8 @@ copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
     auto& disgas = segments.dissolved_gas_rate;
     auto& vapoil = segments.vaporized_oil_rate;
     auto& segment_pressure = segments.pressure;
-    for (size_t seg = 0; seg < value_.size(); ++seg) {
-        std::vector<double> fractions(well_.numPhases(), 0.0);
+    for (std::size_t seg = 0; seg < value_.size(); ++seg) {
+        std::vector<Scalar> fractions(well_.numPhases(), 0.0);
         fractions[oil_pos] = 1.0;
 
         if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
@@ -245,20 +252,20 @@ copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
 
         // convert the fractions to be Q_p / G_total to calculate the phase rates
         for (int p = 0; p < well_.numPhases(); ++p) {
-            const double scale = well_.scalingFactor(p);
+            const Scalar scale = well_.scalingFactor(p);
             // for injection wells, there should only one non-zero scaling factor
             if (scale > 0.) {
                 fractions[p] /= scale;
             } else {
-                // this should only happens to injection wells
+                // this should only happen to injection wells
                 fractions[p] = 0.;
             }
         }
 
         // calculate the phase rates based on the primary variables
-        const double g_total = value_[seg][WQTotal];
+        const Scalar g_total = value_[seg][WQTotal];
         for (int p = 0; p < well_.numPhases(); ++p) {
-            const double phase_rate = g_total * fractions[p];
+            const Scalar phase_rate = g_total * fractions[p];
             segment_rates[seg * well_.numPhases() + p] = phase_rate;
             if (seg == 0) { // top segment
                 ws.surface_rates[p] = phase_rate;
@@ -274,12 +281,12 @@ copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
 
         // Calculate other per-phase dynamic quantities.
 
-        const auto temperature = 0.0; // Ignore thermal effects
-        const auto saltConc = 0.0;    // Ignore salt precipitation
-        const auto Rvw = 0.0;         // Ignore vaporised water.
+        const Scalar temperature = 0.0; // Ignore thermal effects
+        const Scalar saltConc = 0.0;    // Ignore salt precipitation
+        const Scalar Rvw = 0.0;         // Ignore vaporised water.
 
-        auto rsMax = 0.0;
-        auto rvMax = 0.0;
+        Scalar rsMax = 0.0;
+        Scalar rvMax = 0.0;
         if (FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx)) {
             // Both oil and gas active.
             rsMax = FluidSystem::oilPvt()
@@ -314,10 +321,10 @@ copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
 
             well_.rateConverter().calcReservoirVoidageRates
                 (pvtReg, segment_pressure[seg],
-                 std::max(0.0, Rs),
-                 std::max(0.0, Rv),
-                 0.0, // Rsw
-                 0.0, // Rvw
+                 std::max(Scalar{0.0}, Rs),
+                 std::max(Scalar{0.0}, Rv),
+                 Scalar{0.0}, // Rsw
+                 Scalar{0.0}, // Rvw
                  temperature, saltConc, surf_rates, resv_rates);
         }
 
@@ -330,7 +337,7 @@ copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
         std::transform(segments.phase_resv_rates.begin() + (seg + 0) * well_.numPhases(),
                        segments.phase_resv_rates.begin() + (seg + 1) * well_.numPhases(),
                        segments.phase_holdup.begin()     + (seg + 0) * well_.numPhases(),
-                       [tot_resv](const auto qr) { return std::clamp(qr / tot_resv, 0.0, 1.0); });
+                       [tot_resv](const auto qr) -> Scalar { return std::clamp(qr / tot_resv, 0.0, 1.0); });
 
         // 4) Local condition flow velocities for segments other than top segment.
         if (seg > 0) {
@@ -353,7 +360,9 @@ copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
 
         if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
             segments.phase_viscosity[seg * well_.numPhases() + pu.phase_pos[Water]] =
-                FluidSystem::waterPvt().viscosity(pvtReg, temperature, segment_pressure[seg],  0.0 /*Rsw*/, saltConc);
+                FluidSystem::waterPvt().viscosity(pvtReg, temperature,
+                                                  segment_pressure[seg],
+                                                  Scalar{0.0} /*Rsw*/, saltConc);
         }
 
         if (FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx)) {
@@ -393,16 +402,18 @@ copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
                        [maxVel](const auto hf) { return (hf > 0.0) ? maxVel : 0.0; });
     }
 
+    // Note: for the ALQ value, in the StandardWell, WellInterfaceGeneric::getALQ(well_state) is used.
+    // We might want to unify the way regarding AQL value.
     WellBhpThpCalculator(well_)
-        .updateThp(rho, stop_or_zero_rate_target, [this]() { return well_.wellEcl().alq_value(); },
+        .updateThp(rho, [this, &summary_state]() { return well_.wellEcl().alq_value(summary_state); },
                    {FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx),
                     FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx),
                     FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx)},
-                   well_state, deferred_logger);
+                   well_state, summary_state, deferred_logger);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 processFractions(const int seg)
 {
     static constexpr int Water = BlackoilPhases::Aqua;
@@ -411,7 +422,7 @@ processFractions(const int seg)
 
     const PhaseUsage& pu = well_.phaseUsage();
 
-    std::vector<double> fractions(well_.numPhases(), 0.0);
+    std::vector<Scalar> fractions(well_.numPhases(), 0.0);
 
     assert(FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx));
     const int oil_pos = pu.phase_pos[Oil];
@@ -470,9 +481,9 @@ processFractions(const int seg)
     }
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<typename FluidSystem, typename Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 volumeFraction(const int seg,
                const unsigned compIdx) const
 {
@@ -496,19 +507,25 @@ volumeFraction(const int seg,
     /* if (has_solvent) {
         oil_fraction -= evaluation_[seg][SFrac];
     } */
+
+    // oil_fraction may turn out negative due to round-off, in that case
+    // set to zero (but keep derivatives)
+    if (oil_fraction.value() < 0.0) {
+        oil_fraction.setValue(0.0);
+    }
     return oil_fraction;
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 volumeFractionScaled(const int seg,
                      const int comp_idx) const
 {
     // For reservoir rate control, the distr in well control is used for the
     // rate conversion coefficients. For the injection well, only the distr of the injection
     // phase is not zero.
-    const double scale = well_.scalingFactor(well_.ebosCompIdxToFlowCompIdx(comp_idx));
+    const Scalar scale = well_.scalingFactor(well_.modelCompIdxToFlowCompIdx(comp_idx));
     if (scale > 0.) {
         return this->volumeFraction(seg, comp_idx) / scale;
     }
@@ -516,9 +533,9 @@ volumeFractionScaled(const int seg,
     return this->volumeFraction(seg, comp_idx);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 surfaceVolumeFraction(const int seg,
                       const int comp_idx) const
 {
@@ -532,12 +549,12 @@ surfaceVolumeFraction(const int seg,
     return this->volumeFractionScaled(seg, comp_idx) / sum_volume_fraction_scaled;
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 getSegmentRateUpwinding(const int seg,
                         const int seg_upwind,
-                        const size_t comp_idx) const
+                        const std::size_t comp_idx) const
 {
     // the result will contain the derivative with respect to WQTotal in segment seg,
     // and the derivatives with respect to WFrac GFrac in segment seg_upwind.
@@ -549,17 +566,17 @@ getSegmentRateUpwinding(const int seg,
         if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)
                 && Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx) == comp_idx
                 && phase == InjectorType::WATER)
-            return evaluation_[seg][WQTotal] / well_.scalingFactor(well_.ebosCompIdxToFlowCompIdx(comp_idx));
+            return evaluation_[seg][WQTotal] / well_.scalingFactor(well_.modelCompIdxToFlowCompIdx(comp_idx));
 
         if (FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx)
                 && Indices::canonicalToActiveComponentIndex(FluidSystem::oilCompIdx) == comp_idx
                 && phase == InjectorType::OIL)
-            return evaluation_[seg][WQTotal] / well_.scalingFactor(well_.ebosCompIdxToFlowCompIdx(comp_idx));
+            return evaluation_[seg][WQTotal] / well_.scalingFactor(well_.modelCompIdxToFlowCompIdx(comp_idx));
 
         if (FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx)
                 && Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx) == comp_idx
                 && phase == InjectorType::GAS)
-            return evaluation_[seg][WQTotal] / well_.scalingFactor(well_.ebosCompIdxToFlowCompIdx(comp_idx));
+            return evaluation_[seg][WQTotal] / well_.scalingFactor(well_.modelCompIdxToFlowCompIdx(comp_idx));
 
         return 0.0;
     }
@@ -572,75 +589,105 @@ getSegmentRateUpwinding(const int seg,
     return segment_rate;
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 getSegmentPressure(const int seg) const
 {
     return evaluation_[seg][SPres];
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 getBhp() const
 {
     return this->getSegmentPressure(0);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 getSegmentRate(const int seg,
                const int comp_idx) const
 {
     return evaluation_[seg][WQTotal] * this->volumeFractionScaled(seg, comp_idx);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 getQs(const int comp_idx) const
 {
     return this->getSegmentRate(0, comp_idx);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellPrimaryVariables<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellPrimaryVariables<FluidSystem,Indices>::EvalWell
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
 getWQTotal() const
 {
     return evaluation_[0][WQTotal];
 }
 
-#define INSTANCE(...) \
-template class MultisegmentWellPrimaryVariables<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,__VA_ARGS__,double>;
+template<class FluidSystem, class Indices>
+void
+MultisegmentWellPrimaryVariables<FluidSystem,Indices>::
+outputLowLimitPressureSegments(DeferredLogger& deferred_logger) const
+{
+    std::string msg = fmt::format("outputting the segments for well {} with pressures close to the lower limits "
+                                  "for debugging purpose \n", this->well_.name());
+    bool any_seg_pressure_close_to_limit = false;
+    for (std::size_t seg = 0; seg < value_.size(); ++seg) {
+        const double lower_limit = (seg == 0) ? bhp_lower_limit : seg_pres_lower_limit;
+        const double pres = Opm::getValue(this->getSegmentPressure(seg));
+        if (pres <= std::numeric_limits<double>::epsilon() + lower_limit) {
+            any_seg_pressure_close_to_limit = true;
+            fmt::format_to(std::back_inserter(msg), "seg {} : pressure {}\n", seg, pres / unit::barsa);
+        }
+    }
+    if (any_seg_pressure_close_to_limit) {
+        deferred_logger.debug(msg);
+    }
+}
 
-// One phase
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>)
+template<class Scalar>
+using FS = BlackOilFluidSystem<Scalar,BlackOilDefaultIndexTraits>;
 
-// Two phase
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>)
-// Blackoil
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)
+#define INSTANTIATE(T,...) \
+    template class MultisegmentWellPrimaryVariables<FS<T>,__VA_ARGS__>;
+
+#define INSTANTIATE_TYPE(T)                                                  \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<1u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,true,false,0u,0u>)
+
+INSTANTIATE_TYPE(double)
+
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
 }
