@@ -38,7 +38,7 @@ namespace Opm
 class DeferredLogger;
 template<class Scalar> class MultisegmentWellGeneric;
 template<class FluidSystem, class Indices> class WellInterfaceIndices;
-class WellState;
+template<class Scalar> class WellState;
 
 template<class FluidSystem, class Indices>
 class MultisegmentWellPrimaryVariables
@@ -76,7 +76,7 @@ public:
     static constexpr int numWellEq = Indices::numPhases + 1;
 
     using Scalar = typename FluidSystem::Scalar;
-    using EvalWell = DenseAd::Evaluation<double, /*size=*/Indices::numEq + numWellEq>;
+    using EvalWell = DenseAd::Evaluation<Scalar, /*size=*/Indices::numEq + numWellEq>;
 
     using Equations = MultisegmentWellEquations<Scalar,numWellEq,Indices::numEq>;
     using BVectorWell = typename Equations::BVectorWell;
@@ -92,20 +92,20 @@ public:
     void init();
 
     //! \brief Copy values from well state.
-    void update(const WellState& well_state, const bool stop_or_zero_rate_target);
+    void update(const WellState<Scalar>& well_state,
+                const bool stop_or_zero_rate_target);
 
     //! \brief Update values from newton update vector.
     void updateNewton(const BVectorWell& dwells,
-                      const double relaxation_factor,
-                      const double DFLimit,
+                      const Scalar relaxation_factor,
+                      const Scalar DFLimit,
                       const bool stop_or_zero_rate_target,
-                      const double max_pressure_change);
+                      const Scalar max_pressure_change);
 
     //! \brief Copy values to well state.
     void copyToWellState(const MultisegmentWellGeneric<Scalar>& mswell,
-                         const double rho,
-                         const bool stop_or_zero_rate_target,
-                         WellState& well_state,
+                         const Scalar rho,
+                         WellState<Scalar>& well_state,
                          const SummaryState& summary_state,
                          DeferredLogger& deferred_logger) const;
 
@@ -165,7 +165,7 @@ private:
 
     //! \brief The values for the primary variables
     //! \details Based on different solution strategies, the wells can have different primary variables
-    std::vector<std::array<double, numWellEq>> value_;
+    std::vector<std::array<Scalar, numWellEq>> value_;
 
     //! \brief The Evaluation for the well primary variables.
     //! \details Contains derivatives and are used in AD calculation
