@@ -23,6 +23,7 @@
 
 #include <opm/common/ErrorMacros.hpp>
 
+#include <opm/input/eclipse/Schedule/MSW/AICD.hpp>
 #include <opm/input/eclipse/Schedule/MSW/WellSegments.hpp>
 #include <opm/input/eclipse/Schedule/Well/WellConnections.hpp>
 
@@ -37,17 +38,29 @@
 
 #include <opm/simulators/utils/DeferredLogger.hpp>
 #include <opm/simulators/wells/MSWellHelpers.hpp>
+#include <opm/simulators/wells/SegmentState.hpp>
 #include <opm/simulators/wells/WellInterfaceGeneric.hpp>
 
+#include <opm/simulators/utils/BlackoilPhases.hpp>
+
 #include <fmt/format.h>
 
+#include <array>
+#include <cmath>
+#include <cstddef>
+#include <functional>
+#include <stdexcept>
+#include <string>
+#include <utility>
+#include <vector>
+
 namespace Opm
 {
 
-template<class FluidSystem, class Indices, class Scalar>
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+MultisegmentWellSegments<FluidSystem,Indices>::
 MultisegmentWellSegments(const int numSegments,
-                         WellInterfaceGeneric& well)
+                         WellInterfaceGeneric<Scalar>& well)
     : perforations_(numSegments)
     , perforation_depth_diffs_(well.numPerfs(), 0.0)
     , inlets_(well.wellEcl().getSegments().size())
@@ -74,7 +87,7 @@ MultisegmentWellSegments(const int numSegments,
     int i_perf_wells = 0;
     well.perfDepth().resize(well_.numPerfs(), 0.);
     const auto& segment_set = well_.wellEcl().getSegments();
-    for (size_t perf = 0; perf < completion_set.size(); ++perf) {
+    for (std::size_t perf = 0; perf < completion_set.size(); ++perf) {
         const Connection& connection = completion_set.get(perf);
         if (connection.state() == Connection::State::OPEN) {
             const int segment_index = segment_set.segmentNumberToIndex(connection.segment());
@@ -87,7 +100,7 @@ MultisegmentWellSegments(const int numSegments,
             }
             perforations_[segment_index].push_back(i_perf_wells);
             well.perfDepth()[i_perf_wells] = connection.depth();
-            const double segment_depth = segment_set[segment_index].depth();
+            const Scalar segment_depth = segment_set[segment_index].depth();
             perforation_depth_diffs_[i_perf_wells] = well_.perfDepth()[i_perf_wells] - segment_depth;
             i_perf_wells++;
         }
@@ -107,23 +120,23 @@ MultisegmentWellSegments(const int numSegments,
     // calculating the depth difference between the segment and its oulet_segments
     // for the top segment, we will make its zero unless we find other purpose to use this value
     for (int seg = 1; seg < numSegments; ++seg) {
-        const double segment_depth = segment_set[seg].depth();
+        const Scalar segment_depth = segment_set[seg].depth();
         const int outlet_segment_number = segment_set[seg].outletSegment();
         const Segment& outlet_segment = segment_set[segment_set.segmentNumberToIndex(outlet_segment_number)];
-        const double outlet_depth = outlet_segment.depth();
+        const Scalar outlet_depth = outlet_segment.depth();
         depth_diffs_[seg] = segment_depth - outlet_depth;
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellSegments<FluidSystem,Indices>::
 computeFluidProperties(const EvalWell& temperature,
                        const EvalWell& saltConcentration,
                        const PrimaryVariables& primary_variables,
                        int pvt_region_index,
                        DeferredLogger& deferred_logger)
 {
-    std::vector<double> surf_dens(well_.numComponents());
+    std::vector<Scalar> surf_dens(well_.numComponents());
     // Surface density.
     for (unsigned phaseIdx = 0; phaseIdx < FluidSystem::numPhases; ++phaseIdx) {
         if (!FluidSystem::phaseIsActive(phaseIdx)) {
@@ -134,7 +147,7 @@ computeFluidProperties(const EvalWell& temperature,
         surf_dens[compIdx] = FluidSystem::referenceDensity( phaseIdx, pvt_region_index);
     }
 
-    for (size_t seg = 0; seg < perforations_.size(); ++seg) {
+    for (std::size_t seg = 0; seg < perforations_.size(); ++seg) {
         // the compostion of the components inside wellbore under surface condition
         std::vector<EvalWell> mix_s(well_.numComponents(), 0.0);
         for (int comp_idx = 0; comp_idx < well_.numComponents(); ++comp_idx) {
@@ -248,7 +261,7 @@ computeFluidProperties(const EvalWell& temperature,
                                 "rv {} and pressure {}. "
                                 "Continue as if no dissolution (rs = 0) and "
                                 "vaporization (rv = 0) for this connection.",
-                                d, well_.name(), rv, seg_pressure);
+                                d, well_.name(), rs, rv, seg_pressure);
                 deferred_logger.debug(str);
             } else {
                 if (rs > 0.0) {
@@ -292,11 +305,11 @@ computeFluidProperties(const EvalWell& temperature,
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void MultisegmentWellSegments<FluidSystem,Indices>::
 updateUpwindingSegments(const PrimaryVariables& primary_variables)
 {
-    for (size_t seg = 0; seg < perforations_.size(); ++seg) {
+    for (std::size_t seg = 0; seg < perforations_.size(); ++seg) {
         // special treatment is needed for segment 0
         if (seg == 0) {
             // we are not supposed to have injecting producers and producing injectors
@@ -317,26 +330,27 @@ updateUpwindingSegments(const PrimaryVariables& primary_variables)
     }
 }
 
-
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellSegments<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
-getHydroPressureLoss(const int seg) const
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::EvalWell
+MultisegmentWellSegments<FluidSystem,Indices>::
+getHydroPressureLoss(const int seg,
+                     const int seg_density) const
 {
-    return densities_[seg] * well_.gravity() * depth_diffs_[seg];
+    return densities_[seg_density] * well_.gravity() * depth_diffs_[seg];
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-Scalar MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::Scalar
+MultisegmentWellSegments<FluidSystem,Indices>::
 getPressureDiffSegPerf(const int seg,
                        const int perf) const
 {
     return well_.gravity() * densities_[seg].value() * perforation_depth_diffs_[perf];
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellSegments<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::EvalWell
+MultisegmentWellSegments<FluidSystem,Indices>::
 getSurfaceVolume(const EvalWell& temperature,
                  const EvalWell& saltConcentration,
                  const PrimaryVariables& primary_variables,
@@ -476,44 +490,68 @@ getSurfaceVolume(const EvalWell& temperature,
     }
 
     // We increase the segment volume with a factor 10 to stabilize the system.
-    const double volume = well_.wellEcl().getSegments()[seg_idx].volume();
+    const Scalar volume = well_.wellEcl().getSegments()[seg_idx].volume();
 
     return volume / vol_ratio;
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellSegments<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
-getFrictionPressureLoss(const int seg) const
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::EvalWell
+MultisegmentWellSegments<FluidSystem,Indices>::
+getFrictionPressureLoss(const int seg, 
+                        const bool extra_reverse_flow_derivatives /*false*/) const
 {
-    const EvalWell mass_rate = mass_rates_[seg];
+    EvalWell mass_rate = mass_rates_[seg];
     const int seg_upwind = upwinding_segments_[seg];
     EvalWell density = densities_[seg_upwind];
     EvalWell visc = viscosities_[seg_upwind];
-    // WARNING
-    // We disregard the derivatives from the upwind density to make sure derivatives
-    // wrt. to different segments dont get mixed.
+    // In the reverse flow case, we don't have enough slots for all derivatives, e.g.,
+    // upwind pressure and flow. We amend this by a second function call optioin, where
+    // only these remaining derivatives are considered.
+    // For reference: the pressure equation assumes pressure/flow derivatives are given
+    // at segment node while fraction derivatives are given at upwind node.
+
     if (seg != seg_upwind) {
-        density.clearDerivatives();
-        visc.clearDerivatives();
+        if (!extra_reverse_flow_derivatives){
+            constexpr int WQTotal = Indices::numEq + PrimaryVariables::WQTotal;
+            constexpr int SPres = Indices::numEq + PrimaryVariables::SPres;
+            density.setDerivative(WQTotal, 0.0);
+            density.setDerivative(SPres, 0.0);
+            visc.setDerivative(WQTotal, 0.0);
+            visc.setDerivative(SPres, 0.0);
+        } else {
+            if (PrimaryVariables::has_water){
+                constexpr int WFrac = Indices::numEq + PrimaryVariables::WFrac;
+                density.setDerivative(WFrac, 0.0);
+                visc.setDerivative(WFrac, 0.0);
+            }
+            if (PrimaryVariables::has_gas){
+                constexpr int GFrac = Indices::numEq + PrimaryVariables::GFrac;
+                density.setDerivative(GFrac, 0.0);
+                visc.setDerivative(GFrac, 0.0);
+            }
+            mass_rate.clearDerivatives();
+        }
     }
+
     const auto& segment_set = well_.wellEcl().getSegments();
     const int outlet_segment_index = segment_set.segmentNumberToIndex(segment_set[seg].outletSegment());
-    const double length = segment_set[seg].totalLength() - segment_set[outlet_segment_index].totalLength();
+    const Scalar length = segment_set[seg].totalLength() - segment_set[outlet_segment_index].totalLength();
     assert(length > 0.);
-    const double roughness = segment_set[seg].roughness();
-    const double area = segment_set[seg].crossArea();
-    const double diameter = segment_set[seg].internalDiameter();
+    const Scalar roughness = segment_set[seg].roughness();
+    const Scalar area = segment_set[seg].crossArea();
+    const Scalar diameter = segment_set[seg].internalDiameter();
 
-    const double sign = mass_rate < 0. ? 1.0 : - 1.0;
+    const Scalar sign = mass_rate < 0. ? 1.0 : - 1.0;
 
     return sign * mswellhelpers::frictionPressureLoss(length, diameter, area, roughness, density, mass_rate, visc);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellSegments<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
-pressureDropSpiralICD(const int seg) const
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::EvalWell
+MultisegmentWellSegments<FluidSystem,Indices>::
+pressureDropSpiralICD(const int seg,
+                      const bool extra_reverse_flow_derivatives /*false*/) const
 {
     const auto& segment_set = well_.wellEcl().getSegments();
     const SICD& sicd = segment_set[seg].spiralICD();
@@ -547,54 +585,78 @@ pressureDropSpiralICD(const int seg) const
     }
 
     EvalWell density = densities_[seg_upwind];
-    // WARNING
-    // We disregard the derivatives from the upwind density to make sure derivatives
-    // wrt. to different segments dont get mixed.
+    EvalWell mass_rate = mass_rates_[seg];
+    // In the reverse flow case, we don't have enough slots for all derivatives, e.g.,
+    // upwind pressure and flow. We amend this by a second function call option, where 
+    // only these remaining derivatives are considered.
+    // For reference: the pressure equation assumes pressure/flow derivatives are given 
+    // at segment node while fraction derivatives are given at upwind node.   
     if (seg != seg_upwind) {
-        water_fraction.clearDerivatives();
-        water_viscosity.clearDerivatives();
-        oil_fraction.clearDerivatives();
-        oil_viscosity.clearDerivatives();
-        gas_fraction.clearDerivatives();
-        gas_viscosity.clearDerivatives();
-        density.clearDerivatives();
+        constexpr int nvar = FluidSystem::numPhases + 1;
+        std::vector<bool> zero_mask(nvar, false);
+        if (!extra_reverse_flow_derivatives){
+            zero_mask[PrimaryVariables::WQTotal] = true;
+            zero_mask[PrimaryVariables::SPres] = true;
+        } else {
+            if constexpr (PrimaryVariables::has_water){
+                zero_mask[PrimaryVariables::WFrac] = true;
+            }
+            if constexpr (PrimaryVariables::has_gas){
+                zero_mask[PrimaryVariables::GFrac] = true;
+            }
+            // mass_rate has no extra derivatives (they are organized as in equations)
+            mass_rate.clearDerivatives();
+        }
+        for (int ii = 0; ii < nvar; ++ii) {
+            if (zero_mask[ii]) {
+                water_fraction.setDerivative(Indices::numEq + ii, 0.0);
+                water_viscosity.setDerivative(Indices::numEq + ii, 0.0);
+                oil_fraction.setDerivative(Indices::numEq + ii, 0.0);
+                oil_viscosity.setDerivative(Indices::numEq + ii, 0.0);
+                gas_fraction.setDerivative(Indices::numEq + ii, 0.0);
+                gas_viscosity.setDerivative(Indices::numEq + ii, 0.0);
+                density.setDerivative(Indices::numEq + ii, 0.0);
+            }
+        }
     }
 
     const EvalWell liquid_fraction = water_fraction + oil_fraction;
 
     // viscosity contribution from the liquid
     const EvalWell liquid_viscosity_fraction = liquid_fraction < 1.e-30 ? oil_fraction * oil_viscosity + water_fraction * water_viscosity :
-            liquid_fraction * mswellhelpers::emulsionViscosity(water_fraction, water_viscosity, oil_fraction, oil_viscosity, sicd);
+            liquid_fraction * mswellhelpers::emulsionViscosity<EvalWell,Scalar>(water_fraction, water_viscosity, oil_fraction, oil_viscosity, sicd);
 
     const EvalWell mixture_viscosity = liquid_viscosity_fraction + gas_fraction * gas_viscosity;
 
-    const EvalWell reservoir_rate = mass_rates_[seg] / density;
+    const EvalWell reservoir_rate = mass_rate / density;
 
     const EvalWell reservoir_rate_icd = reservoir_rate * sicd.scalingFactor();
 
-    const double viscosity_cali = sicd.viscosityCalibration();
+    const Scalar viscosity_cali = sicd.viscosityCalibration();
 
     using MathTool = MathToolbox<EvalWell>;
 
-    const double density_cali = sicd.densityCalibration();
-    const EvalWell temp_value1 = MathTool::pow(density / density_cali, 0.75);
-    const EvalWell temp_value2 = MathTool::pow(mixture_viscosity / viscosity_cali, 0.25);
+    const Scalar density_cali = sicd.densityCalibration();
+    // make sure we don't pass negative base to powers
+    const EvalWell temp_value1 = density > 0.0 ? MathTool::pow(density / density_cali, 0.75) : 0.0;
+    const EvalWell temp_value2 = mixture_viscosity > 0.0 ? MathTool::pow(mixture_viscosity / viscosity_cali, 0.25) : 0.0;
 
     // formulation before 2016, base_strength is used
     // const double base_strength = sicd.strength() / density_cali;
     // formulation since 2016, strength is used instead
-    const double strength = sicd.strength();
+    const Scalar strength = sicd.strength();
 
-    const double sign = reservoir_rate_icd <= 0. ? 1.0 : -1.0;
+    const Scalar sign = reservoir_rate_icd <= 0. ? 1.0 : -1.0;
 
     return sign * temp_value1 * temp_value2 * strength * reservoir_rate_icd * reservoir_rate_icd;
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellSegments<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::EvalWell
+MultisegmentWellSegments<FluidSystem,Indices>::
 pressureDropAutoICD(const int seg,
-                    const UnitSystem& unit_system) const
+                    const UnitSystem& unit_system, 
+                    const bool extra_reverse_flow_derivatives /*false*/) const
 {
     const auto& segment_set = well_.wellEcl().getSegments();
     const AutoICD& aicd = segment_set[seg].autoICD();
@@ -635,154 +697,333 @@ pressureDropAutoICD(const int seg,
     }
 
     EvalWell density = densities_[seg_upwind];
-    // WARNING
-    // We disregard the derivatives from the upwind density to make sure derivatives
-    // wrt. to different segments dont get mixed.
+    EvalWell mass_rate = mass_rates_[seg];
+    // In the reverse flow case, we don't have enough slots for all derivatives, e.g.,
+    // upwind pressure and flow. We amend this by a second function call option, where 
+    // only these remaining derivatives are considered.
+    // For reference: the pressure equation assumes pressure/flow derivatives are given 
+    // at segment node while fraction derivatives are given at upwind node.   
     if (seg != seg_upwind) {
-        water_fraction.clearDerivatives();
-        water_viscosity.clearDerivatives();
-        water_density.clearDerivatives();
-        oil_fraction.clearDerivatives();
-        oil_viscosity.clearDerivatives();
-        oil_density.clearDerivatives();
-        gas_fraction.clearDerivatives();
-        gas_viscosity.clearDerivatives();
-        gas_density.clearDerivatives();
-        density.clearDerivatives();
+        constexpr int nvar = FluidSystem::numPhases + 1;
+        std::vector<bool> zero_mask(nvar, false);
+        if (!extra_reverse_flow_derivatives){
+            zero_mask[PrimaryVariables::WQTotal] = true;
+            zero_mask[PrimaryVariables::SPres] = true;
+        } else {
+            if (PrimaryVariables::has_water){
+                zero_mask[PrimaryVariables::WFrac] = true;
+            }
+            if (PrimaryVariables::has_gas){
+                zero_mask[PrimaryVariables::GFrac] = true;
+            }
+            // mass_rate has no extra derivatives (they are organized as in equations)
+            mass_rate.clearDerivatives();
+        }
+        for (int ii = 0; ii < nvar; ++ii) {
+            if (zero_mask[ii]) {
+                water_fraction.setDerivative(Indices::numEq + ii, 0.0);
+                water_viscosity.setDerivative(Indices::numEq + ii, 0.0);
+                water_density.setDerivative(Indices::numEq + ii, 0.0);
+                oil_fraction.setDerivative(Indices::numEq + ii, 0.0);
+                oil_viscosity.setDerivative(Indices::numEq + ii, 0.0);
+                oil_density.setDerivative(Indices::numEq + ii, 0.0);
+                gas_fraction.setDerivative(Indices::numEq + ii, 0.0);
+                gas_viscosity.setDerivative(Indices::numEq + ii, 0.0);
+                gas_density.setDerivative(Indices::numEq + ii, 0.0);
+                density.setDerivative(Indices::numEq + ii, 0.0);
+            }
+        }
     }
 
     using MathTool = MathToolbox<EvalWell>;
-    const EvalWell mixture_viscosity = MathTool::pow(water_fraction, aicd.waterViscExponent()) * water_viscosity
-                                     + MathTool::pow(oil_fraction, aicd.oilViscExponent()) * oil_viscosity
-                                     + MathTool::pow(gas_fraction, aicd.gasViscExponent()) * gas_viscosity;
+    // make sure we don't pass negative base to powers
+    auto safe_pow = [](const auto& a, const Scalar b) {
+        return a > 0.0 ? MathTool::pow(a,b) : 0.0;
+    };
 
-    const EvalWell mixture_density = MathTool::pow(water_fraction, aicd.waterDensityExponent()) * water_density
-                                   + MathTool::pow(oil_fraction, aicd.oilDensityExponent()) * oil_density
-                                   + MathTool::pow(gas_fraction, aicd.gasDensityExponent()) * gas_density;
+    const EvalWell mixture_viscosity = safe_pow(water_fraction, aicd.waterViscExponent()) * water_viscosity
+                                     + safe_pow(oil_fraction, aicd.oilViscExponent()) * oil_viscosity
+                                     + safe_pow(gas_fraction, aicd.gasViscExponent()) * gas_viscosity;
 
-    const double rho_reference = aicd.densityCalibration();
-    const double visc_reference = aicd.viscosityCalibration();
-    const auto volume_rate_icd = mass_rates_[seg] * aicd.scalingFactor() / mixture_density;
-    const double sign = volume_rate_icd <= 0. ? 1.0 : -1.0;
+    const EvalWell mixture_density = safe_pow(water_fraction, aicd.waterDensityExponent()) * water_density
+                                   + safe_pow(oil_fraction, aicd.oilDensityExponent()) * oil_density
+                                   + safe_pow(gas_fraction, aicd.gasDensityExponent()) * gas_density;
+
+    const Scalar rho_reference = aicd.densityCalibration();
+    const Scalar visc_reference = aicd.viscosityCalibration();
+    const auto volume_rate_icd = mass_rate * aicd.scalingFactor() / mixture_density;
+    const Scalar sign = volume_rate_icd <= 0. ? 1.0 : -1.0;
     // convert 1 unit volume rate
     using M  = UnitSystem::measure;
-    const double unit_volume_rate = unit_system.to_si(M::geometric_volume_rate, 1.);
+    const Scalar unit_volume_rate = unit_system.to_si(M::geometric_volume_rate, 1.);
 
     // TODO: we did not consider the maximum allowed rate here
     const auto result = sign / rho_reference * mixture_density * mixture_density
-                      * MathTool::pow(visc_reference/mixture_viscosity, aicd.viscExponent())
-                      * aicd.strength() * MathTool::pow( -sign * volume_rate_icd, aicd.flowRateExponent())
+                      * safe_pow(visc_reference/mixture_viscosity, aicd.viscExponent())
+                      * aicd.strength() * safe_pow( -sign * volume_rate_icd, aicd.flowRateExponent())
                       * std::pow(unit_volume_rate, (2. - aicd.flowRateExponent())) ;
     return result;
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellSegments<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
-pressureDropValve(const int seg) const
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::EvalWell
+MultisegmentWellSegments<FluidSystem,Indices>::
+pressureDropValve(const int seg,
+                  const SummaryState& summary_state,
+                  const bool extra_reverse_flow_derivatives /*false*/) const
 {
-    const auto& segment_set = well_.wellEcl().getSegments();
+    const Opm::WellSegments& segment_set = well_.wellEcl().getSegments();
     const Valve& valve = segment_set[seg].valve();
 
-    const EvalWell& mass_rate = mass_rates_[seg];
+    EvalWell mass_rate = mass_rates_[seg];
     const int seg_upwind = upwinding_segments_[seg];
     EvalWell visc = viscosities_[seg_upwind];
     EvalWell density = densities_[seg_upwind];
-    // WARNING
-    // We disregard the derivatives from the upwind density to make sure derivatives
-    // wrt. to different segments dont get mixed.
+    // In the reverse flow case, we don't have enough slots for all derivatives, e.g.,
+    // upwind pressure and flow. We amend this by a second function call optioin, where 
+    // only these remaining derivatives are considered.
+    // For reference: the pressure equation assumes pressure/flow derivatives are given 
+    // at segment node while fraction derivatives are given at upwind node. 
     if (seg != seg_upwind) {
-        visc.clearDerivatives();
-        density.clearDerivatives();
+        if (!extra_reverse_flow_derivatives){
+            constexpr int WQTotal = Indices::numEq + PrimaryVariables::WQTotal;
+            constexpr int SPres = Indices::numEq + PrimaryVariables::SPres;
+            density.setDerivative(WQTotal, 0.0);
+            density.setDerivative(SPres, 0.0);
+            visc.setDerivative(WQTotal, 0.0);
+            visc.setDerivative(SPres, 0.0);
+        } else {
+            if (PrimaryVariables::has_water){
+                constexpr int WFrac = Indices::numEq + PrimaryVariables::WFrac;
+                density.setDerivative(WFrac, 0.0);
+                visc.setDerivative(WFrac, 0.0);
+            }
+            if (PrimaryVariables::has_gas){
+                constexpr int GFrac = Indices::numEq + PrimaryVariables::GFrac;
+                density.setDerivative(GFrac, 0.0);
+                visc.setDerivative(GFrac, 0.0);
+            }
+            mass_rate.clearDerivatives();
+        }
     }
 
-    const double additional_length = valve.pipeAdditionalLength();
-    const double roughness = valve.pipeRoughness();
-    const double diameter = valve.pipeDiameter();
-    const double area = valve.pipeCrossArea();
+    const Scalar additional_length = valve.pipeAdditionalLength();
+    const Scalar roughness = valve.pipeRoughness();
+    const Scalar diameter = valve.pipeDiameter();
+    const Scalar area = valve.pipeCrossArea();
 
     const EvalWell friction_pressure_loss =
         mswellhelpers::frictionPressureLoss(additional_length, diameter, area, roughness, density, mass_rate, visc);
 
-    const double area_con = valve.conCrossArea();
-    const double cv = valve.conFlowCoefficient();
+    const ValveUDAEval uda_eval {summary_state, this->well_.name(), static_cast<std::size_t>(segment_set[seg].segmentNumber())};
+    const Scalar area_con = valve.conCrossArea(uda_eval);
+    const Scalar cv = valve.conFlowCoefficient();
 
     const EvalWell constriction_pressure_loss =
         mswellhelpers::valveContrictionPressureLoss(mass_rate, density, area_con, cv);
 
-    const double sign = mass_rate <= 0. ? 1.0 : -1.0;
+    const Scalar sign = mass_rate <= 0. ? 1.0 : -1.0;
     return sign * (friction_pressure_loss + constriction_pressure_loss);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-typename MultisegmentWellSegments<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellSegments<FluidSystem,Indices,Scalar>::
-accelerationPressureLoss(const int seg) const
+template<class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::EvalWell
+MultisegmentWellSegments<FluidSystem,Indices>::
+accelerationPressureLossContribution(const int seg,
+                                     const Scalar area,
+                                     const bool extra_reverse_flow_derivatives /*false*/) const
 {
-    const auto& segment_set = well_.wellEcl().getSegments();
-    const double area = segment_set[seg].crossArea();
-    const EvalWell mass_rate = mass_rates_[seg];
+    // Compute the *signed* velocity head for given segment (sign is positive for flow towards surface, i.e., negative rate) 
+    // Optionally return derivatives for reversed flow case
+    EvalWell mass_rate = mass_rates_[seg];
     const int seg_upwind = upwinding_segments_[seg];
-    EvalWell density = densities_[seg_upwind];
-    // WARNING
-    // We disregard the derivatives from the upwind density to make sure derivatives
-    // wrt. to different segments dont get mixed.
+    EvalWell density = densities_[seg_upwind];    
     if (seg != seg_upwind) {
-        density.clearDerivatives();
+        if (!extra_reverse_flow_derivatives){
+            constexpr int WQTotal = Indices::numEq + PrimaryVariables::WQTotal;
+            constexpr int SPres = Indices::numEq + PrimaryVariables::SPres;
+            density.setDerivative(WQTotal, 0.0);
+            density.setDerivative(SPres, 0.0);
+        } else {
+            if (PrimaryVariables::has_water){
+                constexpr int WFrac = Indices::numEq + PrimaryVariables::WFrac;
+                density.setDerivative(WFrac, 0.0);
+            }
+            if (PrimaryVariables::has_gas){
+                constexpr int GFrac = Indices::numEq + PrimaryVariables::GFrac;
+                density.setDerivative(GFrac, 0.0);
+            }
+            mass_rate.clearDerivatives();
+        }
     }
+    const Scalar sign = mass_rate > 0 ? -1.0 : 1.0;
+    return sign*mswellhelpers::velocityHead(area, mass_rate, density);
+}                                     
 
-    EvalWell accelerationPressureLoss = mswellhelpers::velocityHead(area, mass_rate, density);
-    // handling the velocity head of intlet segments
-    for (const int inlet : inlets_[seg]) {
-        const int seg_upwind_inlet = upwinding_segments_[inlet];
-        const double inlet_area = segment_set[inlet].crossArea();
-        EvalWell inlet_density = densities_[seg_upwind_inlet];
-        // WARNING
-        // We disregard the derivatives from the upwind density to make sure derivatives
-        // wrt. to different segments dont get mixed.
-        if (inlet != seg_upwind_inlet) {
-            inlet_density.clearDerivatives();
+template <class FluidSystem, class Indices>
+void
+MultisegmentWellSegments<FluidSystem,Indices>::
+copyPhaseDensities(const PhaseUsage& pu, SegmentState<Scalar>& segSol) const
+{
+    auto* rho = segSol.phase_density.data();
+
+    const auto phaseMap = std::vector {
+        std::pair { BlackoilPhases::Liquid, FluidSystem::oilPhaseIdx },
+        std::pair { BlackoilPhases::Vapour, FluidSystem::gasPhaseIdx },
+        std::pair { BlackoilPhases::Aqua  , FluidSystem::waterPhaseIdx },
+    };
+
+    // Densities stored in 'rho' as
+    // [{ p0, p1, ..., (np - 1), mixture, mixture_with_exponents },
+    //  { p0, p1, ..., (np - 1), mixture, mixture_with_exponents },
+    //  ...
+    //  { p0, p1, ..., (np - 1), mixture, mixture_with_exponents }]
+    // Stride is np + 2.
+    for (const auto& [boPhase, fsPhaseIdx] : phaseMap) {
+        if (pu.phase_used[boPhase]) {
+            this->copyPhaseDensities(fsPhaseIdx, pu.num_phases + 2,
+                                     rho + pu.phase_pos[boPhase]);
         }
-        const EvalWell inlet_mass_rate = mass_rates_[inlet];
-        accelerationPressureLoss -= mswellhelpers::velocityHead(std::max(inlet_area, area), inlet_mass_rate, inlet_density);
     }
 
-    // We change the sign of the accelerationPressureLoss for injectors.
-    // Is this correct? Testing indicates that this is what the reference simulator does
-    const double sign = mass_rate < 0. ? 1.0 : - 1.0;
-    accelerationPressureLoss *= sign;
+    // Mixture densities.
+    for (auto seg = 0*this->densities_.size(); seg < this->densities_.size(); ++seg) {
+        const auto mixOffset = seg*(pu.num_phases + 2) + pu.num_phases;
 
-    return accelerationPressureLoss;
+        rho[mixOffset + 0] = this->mixtureDensity(seg);
+        rho[mixOffset + 1] = this->mixtureDensityWithExponents(seg);
+    }
 }
 
-#define INSTANCE(...) \
-template class MultisegmentWellSegments<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,__VA_ARGS__,double>;
+template <class FluidSystem, class Indices>
+void
+MultisegmentWellSegments<FluidSystem,Indices>::
+copyPhaseDensities(const unsigned    phaseIdx,
+                   const std::size_t stride,
+                   Scalar*           dens) const
+{
+    const auto compIdx = Indices::canonicalToActiveComponentIndex
+        (FluidSystem::solventComponentIndex(phaseIdx));
 
-// One phase
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>)
+    for (const auto& phase_density : this->phase_densities_) {
+        *dens = phase_density[compIdx].value();
+        dens += stride;
+    }
+}
 
-// Two phase
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)
-INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>)
-// Blackoil
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)
-INSTANCE(BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)
-INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)
+template <class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::Scalar
+MultisegmentWellSegments<FluidSystem,Indices>::
+mixtureDensity(const int seg) const
+{
+    auto mixDens = 0.0;
+
+    const auto& rho = this->phase_densities_[seg];
+    const auto& q   = this->phase_fractions_[seg];
+
+    for (const auto& phIdx : {
+            FluidSystem::oilPhaseIdx,
+            FluidSystem::gasPhaseIdx,
+            FluidSystem::waterPhaseIdx
+        })
+    {
+        if (! FluidSystem::phaseIsActive(phIdx)) {
+            continue;
+        }
+
+        const auto compIdx = Indices::
+            canonicalToActiveComponentIndex(phIdx);
+
+        mixDens += q[compIdx].value() * rho[compIdx].value();
+    }
+
+    return mixDens;
+}
+
+template <class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::Scalar
+MultisegmentWellSegments<FluidSystem,Indices>::
+mixtureDensityWithExponents(const int seg) const
+{
+    if (const auto& segment = this->well_.wellEcl().getSegments()[seg];
+        segment.isAICD())
+    {
+        return this->mixtureDensityWithExponents(segment.autoICD(), seg);
+    }
+
+    // No other segment type includes exponents of flowing fractions when
+    // calculating the mixture/emulsion density.
+    return this->mixtureDensity(seg);
+}
+
+template <class FluidSystem, class Indices>
+typename MultisegmentWellSegments<FluidSystem,Indices>::Scalar
+MultisegmentWellSegments<FluidSystem,Indices>::
+mixtureDensityWithExponents(const AutoICD& aicd, const int seg) const
+{
+    auto mixDens = 0.0;
+
+    const auto& rho = this->phase_densities_[seg];
+    const auto& q   = this->phase_fractions_[seg];
+
+    constexpr auto densityExponents = std::array {
+        std::pair { FluidSystem::oilPhaseIdx  , &AutoICD::oilDensityExponent   },
+        std::pair { FluidSystem::gasPhaseIdx  , &AutoICD::gasDensityExponent   },
+        std::pair { FluidSystem::waterPhaseIdx, &AutoICD::waterDensityExponent },
+    };
+
+    for (const auto& [fsPhaseIdx, densityExponent] : densityExponents) {
+        if (FluidSystem::phaseIsActive(fsPhaseIdx)) {
+            const auto compIdx = Indices::
+                canonicalToActiveComponentIndex(fsPhaseIdx);
+
+            // exp = (aicd.*densityExponent)() in native syntax.
+            const auto exp = std::invoke(densityExponent, aicd);
+
+            mixDens += std::pow(q[compIdx].value(), exp) * rho[compIdx].value();
+        }
+    }
+
+    return mixDens;
+}
+
+template<class Scalar>
+using FS = BlackOilFluidSystem<Scalar,BlackOilDefaultIndexTraits>;
+
+#define INSTANTIATE(T,...) \
+    template class MultisegmentWellSegments<FS<T>,__VA_ARGS__>;
+
+#define INSTANTIATE_TYPE(T)                                                  \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,5u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,1u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,true,0u,0u,0u>)  \
+    INSTANTIATE(T,BlackOilTwoPhaseIndices<1u,0u,0u,0u,false,false,0u,0u,0u>) \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,true,2u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,1u,0u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)            \
+    INSTANTIATE(T,BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)             \
+    INSTANTIATE(T,BlackOilIndices<1u,0u,0u,0u,true,false,0u,0u>)
+
+INSTANTIATE_TYPE(double)
+
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
 } // namespace Opm
