@@ -29,25 +29,27 @@
 #include <opm/simulators/wells/SegmentState.hpp>
 #include <opm/simulators/wells/PerfData.hpp>
 #include <opm/simulators/wells/ParallelWellInfo.hpp>
-#include <opm/core/props/BlackoilPhases.hpp>
+
+#include <opm/simulators/utils/BlackoilPhases.hpp>
 
 namespace Opm {
 
-struct PerforationData;
+template<class Scalar> struct PerforationData;
 class SummaryState;
 class Well;
 
+template<class Scalar>
 class SingleWellState {
 public:
     SingleWellState(const std::string& name,
-                    const ParallelWellInfo& pinfo,
+                    const ParallelWellInfo<Scalar>& pinfo,
                     bool is_producer,
-                    double presssure_first_connection,
-                    const std::vector<PerforationData>& perf_input,
+                    Scalar presssure_first_connection,
+                    const std::vector<PerforationData<Scalar>>& perf_input,
                     const PhaseUsage& pu,
-                    double temp);
+                    Scalar temp);
 
-    static SingleWellState serializationTestObject(const ParallelWellInfo& pinfo);
+    static SingleWellState serializationTestObject(const ParallelWellInfo<Scalar>& pinfo);
 
     template<class Serializer>
     void serializeOp(Serializer& serializer)
@@ -78,19 +80,19 @@ public:
     bool operator==(const SingleWellState&) const;
 
     std::string name;
-    std::reference_wrapper<const ParallelWellInfo> parallel_info;
+    std::reference_wrapper<const ParallelWellInfo<Scalar>> parallel_info;
 
     WellStatus status{WellStatus::OPEN};
     bool producer;
     PhaseUsage pu;
-    double bhp{0};
-    double thp{0};
-    double temperature{0};
+    Scalar bhp{0};
+    Scalar thp{0};
+    Scalar temperature{0};
 
     // filtration injection concentration
-    double filtrate_conc{0};
+    Scalar filtrate_conc{0};
 
-    std::array<double,4> phase_mixing_rates{};
+    std::array<Scalar,4> phase_mixing_rates{};
     enum RateIndices {
       dissolved_gas = 0,
       dissolved_gas_in_water = 1,
@@ -98,16 +100,16 @@ public:
       vaporized_water = 3
     };
 
-    std::vector<double> well_potentials;
-    std::vector<double> productivity_index;
-    std::vector<double> implicit_ipr_a;
-    std::vector<double> implicit_ipr_b;    
-    std::vector<double> surface_rates;
-    std::vector<double> reservoir_rates;
-    std::vector<double> prev_surface_rates;
-    PerfData perf_data;
+    std::vector<Scalar> well_potentials;
+    std::vector<Scalar> productivity_index;
+    std::vector<Scalar> implicit_ipr_a;
+    std::vector<Scalar> implicit_ipr_b;
+    std::vector<Scalar> surface_rates;
+    std::vector<Scalar> reservoir_rates;
+    std::vector<Scalar> prev_surface_rates;
+    PerfData<Scalar> perf_data;
     bool trivial_target;
-    SegmentState segments;
+    SegmentState<Scalar> segments;
     Events events;
     WellInjectorCMode injection_cmode{WellInjectorCMode::CMODE_UNDEFINED};
     WellProducerCMode production_cmode{WellProducerCMode::CMODE_UNDEFINED};
@@ -119,7 +121,7 @@ public:
     /// \param[in] new_perf_data New perforation data.  Only
     ///    PerforationData::connection_transmissibility_factor actually
     ///    used (overwrites existing internal values).
-    void reset_connection_factors(const std::vector<PerforationData>& new_perf_data);
+    void reset_connection_factors(const std::vector<PerforationData<Scalar>>& new_perf_data);
     void update_producer_targets(const Well& ecl_well, const SummaryState& st);
     void update_injector_targets(const Well& ecl_well, const SummaryState& st);
     void update_targets(const Well& ecl_well, const SummaryState& st);
@@ -132,20 +134,17 @@ public:
     // The sum_xxx_rates() functions sum over all connection rates of pertinent
     // types. In the case of distributed wells this involves an MPI
     // communication.
-    double sum_solvent_rates() const;
-    double sum_polymer_rates() const;
-    double sum_brine_rates() const;
+    Scalar sum_solvent_rates() const;
+    Scalar sum_polymer_rates() const;
+    Scalar sum_brine_rates() const;
 
-    double sum_filtrate_rate() const;
-    double sum_filtrate_total() const;
+    Scalar sum_filtrate_rate() const;
+    Scalar sum_filtrate_total() const;
 
 private:
-    double sum_connection_rates(const std::vector<double>& connection_rates) const;
+    Scalar sum_connection_rates(const std::vector<Scalar>& connection_rates) const;
 };
 
-
 }
 
-
-
 #endif
