@@ -29,13 +29,13 @@ namespace Opm
 {
 
 class DeferredLogger;
-class GroupState;
+template<class Scalar> class GroupState;
 class Schedule;
 template<class Scalar, int numEq> class StandardWellEquations;
 template<class FluidSystem, class Indices> class StandardWellPrimaryVariables;
 class SummaryState;
 template<class FluidSystem> class WellInterfaceFluidSystem;
-class WellState;
+template<class Scalar> class WellState;
 
 //! \brief Class handling assemble of the equation system for StandardWell.
 template<class FluidSystem, class Indices>
@@ -52,15 +52,16 @@ public:
     {}
 
     //! \brief Assemble control equation.
-    void assembleControlEq(const WellState& well_state,
-                           const GroupState& group_state,
+    void assembleControlEq(const WellState<Scalar>& well_state,
+                           const GroupState<Scalar>& group_state,
                            const Schedule& schedule,
                            const SummaryState& summaryState,
                            const Well::InjectionControls& inj_controls,
                            const Well::ProductionControls& prod_controls,
                            const PrimaryVariables& primary_variables,
-                           const double rho,
+                           const Scalar rho,
                            StandardWellEquations<Scalar,Indices::numEq>& eqns,
+                           const bool stopped_or_zero_target,
                            DeferredLogger& deferred_logger) const;
 
     //! \brief Assemble injectivity equation.
