@@ -26,24 +26,20 @@
 #include <opm/simulators/utils/DeferredLoggingErrorHelpers.hpp>
 #include <opm/simulators/wells/GroupState.hpp>
 
-#include <algorithm>
 #include <cassert>
 #include <stdexcept>
-#include <type_traits>
 
-namespace Opm
-{
+namespace Opm::WGHelpers {
 
-namespace WellGroupHelpers
-{
-
-TargetCalculator::TargetCalculator(const Group::ProductionCMode cmode,
-                                   const PhaseUsage& pu,
-                                   const std::vector<double>& resv_coeff,
-                                   const double group_grat_target_from_sales,
-                                   const std::string& group_name,
-                                   const GroupState& group_state,
-                                   const bool use_gpmaint)
+template<class Scalar>
+TargetCalculator<Scalar>::
+TargetCalculator(const Group::ProductionCMode cmode,
+                 const PhaseUsage& pu,
+                 const std::vector<Scalar>& resv_coeff,
+                 const Scalar group_grat_target_from_sales,
+                 const std::string& group_name,
+                 const GroupState<Scalar>& group_state,
+                 const bool use_gpmaint)
     : cmode_(cmode)
     , pu_(pu)
     , resv_coeff_(resv_coeff)
@@ -55,8 +51,9 @@ TargetCalculator::TargetCalculator(const Group::ProductionCMode cmode,
 {
 }
 
+template<class Scalar>
 template <typename RateType>
-RateType TargetCalculator::calcModeRateFromRates(const RateType* rates) const
+RateType TargetCalculator<Scalar>::calcModeRateFromRates(const RateType* rates) const
 {
     switch (cmode_) {
     case Group::ProductionCMode::ORAT: {
@@ -95,7 +92,10 @@ RateType TargetCalculator::calcModeRateFromRates(const RateType* rates) const
     }
 }
 
-double TargetCalculator::groupTarget(const std::optional<Group::ProductionControls>& ctrl, Opm::DeferredLogger& deferred_logger) const
+template<class Scalar>
+Scalar TargetCalculator<Scalar>::
+groupTarget(const std::optional<Group::ProductionControls>& ctrl,
+            DeferredLogger& deferred_logger) const
 {
     if (!ctrl && !use_gpmaint_) {
         OPM_DEFLOG_THROW(std::logic_error,
@@ -111,7 +111,7 @@ double TargetCalculator::groupTarget(const std::optional<Group::ProductionContro
     case Group::ProductionCMode::GRAT:
     {
         // gas target may have been adjusted by GCONSALE
-        if ( group_grat_target_from_sales_ > 0)
+        if (group_grat_target_from_sales_ > 0)
             return group_grat_target_from_sales_;
 
         return ctrl->gas_target;
@@ -120,7 +120,7 @@ double TargetCalculator::groupTarget(const std::optional<Group::ProductionContro
         return ctrl->liquid_target;
     case Group::ProductionCMode::RESV:
     {
-        if(use_gpmaint_ && this->group_state_.has_gpmaint_target(this->group_name_))
+        if (use_gpmaint_ && this->group_state_.has_gpmaint_target(this->group_name_))
             return this->group_state_.gpmaint_target(this->group_name_);
 
         return ctrl->resv_target;
@@ -132,7 +132,9 @@ double TargetCalculator::groupTarget(const std::optional<Group::ProductionContro
     }
 }
 
-GuideRateModel::Target TargetCalculator::guideTargetMode() const
+template<class Scalar>
+GuideRateModel::Target
+TargetCalculator<Scalar>::guideTargetMode() const
 {
     switch (cmode_) {
     case Group::ProductionCMode::ORAT:
@@ -152,15 +154,17 @@ GuideRateModel::Target TargetCalculator::guideTargetMode() const
     }
 }
 
-InjectionTargetCalculator::InjectionTargetCalculator(const Group::InjectionCMode& cmode,
-                                                     const PhaseUsage& pu,
-                                                     const std::vector<double>& resv_coeff,
-                                                     const std::string& group_name,
-                                                     const double sales_target,
-                                                     const GroupState& group_state,
-                                                     const Phase& injection_phase,
-                                                     const bool use_gpmaint,
-                                                     DeferredLogger& deferred_logger)
+template<class Scalar>
+InjectionTargetCalculator<Scalar>::
+InjectionTargetCalculator(const Group::InjectionCMode& cmode,
+                          const PhaseUsage& pu,
+                          const std::vector<Scalar>& resv_coeff,
+                          const std::string& group_name,
+                          const Scalar sales_target,
+                          const GroupState<Scalar>& group_state,
+                          const Phase& injection_phase,
+                          const bool use_gpmaint,
+                          DeferredLogger& deferred_logger)
     : cmode_(cmode)
     , pu_(pu)
     , resv_coeff_(resv_coeff)
@@ -197,8 +201,10 @@ InjectionTargetCalculator::InjectionTargetCalculator(const Group::InjectionCMode
     }
 }
 
-
-double InjectionTargetCalculator::groupTarget(const std::optional<Group::InjectionControls>& ctrl, Opm::DeferredLogger& deferred_logger) const
+template<class Scalar>
+Scalar InjectionTargetCalculator<Scalar>::
+groupTarget(const std::optional<Group::InjectionControls>& ctrl,
+            DeferredLogger& deferred_logger) const
 {
     if (!ctrl && !use_gpmaint_) {
         OPM_DEFLOG_THROW(std::logic_error,
@@ -208,32 +214,32 @@ double InjectionTargetCalculator::groupTarget(const std::optional<Group::Injecti
     }
     switch (cmode_) {
     case Group::InjectionCMode::RATE:
-        if(use_gpmaint_ && this->group_state_.has_gpmaint_target(this->group_name_))
+        if (use_gpmaint_ && this->group_state_.has_gpmaint_target(this->group_name_))
             return this->group_state_.gpmaint_target(this->group_name_);
 
         return ctrl->surface_max_rate;
     case Group::InjectionCMode::RESV:
-        if(use_gpmaint_ && this->group_state_.has_gpmaint_target(this->group_name_))
+        if (use_gpmaint_ && this->group_state_.has_gpmaint_target(this->group_name_))
             return this->group_state_.gpmaint_target(this->group_name_) / resv_coeff_[pos_];
 
         return ctrl->resv_max_rate / resv_coeff_[pos_];
     case Group::InjectionCMode::REIN: {
-        double production_rate = this->group_state_.injection_rein_rates(ctrl->reinj_group)[pos_];
+        Scalar production_rate = this->group_state_.injection_rein_rates(ctrl->reinj_group)[pos_];
         return ctrl->target_reinj_fraction * production_rate;
     }
     case Group::InjectionCMode::VREP: {
-        const std::vector<double>& group_injection_reductions = this->group_state_.injection_reduction_rates(this->group_name_);
-        double voidage_rate = group_state_.injection_vrep_rate(ctrl->voidage_group) * ctrl->target_void_fraction;
-        double inj_reduction = 0.0;
+        const std::vector<Scalar>& group_injection_reductions = this->group_state_.injection_reduction_rates(this->group_name_);
+        Scalar voidage_rate = group_state_.injection_vrep_rate(ctrl->voidage_group) * ctrl->target_void_fraction;
+        Scalar inj_reduction = 0.0;
         if (ctrl->phase != Phase::WATER)
             inj_reduction += group_injection_reductions[pu_.phase_pos[BlackoilPhases::Aqua]]
-                    * resv_coeff_[pu_.phase_pos[BlackoilPhases::Aqua]];
+                           * resv_coeff_[pu_.phase_pos[BlackoilPhases::Aqua]];
         if (ctrl->phase != Phase::OIL)
             inj_reduction += group_injection_reductions[pu_.phase_pos[BlackoilPhases::Liquid]]
-                    * resv_coeff_[pu_.phase_pos[BlackoilPhases::Liquid]];
+                           * resv_coeff_[pu_.phase_pos[BlackoilPhases::Liquid]];
         if (ctrl->phase != Phase::GAS)
             inj_reduction += group_injection_reductions[pu_.phase_pos[BlackoilPhases::Vapour]]
-                    * resv_coeff_[pu_.phase_pos[BlackoilPhases::Vapour]];
+                           * resv_coeff_[pu_.phase_pos[BlackoilPhases::Vapour]];
         voidage_rate -= inj_reduction;
         return voidage_rate / resv_coeff_[pos_];
     }
@@ -241,7 +247,7 @@ double InjectionTargetCalculator::groupTarget(const std::optional<Group::Injecti
         assert(pos_ == pu_.phase_pos[BlackoilPhases::Vapour]);
         // Gas injection rate = Total gas production rate + gas import rate - gas consumption rate - sales rate;
         // Gas import and consumption is already included in the REIN rates
-        double inj_rate = group_state_.injection_rein_rates(this->group_name_)[pos_];
+        Scalar inj_rate = group_state_.injection_rein_rates(this->group_name_)[pos_];
         inj_rate -= sales_target_;
         return inj_rate;
     }
@@ -253,31 +259,42 @@ double InjectionTargetCalculator::groupTarget(const std::optional<Group::Injecti
     }
 }
 
-GuideRateModel::Target InjectionTargetCalculator::guideTargetMode() const
+template<class Scalar>
+GuideRateModel::Target
+InjectionTargetCalculator<Scalar>::guideTargetMode() const
 {
     return target_;
 }
 
-#define INSTANCE_TARGET_CALCULATOR(...) \
-template __VA_ARGS__ TargetCalculator::calcModeRateFromRates<__VA_ARGS__>(const __VA_ARGS__* rates) const;
+#define INSTANTIATE_TARGET_CALCULATOR(T,...) \
+    template __VA_ARGS__                     \
+    TargetCalculator<T>::calcModeRateFromRates(const __VA_ARGS__* rates) const;
 
-INSTANCE_TARGET_CALCULATOR(double)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,3,0>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,4,0>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,5,0>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,6,0>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,7,0>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,8,0>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,9,0>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,-1,4>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,-1,5>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,-1,6>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,-1,7>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,-1,8>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,-1,9>)
-INSTANCE_TARGET_CALCULATOR(DenseAd::Evaluation<double,-1,10>)
+#define INSTANTIATE_TYPE(T)                                       \
+    template class TargetCalculator<T>;                           \
+    template class InjectionTargetCalculator<T>;                  \
+    INSTANTIATE_TARGET_CALCULATOR(T,T)                            \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,3,0>)   \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,4,0>)   \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,5,0>)   \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,6,0>)   \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,7,0>)   \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,8,0>)   \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,9,0>)   \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,10,0>)  \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,4>)  \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,5>)  \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,6>)  \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,7>)  \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,8>)  \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,9>)  \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,10>) \
+    INSTANTIATE_TARGET_CALCULATOR(T,DenseAd::Evaluation<T,-1,11>)
 
+INSTANTIATE_TYPE(double)
 
-} // namespace WellGroupHelpers
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
-} // namespace Opm
+} // namespace Opm::WGHelpers
