@@ -30,6 +30,7 @@
 
 #include <opm/material/densead/Evaluation.hpp>
 
+#include <opm/simulators/wells/FractionCalculator.hpp>
 #include <opm/simulators/wells/GroupState.hpp>
 #include <opm/simulators/wells/TargetCalculator.hpp>
 #include <opm/simulators/wells/WellGroupHelpers.hpp>
@@ -37,23 +38,24 @@
 #include <opm/simulators/wells/WellState.hpp>
 
 #include <algorithm>
+#include <cstddef>
 #include <cassert>
 
-namespace Opm
-{
+namespace Opm {
 
+template<class Scalar>
 template<class EvalWell>
-void WellGroupControls::
+void WellGroupControls<Scalar>::
 getGroupInjectionControl(const Group& group,
-                         const WellState& well_state,
-                         const GroupState& group_state,
+                         const WellState<Scalar>& well_state,
+                         const GroupState<Scalar>& group_state,
                          const Schedule& schedule,
                          const SummaryState& summaryState,
                          const InjectorType& injectorType,
                          const EvalWell& bhp,
                          const EvalWell& injection_rate,
                          const RateConvFunc& rateConverter,
-                         double efficiencyFactor,
+                         Scalar efficiencyFactor,
                          EvalWell& control_eq,
                          DeferredLogger& deferred_logger) const
 {
@@ -126,32 +128,42 @@ getGroupInjectionControl(const Group& group,
     // This is the group containing the control we will check against.
 
     // Make conversion factors for RESV <-> surface rates.
-    std::vector<double> resv_coeff(pu.num_phases, 1.0);
+    std::vector<Scalar> resv_coeff(pu.num_phases, 1.0);
     rateConverter(0, well_.pvtRegionIdx(), std::nullopt, resv_coeff); // FIPNUM region 0 here, should use FIPNUM from WELSPECS.
 
-    double sales_target = 0;
+    Scalar sales_target = 0;
     if (schedule[well_.currentStep()].gconsale().has(group.name())) {
         const auto& gconsale = schedule[well_.currentStep()].gconsale().get(group.name(), summaryState);
         sales_target = gconsale.sales_target;
     }
-    WellGroupHelpers::InjectionTargetCalculator tcalc(currentGroupControl, pu,
-                                                      resv_coeff, group.name(),
-                                                      sales_target, group_state,
-                                                      injectionPhase,
-                                                      group.has_gpmaint_control(injectionPhase, currentGroupControl),
-                                                      deferred_logger);
-    WellGroupHelpers::FractionCalculator fcalc(schedule, well_state,
-                                               group_state, well_.currentStep(),
-                                               well_.guideRate(),
-                                               tcalc.guideTargetMode(),
-                                               pu, false, injectionPhase);
+
+    WGHelpers::InjectionTargetCalculator tcalc(currentGroupControl,
+                                               pu,
+                                               resv_coeff,
+                                               group.name(),
+                                               sales_target,
+                                               group_state,
+                                               injectionPhase,
+                                               group.has_gpmaint_control(injectionPhase,
+                                                                         currentGroupControl),
+                                               deferred_logger);
+
+    WGHelpers::FractionCalculator fcalc(schedule,
+                                        well_state,
+                                        group_state,
+                                        well_.currentStep(),
+                                        well_.guideRate(),
+                                        tcalc.guideTargetMode(),
+                                        pu,
+                                        false,
+                                        injectionPhase);
 
     auto localFraction = [&](const std::string& child) {
         return fcalc.localFraction(child, child);
     };
 
     auto localReduction = [&](const std::string& group_name) {
-        const std::vector<double>& groupTargetReductions = group_state.injection_reduction_rates(group_name);
+        const std::vector<Scalar>& groupTargetReductions = group_state.injection_reduction_rates(group_name);
         return tcalc.calcModeRateFromRates(groupTargetReductions);
     };
 
@@ -159,14 +171,13 @@ getGroupInjectionControl(const Group& group,
     if (!group.has_gpmaint_control(injectionPhase, currentGroupControl))
         ctrl = group.injectionControls(injectionPhase, summaryState);
 
-    const double orig_target = tcalc.groupTarget(ctrl,
-                                                deferred_logger);
-    const auto chain = WellGroupHelpers::groupChainTopBot(well_.name(), group.name(),
-                                                          schedule, well_.currentStep());
+    const Scalar orig_target = tcalc.groupTarget(ctrl, deferred_logger);
+    const auto chain = WellGroupHelpers<Scalar>::groupChainTopBot(well_.name(), group.name(),
+                                                                  schedule, well_.currentStep());
     // Because 'name' is the last of the elements, and not an ancestor, we subtract one below.
-    const size_t num_ancestors = chain.size() - 1;
-    double target = orig_target;
-    for (size_t ii = 0; ii < num_ancestors; ++ii) {
+    const std::size_t num_ancestors = chain.size() - 1;
+    Scalar target = orig_target;
+    for (std::size_t ii = 0; ii < num_ancestors; ++ii) {
         if ((ii == 0) || well_.guideRate()->has(chain[ii], injectionPhase)) {
             // Apply local reductions only at the control level
             // (top) and for levels where we have a specified
@@ -176,22 +187,23 @@ getGroupInjectionControl(const Group& group,
         target *= localFraction(chain[ii+1]);
     }
     // Avoid negative target rates coming from too large local reductions.
-    const double target_rate = std::max(0.0, target / efficiencyFactor);
+    const Scalar target_rate = std::max(Scalar(0.0), target / efficiencyFactor);
     const auto current_rate = injection_rate;
 
     control_eq = current_rate - target_rate;
 }
 
-std::optional<double>
-WellGroupControls::
+template<class Scalar>
+std::optional<Scalar>
+WellGroupControls<Scalar>::
 getGroupInjectionTargetRate(const Group& group,
-                            const WellState& well_state,
-                            const GroupState& group_state,
+                            const WellState<Scalar>& well_state,
+                            const GroupState<Scalar>& group_state,
                             const Schedule& schedule,
                             const SummaryState& summaryState,
                             const InjectorType& injectorType,
                             const RateConvFunc& rateConverter,
-                            double efficiencyFactor,
+                            Scalar efficiencyFactor,
                             DeferredLogger& deferred_logger) const
 {
     // Setting some defaults to silence warnings below.
@@ -250,29 +262,41 @@ getGroupInjectionTargetRate(const Group& group,
     // This is the group containing the control we will check against.
 
     // Make conversion factors for RESV <-> surface rates.
-    std::vector<double> resv_coeff(pu.num_phases, 1.0);
+    std::vector<Scalar> resv_coeff(pu.num_phases, 1.0);
     rateConverter(0, well_.pvtRegionIdx(), std::nullopt, resv_coeff); // FIPNUM region 0 here, should use FIPNUM from WELSPECS.
 
-    double sales_target = 0;
+    Scalar sales_target = 0;
     if (schedule[well_.currentStep()].gconsale().has(group.name())) {
         const auto& gconsale = schedule[well_.currentStep()].gconsale().get(group.name(), summaryState);
         sales_target = gconsale.sales_target;
     }
-    WellGroupHelpers::InjectionTargetCalculator tcalc(currentGroupControl, pu, resv_coeff,
-                                                      group.name(), sales_target, group_state,
-                                                      injectionPhase,
-                                                      group.has_gpmaint_control(injectionPhase, currentGroupControl),
-                                                      deferred_logger);
-    WellGroupHelpers::FractionCalculator fcalc(schedule, well_state, group_state,
-                                               well_.currentStep(), well_.guideRate(),
-                                               tcalc.guideTargetMode(), pu, false, injectionPhase);
+    WGHelpers::InjectionTargetCalculator tcalc(currentGroupControl,
+                                               pu,
+                                               resv_coeff,
+                                               group.name(),
+                                               sales_target,
+                                               group_state,
+                                               injectionPhase,
+                                               group.has_gpmaint_control(injectionPhase,
+                                                                         currentGroupControl),
+                                               deferred_logger);
+
+    WGHelpers::FractionCalculator fcalc(schedule,
+                                        well_state,
+                                        group_state,
+                                        well_.currentStep(),
+                                        well_.guideRate(),
+                                        tcalc.guideTargetMode(),
+                                        pu,
+                                        false,
+                                        injectionPhase);
 
     auto localFraction = [&](const std::string& child) {
         return fcalc.localFraction(child, child); //Note child needs to be passed to always include since the global isGrup map is not updated yet.
     };
 
     auto localReduction = [&](const std::string& group_name) {
-        const std::vector<double>& groupTargetReductions = group_state.injection_reduction_rates(group_name);
+        const std::vector<Scalar>& groupTargetReductions = group_state.injection_reduction_rates(group_name);
         return tcalc.calcModeRateFromRates(groupTargetReductions);
     };
 
@@ -280,13 +304,16 @@ getGroupInjectionTargetRate(const Group& group,
     if (!group.has_gpmaint_control(injectionPhase, currentGroupControl))
         ctrl = group.injectionControls(injectionPhase, summaryState);
 
-    const double orig_target = tcalc.groupTarget(ctrl, deferred_logger);
+    const Scalar orig_target = tcalc.groupTarget(ctrl, deferred_logger);
 
-    const auto chain = WellGroupHelpers::groupChainTopBot(well_.name(), group.name(), schedule, well_.currentStep());
+    const auto chain = WellGroupHelpers<Scalar>::groupChainTopBot(well_.name(),
+                                                                  group.name(),
+                                                                  schedule,
+                                                                  well_.currentStep());
     // Because 'name' is the last of the elements, and not an ancestor, we subtract one below.
-    const size_t num_ancestors = chain.size() - 1;
-    double target = orig_target;
-    for (size_t ii = 0; ii < num_ancestors; ++ii) {
+    const std::size_t num_ancestors = chain.size() - 1;
+    Scalar target = orig_target;
+    for (std::size_t ii = 0; ii < num_ancestors; ++ii) {
         if ((ii == 0) || well_.guideRate()->has(chain[ii], injectionPhase)) {
             // Apply local reductions only at the control level
             // (top) and for levels where we have a specified
@@ -295,21 +322,23 @@ getGroupInjectionTargetRate(const Group& group,
         }
         target *= localFraction(chain[ii+1]);
     }
-    return std::max(0.0, target / efficiencyFactor);
+    return std::max(Scalar(0.0), target / efficiencyFactor);
 }
 
+template<class Scalar>
 template<class EvalWell>
-void WellGroupControls::getGroupProductionControl(const Group& group,
-                                                  const WellState& well_state,
-                                                  const GroupState& group_state,
-                                                  const Schedule& schedule,
-                                                  const SummaryState& summaryState,
-                                                  const EvalWell& bhp,
-                                                  const std::vector<EvalWell>& rates,
-                                                  const RateConvFunc& rateConverter,
-                                                  double efficiencyFactor,
-                                                  EvalWell& control_eq,
-                                                  DeferredLogger& deferred_logger) const
+void WellGroupControls<Scalar>::
+getGroupProductionControl(const Group& group,
+                          const WellState<Scalar>& well_state,
+                          const GroupState<Scalar>& group_state,
+                          const Schedule& schedule,
+                          const SummaryState& summaryState,
+                          const EvalWell& bhp,
+                          const std::vector<EvalWell>& rates,
+                          const RateConvFunc& rateConverter,
+                          Scalar efficiencyFactor,
+                          EvalWell& control_eq,
+                          DeferredLogger& deferred_logger) const
 {
     const Group::ProductionCMode& currentGroupControl = group_state.production_control(group.name());
     if (currentGroupControl == Group::ProductionCMode::FLD ||
@@ -352,31 +381,39 @@ void WellGroupControls::getGroupProductionControl(const Group& group,
     // This is the group containing the control we will check against.
 
     // Make conversion factors for RESV <-> surface rates.
-    std::vector<double> resv_coeff(well_.phaseUsage().num_phases, 1.0);
+    std::vector<Scalar> resv_coeff(well_.phaseUsage().num_phases, 1.0);
     rateConverter(0, well_.pvtRegionIdx(), group.name(), resv_coeff); // FIPNUM region 0 here, should use FIPNUM from WELSPECS.
 
     // gconsale may adjust the grat target.
     // the adjusted rates is send to the targetCalculator
-    double gratTargetFromSales = 0.0;
+    Scalar gratTargetFromSales = 0.0;
     if (group_state.has_grat_sales_target(group.name()))
         gratTargetFromSales = group_state.grat_sales_target(group.name());
 
-    WellGroupHelpers::TargetCalculator tcalc(currentGroupControl, pu, resv_coeff,
-                                             gratTargetFromSales, group.name(),
-                                             group_state,
-                                             group.has_gpmaint_control(currentGroupControl));
-    WellGroupHelpers::FractionCalculator fcalc(schedule, well_state, group_state,
-                                               well_.currentStep(),
-                                               well_.guideRate(),
-                                               tcalc.guideTargetMode(),
-                                               pu, true, Phase::OIL);
+    WGHelpers::TargetCalculator tcalc(currentGroupControl,
+                                      pu,
+                                      resv_coeff,
+                                      gratTargetFromSales,
+                                      group.name(),
+                                      group_state,
+                                      group.has_gpmaint_control(currentGroupControl));
+
+    WGHelpers::FractionCalculator fcalc(schedule,
+                                        well_state,
+                                        group_state,
+                                        well_.currentStep(),
+                                        well_.guideRate(),
+                                        tcalc.guideTargetMode(),
+                                        pu,
+                                        true,
+                                        Phase::OIL);
 
     auto localFraction = [&](const std::string& child) {
         return fcalc.localFraction(child, child);
     };
 
     auto localReduction = [&](const std::string& group_name) {
-        const std::vector<double>& groupTargetReductions = group_state.production_reduction_rates(group_name);
+        const std::vector<Scalar>& groupTargetReductions = group_state.production_reduction_rates(group_name);
         return tcalc.calcModeRateFromRates(groupTargetReductions);
     };
 
@@ -384,13 +421,13 @@ void WellGroupControls::getGroupProductionControl(const Group& group,
     if (!group.has_gpmaint_control(currentGroupControl))
         ctrl = group.productionControls(summaryState);
 
-    const double orig_target = tcalc.groupTarget(ctrl, deferred_logger);
-    const auto chain = WellGroupHelpers::groupChainTopBot(well_.name(), group.name(),
-                                                          schedule, well_.currentStep());
+    const Scalar orig_target = tcalc.groupTarget(ctrl, deferred_logger);
+    const auto chain = WellGroupHelpers<Scalar>::groupChainTopBot(well_.name(), group.name(),
+                                                                  schedule, well_.currentStep());
     // Because 'name' is the last of the elements, and not an ancestor, we subtract one below.
-    const size_t num_ancestors = chain.size() - 1;
-    double target = orig_target;
-    for (size_t ii = 0; ii < num_ancestors; ++ii) {
+    const std::size_t num_ancestors = chain.size() - 1;
+    Scalar target = orig_target;
+    for (std::size_t ii = 0; ii < num_ancestors; ++ii) {
         if ((ii == 0) || well_.guideRate()->has(chain[ii])) {
             // Apply local reductions only at the control level
             // (top) and for levels where we have a specified
@@ -400,19 +437,20 @@ void WellGroupControls::getGroupProductionControl(const Group& group,
         target *= localFraction(chain[ii+1]);
     }
     // Avoid negative target rates coming from too large local reductions.
-    const double target_rate = std::max(0.0, target / efficiencyFactor);
+    const Scalar target_rate = std::max(Scalar(0.0), target / efficiencyFactor);
     const auto current_rate = -tcalc.calcModeRateFromRates(rates); // Switch sign since 'rates' are negative for producers.
     control_eq = current_rate - target_rate;
 }
 
-double WellGroupControls::
+template<class Scalar>
+Scalar WellGroupControls<Scalar>::
 getGroupProductionTargetRate(const Group& group,
-                             const WellState& well_state,
-                             const GroupState& group_state,
+                             const WellState<Scalar>& well_state,
+                             const GroupState<Scalar>& group_state,
                              const Schedule& schedule,
                              const SummaryState& summaryState,
                              const RateConvFunc& rateConverter,
-                             double efficiencyFactor,
+                             Scalar efficiencyFactor,
                              DeferredLogger& deferred_logger) const
 {
     const Group::ProductionCMode& currentGroupControl = group_state.production_control(group.name());
@@ -441,28 +479,39 @@ getGroupProductionTargetRate(const Group& group,
     // This is the group containing the control we will check against.
 
     // Make conversion factors for RESV <-> surface rates.
-    std::vector<double> resv_coeff(well_.phaseUsage().num_phases, 1.0);
+    std::vector<Scalar> resv_coeff(well_.phaseUsage().num_phases, 1.0);
     rateConverter(0, well_.pvtRegionIdx(), group.name(), resv_coeff); // FIPNUM region 0 here, should use FIPNUM from WELSPECS.
 
     // gconsale may adjust the grat target.
     // the adjusted rates is send to the targetCalculator
-    double gratTargetFromSales = 0.0;
+    Scalar gratTargetFromSales = 0.0;
     if (group_state.has_grat_sales_target(group.name()))
         gratTargetFromSales = group_state.grat_sales_target(group.name());
 
-    WellGroupHelpers::TargetCalculator tcalc(currentGroupControl, pu, resv_coeff, gratTargetFromSales, group.name(), group_state, group.has_gpmaint_control(currentGroupControl));
-    WellGroupHelpers::FractionCalculator fcalc(schedule, well_state, group_state,
-                                               well_.currentStep(),
-                                               well_.guideRate(),
-                                               tcalc.guideTargetMode(),
-                                               pu, true, Phase::OIL);
+    WGHelpers::TargetCalculator tcalc(currentGroupControl,
+                                      pu,
+                                      resv_coeff,
+                                      gratTargetFromSales,
+                                      group.name(),
+                                      group_state,
+                                      group.has_gpmaint_control(currentGroupControl));
+
+    WGHelpers::FractionCalculator fcalc(schedule,
+                                        well_state,
+                                        group_state,
+                                        well_.currentStep(),
+                                        well_.guideRate(),
+                                        tcalc.guideTargetMode(),
+                                        pu,
+                                        true,
+                                        Phase::OIL);
 
     auto localFraction = [&](const std::string& child) {
         return fcalc.localFraction(child, child); //Note child needs to be passed to always include since the global isGrup map is not updated yet.
     };
 
     auto localReduction = [&](const std::string& group_name) {
-        const std::vector<double>& groupTargetReductions = group_state.production_reduction_rates(group_name);
+        const std::vector<Scalar>& groupTargetReductions = group_state.production_reduction_rates(group_name);
         return tcalc.calcModeRateFromRates(groupTargetReductions);
     };
 
@@ -470,13 +519,13 @@ getGroupProductionTargetRate(const Group& group,
     if (!group.has_gpmaint_control(currentGroupControl))
         ctrl = group.productionControls(summaryState);
 
-    const double orig_target = tcalc.groupTarget(ctrl, deferred_logger);
-    const auto chain = WellGroupHelpers::groupChainTopBot(well_.name(), group.name(),
-                                                          schedule, well_.currentStep());
+    const Scalar orig_target = tcalc.groupTarget(ctrl, deferred_logger);
+    const auto chain = WellGroupHelpers<Scalar>::groupChainTopBot(well_.name(), group.name(),
+                                                                  schedule, well_.currentStep());
     // Because 'name' is the last of the elements, and not an ancestor, we subtract one below.
-    const size_t num_ancestors = chain.size() - 1;
-    double target = orig_target;
-    for (size_t ii = 0; ii < num_ancestors; ++ii) {
+    const std::size_t num_ancestors = chain.size() - 1;
+    Scalar target = orig_target;
+    for (std::size_t ii = 0; ii < num_ancestors; ++ii) {
         if ((ii == 0) || well_.guideRate()->has(chain[ii])) {
             // Apply local reductions only at the control level
             // (top) and for levels where we have a specified
@@ -486,60 +535,70 @@ getGroupProductionTargetRate(const Group& group,
         target *= localFraction(chain[ii+1]);
     }
     // Avoid negative target rates coming from too large local reductions.
-    const double target_rate = std::max(0.0, target / efficiencyFactor);
+    const Scalar target_rate = std::max(Scalar(0.0), target / efficiencyFactor);
     const auto& ws = well_state.well(well_.indexOfWell());
     const auto& rates = ws.surface_rates;
     const auto current_rate = -tcalc.calcModeRateFromRates(rates); // Switch sign since 'rates' are negative for producers.
-    double scale = 1.0;
+    Scalar scale = 1.0;
     if (target_rate == 0.0) {
         return 0.0;
     }
 
     if (current_rate > 1e-14)
-        scale = target_rate/current_rate;
+        scale = target_rate / current_rate;
     return scale;
 }
 
-#define INSTANCE(...) \
-template void WellGroupControls:: \
-getGroupInjectionControl<__VA_ARGS__>(const Group&, \
-                                      const WellState&, \
-                                      const GroupState&, \
-                                      const Schedule&, \
-                                      const SummaryState&, \
-                                      const InjectorType&, \
-                                      const __VA_ARGS__& bhp, \
-                                      const __VA_ARGS__& injection_rate, \
-                                      const RateConvFunc& rateConverter, \
-                                      double efficiencyFactor, \
-                                      __VA_ARGS__& control_eq, \
-                                      DeferredLogger& deferred_logger) const; \
-template void WellGroupControls:: \
-getGroupProductionControl<__VA_ARGS__>(const Group&, \
-                                       const WellState&, \
-                                       const GroupState&, \
-                                       const Schedule&, \
-                                       const SummaryState&, \
-                                       const __VA_ARGS__& bhp, \
-                                       const std::vector<__VA_ARGS__>&, \
-                                       const RateConvFunc& rateConverter, \
-                                       double efficiencyFactor, \
-                                       __VA_ARGS__& control_eq, \
-                                       DeferredLogger& deferred_logger) const; \
+#define INSTANTIATE(T,...)                                               \
+    template void WellGroupControls<T>::                                 \
+        getGroupInjectionControl(const Group&,                           \
+                                 const WellState<T>&,                    \
+                                 const GroupState<T>&,                   \
+                                 const Schedule&,                        \
+                                 const SummaryState&,                    \
+                                 const InjectorType&,                    \
+                                 const __VA_ARGS__& bhp,                 \
+                                 const __VA_ARGS__& injection_rate,      \
+                                 const RateConvFunc& rateConverter,      \
+                                 T efficiencyFactor,                     \
+                                 __VA_ARGS__& control_eq,                \
+                                 DeferredLogger& deferred_logger) const; \
+    template void WellGroupControls<T>::                                 \
+        getGroupProductionControl(const Group&,                          \
+                                  const WellState<T>&,                   \
+                                  const GroupState<T>&,                  \
+                                  const Schedule&,                       \
+                                  const SummaryState&,                   \
+                                  const __VA_ARGS__& bhp,                \
+                                  const std::vector<__VA_ARGS__>&,       \
+                                  const RateConvFunc& rateConverter,     \
+                                  T efficiencyFactor,                    \
+                                  __VA_ARGS__& control_eq,               \
+                                  DeferredLogger& deferred_logger) const;
 
-INSTANCE(DenseAd::Evaluation<double,3,0u>)
-INSTANCE(DenseAd::Evaluation<double,4,0u>)
-INSTANCE(DenseAd::Evaluation<double,5,0u>)
-INSTANCE(DenseAd::Evaluation<double,6,0u>)
-INSTANCE(DenseAd::Evaluation<double,7,0u>)
-INSTANCE(DenseAd::Evaluation<double,8,0u>)
-INSTANCE(DenseAd::Evaluation<double,9,0u>)
-INSTANCE(DenseAd::Evaluation<double,-1,4u>)
-INSTANCE(DenseAd::Evaluation<double,-1,5u>)
-INSTANCE(DenseAd::Evaluation<double,-1,6u>)
-INSTANCE(DenseAd::Evaluation<double,-1,7u>)
-INSTANCE(DenseAd::Evaluation<double,-1,8u>)
-INSTANCE(DenseAd::Evaluation<double,-1,9u>)
-INSTANCE(DenseAd::Evaluation<double,-1,10u>)
+#define INSTANTIATE_TYPE(T)                      \
+    template class WellGroupControls<T>;         \
+    INSTANTIATE(T,DenseAd::Evaluation<T,3,0u>)   \
+    INSTANTIATE(T,DenseAd::Evaluation<T,4,0u>)   \
+    INSTANTIATE(T,DenseAd::Evaluation<T,5,0u>)   \
+    INSTANTIATE(T,DenseAd::Evaluation<T,6,0u>)   \
+    INSTANTIATE(T,DenseAd::Evaluation<T,7,0u>)   \
+    INSTANTIATE(T,DenseAd::Evaluation<T,8,0u>)   \
+    INSTANTIATE(T,DenseAd::Evaluation<T,9,0u>)   \
+    INSTANTIATE(T,DenseAd::Evaluation<T,10,0u>)  \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,4u>)  \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,5u>)  \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,6u>)  \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,7u>)  \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,8u>)  \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,9u>)  \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,10u>) \
+    INSTANTIATE(T,DenseAd::Evaluation<T,-1,11u>)
+
+INSTANTIATE_TYPE(double)
+
+#if FLOW_INSTANTIATE_FLOAT
+INSTANTIATE_TYPE(float)
+#endif
 
 } // namespace Opm
