@@ -30,6 +30,7 @@
 
 #include <opm/simulators/utils/ParallelCommunication.hpp>
 #include <opm/simulators/wells/ParallelWellInfo.hpp>
+#include <opm/simulators/wells/PerforationData.hpp>
 #include <opm/grid/common/p2pcommunicator.hh>
 #include <opm/output/data/Wells.hpp>
 
@@ -126,28 +127,33 @@ void PackUnpackXConn::unpack([[maybe_unused]] const int link,
 
 namespace Opm {
 
-WellState::WellState(const ParallelWellInfo& pinfo)
-    : phase_usage_{}
+template<class Scalar>
+WellState<Scalar>::WellState(const ParallelWellInfo<Scalar>& pinfo)
+    : phase_usage_{{BlackoilPhases::Aqua, BlackoilPhases::Liquid}}
 {
     wells_.add("test4",
-               SingleWellState{"dummy", pinfo, false, 0.0, {}, phase_usage_, 0.0});
+               SingleWellState<Scalar>{"dummy", pinfo, false, 0.0, {}, phase_usage_, 0.0});
 }
 
-WellState WellState::serializationTestObject(const ParallelWellInfo& pinfo)
+template<class Scalar>
+WellState<Scalar> WellState<Scalar>::
+serializationTestObject(const ParallelWellInfo<Scalar>& pinfo)
 {
     WellState result(PhaseUsage{});
-    result.alq_state = ALQState::serializationTestObject();
+    result.alq_state = ALQState<Scalar>::serializationTestObject();
     result.well_rates = {{"test2", {true, {1.0}}}, {"test3", {false, {2.0}}}};
-    result.wells_.add("test4", SingleWellState::serializationTestObject(pinfo));
+    result.wells_.add("test4", SingleWellState<Scalar>::serializationTestObject(pinfo));
 
     return result;
 }
 
-void WellState::base_init(const std::vector<double>& cellPressures,
-                          const std::vector<Well>& wells_ecl,
-                          const std::vector<std::reference_wrapper<ParallelWellInfo>>& parallel_well_info,
-                          const std::vector<std::vector<PerforationData>>& well_perf_data,
-                          const SummaryState& summary_state)
+template<class Scalar>
+void WellState<Scalar>::base_init(const std::vector<Scalar>& cellPressures,
+                                  const std::vector<Scalar>& cellTemperatures,
+                                  const std::vector<Well>& wells_ecl,
+                                  const std::vector<std::reference_wrapper<ParallelWellInfo<Scalar>>>& parallel_well_info,
+                                  const std::vector<std::vector<PerforationData<Scalar>>>& well_perf_data,
+                                  const SummaryState& summary_state)
 {
     // clear old name mapping
     this->wells_.clear();
@@ -159,18 +165,20 @@ void WellState::base_init(const std::vector<double>& cellPressures,
             const Well& well = wells_ecl[w];
 
             // Initialize bhp(), thp(), wellRates(), temperature().
-            initSingleWell(cellPressures, well, well_perf_data[w], parallel_well_info[w], summary_state);
+            initSingleWell(cellPressures, cellTemperatures, well, well_perf_data[w], parallel_well_info[w], summary_state);
         }
     }
 }
 
-void WellState::initSingleProducer(const Well& well,
-                                   const ParallelWellInfo& well_info,
-                                   double pressure_first_connection,
-                                   const std::vector<PerforationData>& well_perf_data,
-                                   const SummaryState& summary_state) {
+template<class Scalar>
+void WellState<Scalar>::initSingleProducer(const Well& well,
+                                           const ParallelWellInfo<Scalar>& well_info,
+                                           Scalar pressure_first_connection,
+                                           const std::vector<PerforationData<Scalar>>& well_perf_data,
+                                           const SummaryState& summary_state)
+{
     const auto& pu = this->phase_usage_;
-    const double temp = 273.15 + 15.56;
+    const Scalar temp = 273.15 + 15.56;
 
     auto& ws = this->wells_.add(well.name(),
                                 SingleWellState{well.name(),
@@ -191,22 +199,23 @@ void WellState::initSingleProducer(const Well& well,
     ws.update_producer_targets(well, summary_state);
 }
 
-void WellState::initSingleInjector(const Well& well,
-                                   const ParallelWellInfo& well_info,
-                                   double pressure_first_connection,
-                                   const std::vector<PerforationData>& well_perf_data,
-                                   const SummaryState& summary_state) {
-
+template<class Scalar>
+void WellState<Scalar>::initSingleInjector(const Well& well,
+                                           const ParallelWellInfo<Scalar>& well_info,
+                                           Scalar pressure_first_connection,
+                                           Scalar temperature_first_connection,
+                                           const std::vector<PerforationData<Scalar>>& well_perf_data,
+                                           const SummaryState& summary_state)
+{
     const auto& pu = this->phase_usage_;
-    const double temp = well.temperature();
-
-    auto& ws = this->wells_.add(well.name(), SingleWellState{well.name(),
-                                                             well_info,
-                                                             false,
-                                                             pressure_first_connection,
-                                                             well_perf_data,
-                                                             pu,
-                                                             temp});
+    const Scalar temp = well.hasInjTemperature() ? well.inj_temperature() : temperature_first_connection;
+    auto& ws = this->wells_.add(well.name(), SingleWellState<Scalar>{well.name(),
+                                                                     well_info,
+                                                                     false,
+                                                                     pressure_first_connection,
+                                                                     well_perf_data,
+                                                                     pu,
+                                                                     temp});
 
     // the rest of the code needs to executed even if ws.perf_data is empty
     // as this does not say anything for the whole well if it is distributed.
@@ -218,19 +227,27 @@ void WellState::initSingleInjector(const Well& well,
     ws.update_injector_targets(well, summary_state);
 }
 
-void WellState::initSingleWell(const std::vector<double>& cellPressures,
-                               const Well& well,
-                               const std::vector<PerforationData>& well_perf_data,
-                               const ParallelWellInfo& well_info,
-                               const SummaryState& summary_state)
+template<class Scalar>
+void WellState<Scalar>::initSingleWell(const std::vector<Scalar>& cellPressures,
+                                       const std::vector<Scalar>& cellTemperatures,
+                                       const Well& well,
+                                       const std::vector<PerforationData<Scalar>>& well_perf_data,
+                                       const ParallelWellInfo<Scalar>& well_info,
+                                       const SummaryState& summary_state)
 {
-    double pressure_first_connection = -1;
-    if (!well_perf_data.empty())
+    Scalar pressure_first_connection = -1;
+    if (!well_perf_data.empty()) {
         pressure_first_connection = cellPressures[well_perf_data[0].cell_index];
+    }
     pressure_first_connection = well_info.broadcastFirstPerforationValue(pressure_first_connection);
 
     if (well.isInjector()) {
-        this->initSingleInjector(well, well_info, pressure_first_connection,
+        Scalar temperature_first_connection = -1;
+        if (!well_perf_data.empty()) {
+            temperature_first_connection = cellTemperatures[well_perf_data[0].cell_index];
+        }
+        temperature_first_connection = well_info.broadcastFirstPerforationValue(temperature_first_connection);
+        this->initSingleInjector(well, well_info, pressure_first_connection, temperature_first_connection,
                                  well_perf_data, summary_state);
     } else {
         this->initSingleProducer(well, well_info, pressure_first_connection,
@@ -238,24 +255,26 @@ void WellState::initSingleWell(const std::vector<double>& cellPressures,
     }
 }
 
-void WellState::init(const std::vector<double>& cellPressures,
-                     const Schedule& schedule,
-                     const std::vector<Well>& wells_ecl,
-                     const std::vector<std::reference_wrapper<ParallelWellInfo>>& parallel_well_info,
-                     const int report_step,
-                     const WellState* prevState,
-                     const std::vector<std::vector<PerforationData>>& well_perf_data,
-                     const SummaryState& summary_state)
+template<class Scalar>
+void WellState<Scalar>::init(const std::vector<Scalar>& cellPressures,
+                             const std::vector<Scalar>& cellTemperatures,
+                             const Schedule& schedule,
+                             const std::vector<Well>& wells_ecl,
+                             const std::vector<std::reference_wrapper<ParallelWellInfo<Scalar>>>& parallel_well_info,
+                             const int report_step,
+                             const WellState* prevState,
+                             const std::vector<std::vector<PerforationData<Scalar>>>& well_perf_data,
+                             const SummaryState& summary_state)
 {
     // call init on base class
-    this->base_init(cellPressures, wells_ecl, parallel_well_info,
+    this->base_init(cellPressures, cellTemperatures, wells_ecl, parallel_well_info,
                     well_perf_data, summary_state);
     this->global_well_info = std::make_optional<GlobalWellInfo>(schedule,
                                                                 report_step,
                                                                 wells_ecl);
     for (const auto& wname : schedule.wellNames(report_step))
     {
-        well_rates.insert({wname, std::make_pair(false, std::vector<double>(this->numPhases()))});
+        well_rates.insert({wname, std::make_pair(false, std::vector<Scalar>(this->numPhases()))});
     }
     for (const auto& winfo: parallel_well_info)
     {
@@ -291,7 +310,7 @@ void WellState::init(const std::vector<double>& cellPressures,
         for (int perf = 0; perf < num_perf_this_well; ++perf) {
             if (wells_ecl[w].getStatus() == Well::Status::OPEN) {
                 for (int p = 0; p < this->numPhases(); ++p) {
-                    perf_data.phase_rates[this->numPhases()*perf + p] = ws.surface_rates[p] / double(global_num_perf_this_well);
+                    perf_data.phase_rates[this->numPhases()*perf + p] = ws.surface_rates[p] / Scalar(global_num_perf_this_well);
                 }
             }
             perf_data.pressure[perf] = cellPressures[well_perf_data[w][perf].cell_index];
@@ -302,15 +321,22 @@ void WellState::init(const std::vector<double>& cellPressures,
         auto& ws = this->well(w);
         if (wells_ecl[w].isProducer()) {
             const auto controls = wells_ecl[w].productionControls(summary_state);
-            ws.production_cmode = controls.cmode;
+            if (controls.cmode == Well::ProducerCMode::GRUP && !wells_ecl[w].isAvailableForGroupControl()) {
+                ws.production_cmode = Well::ProducerCMode::BHP; // wells always has a BHP control
+            } else {
+                ws.production_cmode = controls.cmode;
+            }
         }
         else {
             const auto controls = wells_ecl[w].injectionControls(summary_state);
-            ws.injection_cmode = controls.cmode;
+            if (controls.cmode == Well::InjectorCMode::GRUP && !wells_ecl[w].isAvailableForGroupControl()) {
+                ws.injection_cmode = Well::InjectorCMode::BHP; // wells always has a BHP control
+            } else {
+                ws.injection_cmode = controls.cmode;
+            }
         }
     }
 
-
     for (int w = 0; w < nw; ++w) {
         switch (wells_ecl[w].getStatus()) {
         case Well::Status::SHUT:
@@ -370,6 +396,7 @@ void WellState::init(const std::vector<double>& cellPressures,
                 // perforations is equal, otherwise initialize
                 // perfphaserates to well rates divided by the
                 // number of perforations.
+                // TODO: we might still need the values from the prev_well if the connection structure changes
                 if (global_num_perf_same)
                 {
                     auto& perf_data = new_well.perf_data;
@@ -381,50 +408,45 @@ void WellState::init(const std::vector<double>& cellPressures,
                     auto& target_rates = perf_data.phase_rates;
                     for (int perf_index = 0; perf_index < num_perf_this_well; perf_index++) {
                         for (int p = 0; p < np; ++p) {
-                            target_rates[perf_index*np + p] = new_well.surface_rates[p] / double(global_num_perf_this_well);
+                            target_rates[perf_index*np + p] = new_well.surface_rates[p] / Scalar(global_num_perf_this_well);
                         }
                     }
                 }
 
                 // Productivity index.
                 new_well.productivity_index = prev_well.productivity_index;
-            }
 
-            // If in the new step, there is no THP related
-            // target/limit anymore, its thp value should be set to
-            // zero.
-            const bool has_thp = well.isInjector()
-                ? well.injectionControls (summary_state).hasControl(Well::InjectorCMode::THP)
-                : well.productionControls(summary_state).hasControl(Well::ProducerCMode::THP);
-
-            if (!has_thp) {
-                new_well.thp = 0;
+                // if there is no valid VFP table associated, we set the THP value to be 0.
+                if (well.vfp_table_number() == 0) {
+                    new_well.thp = 0.;
+                }
             }
         }
     }
 
-
-    updateWellsDefaultALQ(wells_ecl);
+    updateWellsDefaultALQ(schedule, report_step, summary_state);
 }
 
-void WellState::resize(const std::vector<Well>& wells_ecl,
-                       const std::vector<std::reference_wrapper<ParallelWellInfo>>& parallel_well_info,
-                       const Schedule& schedule,
-                       const bool handle_ms_well,
-                       const size_t numCells,
-                       const std::vector<std::vector<PerforationData>>& well_perf_data,
-                       const SummaryState& summary_state)
+template<class Scalar>
+void WellState<Scalar>::resize(const std::vector<Well>& wells_ecl,
+                               const std::vector<std::reference_wrapper<ParallelWellInfo<Scalar>>>& parallel_well_info,
+                               const Schedule& schedule,
+                               const bool handle_ms_well,
+                               const std::size_t numCells,
+                               const std::vector<std::vector<PerforationData<Scalar>>>& well_perf_data,
+                               const SummaryState& summary_state)
 {
-    const std::vector<double> tmp(numCells, 0.0); // <- UGLY HACK to pass the size
-    init(tmp, schedule, wells_ecl, parallel_well_info, 0, nullptr, well_perf_data, summary_state);
+    const std::vector<Scalar> tmp(numCells, 0.0); // <- UGLY HACK to pass the size
+    init(tmp, tmp, schedule, wells_ecl, parallel_well_info, 0, nullptr, well_perf_data, summary_state);
 
     if (handle_ms_well) {
         initWellStateMSWell(wells_ecl, nullptr);
     }
 }
 
-const std::vector<double>&
-WellState::currentWellRates(const std::string& wellName) const
+template<class Scalar>
+const std::vector<Scalar>&
+WellState<Scalar>::currentWellRates(const std::string& wellName) const
 {
     auto it = well_rates.find(wellName);
 
@@ -435,10 +457,11 @@ WellState::currentWellRates(const std::string& wellName) const
     return it->second.second;
 }
 
-template<class Communication>
-void WellState::gatherVectorsOnRoot(const std::vector<data::Connection>& from_connections,
-                                                         std::vector<data::Connection>& to_connections,
-                                                         const Communication& comm) const
+template<class Scalar>
+void WellState<Scalar>::
+gatherVectorsOnRoot(const std::vector<data::Connection>& from_connections,
+                    std::vector<data::Connection>& to_connections,
+                    const Parallel::Communication& comm) const
 {
     auto send = std::set<int>{};
     auto recv = std::set<int>{};
@@ -460,9 +483,10 @@ void WellState::gatherVectorsOnRoot(const std::vector<data::Connection>& from_co
     toOwnerComm.exchange(lineariser);
 }
 
+template<class Scalar>
 data::Wells
-WellState::report(const int* globalCellIdxMap,
-                  const std::function<bool(const int)>& wasDynamicallyClosed) const
+WellState<Scalar>::report(const int* globalCellIdxMap,
+                          const std::function<bool(const int)>& wasDynamicallyClosed) const
 {
     if (this->numWells() == 0) {
         return {};
@@ -491,6 +515,9 @@ WellState::report(const int* globalCellIdxMap,
         well.bhp = ws.bhp;
         well.thp = ws.thp;
         well.temperature = ws.temperature;
+        well.filtrate.rate = ws.sum_filtrate_rate();
+        well.filtrate.total = ws.sum_filtrate_total();
+        well.filtrate.concentration = ws.filtrate_conc;
 
         if (pu.phase_used[BlackoilPhases::Aqua]) {
             well.rates.set(rt::wat, wv[ pu.phase_pos[BlackoilPhases::Aqua] ] );
@@ -532,9 +559,11 @@ WellState::report(const int* globalCellIdxMap,
             well.rates.set(rt::alq, 0.0);
         }
 
-        well.rates.set(rt::dissolved_gas, ws.dissolved_gas_rate + ws.dissolved_gas_rate_in_water);
-        well.rates.set(rt::vaporized_oil, ws.vaporized_oil_rate);
-        well.rates.set(rt::vaporized_water, ws.vaporized_wat_rate);
+        well.rates.set(rt::dissolved_gas,
+                       ws.phase_mixing_rates[ws.dissolved_gas] +
+                       ws.phase_mixing_rates[ws.dissolved_gas_in_water]);
+        well.rates.set(rt::vaporized_oil, ws.phase_mixing_rates[ws.vaporized_oil]);
+        well.rates.set(rt::vaporized_water, ws.phase_mixing_rates[ws.vaporized_water]);
 
         {
             auto& curr = well.current_control;
@@ -565,17 +594,20 @@ WellState::report(const int* globalCellIdxMap,
     return res;
 }
 
-void WellState::reportConnections(std::vector<data::Connection>& connections,
-                                  const PhaseUsage &pu,
-                                  std::size_t well_index,
-                                  const int* globalCellIdxMap) const
+template<class Scalar>
+void WellState<Scalar>::reportConnections(std::vector<data::Connection>& connections,
+                                          const PhaseUsage &pu,
+                                          std::size_t well_index,
+                                          const int* globalCellIdxMap) const
 {
     using rt = data::Rates::opt;
-    const auto& perf_data = this->well(well_index).perf_data;
+    const auto& ws = this->well(well_index);
+    const auto& perf_data = ws.perf_data;
     const int num_perf_well = perf_data.size();
     connections.resize(num_perf_well);
     const auto& perf_rates = perf_data.rates;
     const auto& perf_pressure = perf_data.pressure;
+    const auto& perf_mixing_rates = perf_data.phase_mixing_rates;
     for (int i = 0; i < num_perf_well; ++i) {
       const auto active_index = perf_data.cell_index[i];
         auto& connection = connections[ i ];
@@ -583,6 +615,22 @@ void WellState::reportConnections(std::vector<data::Connection>& connections,
         connection.pressure = perf_pressure[i];
         connection.reservoir_rate = perf_rates[i];
         connection.trans_factor = perf_data.connection_transmissibility_factor[i];
+        connection.d_factor = perf_data.connection_d_factor[i];
+        connection.compact_mult = perf_data.connection_compaction_tmult[i];
+        connection.rates.set(rt::dissolved_gas, perf_mixing_rates[i][ws.dissolved_gas]);
+        connection.rates.set(rt::vaporized_oil, perf_mixing_rates[i][ws.vaporized_oil]);
+        if (!ws.producer) {
+            const auto& filtrate_data = perf_data.filtrate_data;
+            auto& filtrate = connection.filtrate;
+            filtrate.rate = filtrate_data.rates[i];
+            filtrate.total = filtrate_data.total[i];
+            filtrate.skin_factor = filtrate_data.skin_factor[i];
+            filtrate.thickness = filtrate_data.thickness[i];
+            filtrate.poro = filtrate_data.poro[i];
+            filtrate.perm = filtrate_data.perm[i];
+            filtrate.radius = filtrate_data.radius[i];
+            filtrate.area_of_flow = filtrate_data.area_of_flow[i];
+        }
     }
 
     const int np = pu.num_phases;
@@ -603,7 +651,7 @@ void WellState::reportConnections(std::vector<data::Connection>& connections,
         pi .at( pu.phase_pos[Gas] ) = rt::productivity_index_gas;
     }
 
-    size_t local_conn_index = 0;
+    std::size_t local_conn_index = 0;
     for (auto& comp : connections) {
         const auto * rates = &perf_data.phase_rates[np * local_conn_index];
         const auto * connPI = &perf_data.prod_index[np * local_conn_index];
@@ -625,13 +673,13 @@ void WellState::reportConnections(std::vector<data::Connection>& connections,
             const auto& perf_solvent_rate = perf_data.solvent_rates;
             comp.rates.set( rt::solvent, perf_solvent_rate[local_conn_index] );
         }
-
         ++local_conn_index;
     }
 }
 
-void WellState::initWellStateMSWell(const std::vector<Well>& wells_ecl,
-                                    const WellState* prev_well_state)
+template<class Scalar>
+void WellState<Scalar>::initWellStateMSWell(const std::vector<Well>& wells_ecl,
+                                            const WellState* prev_well_state)
 {
     // still using the order in wells
     const int nw = wells_ecl.size();
@@ -652,14 +700,14 @@ void WellState::initWellStateMSWell(const std::vector<Well>& wells_ecl,
             // assuming the order of the perforations in well_ecl is the same with Wells
             const WellConnections& completion_set = well_ecl.getConnections();
             // number of segment for this single well
-            ws.segments = SegmentState{np, segment_set};
+            ws.segments = SegmentState<Scalar>{np, segment_set};
             const int well_nseg = segment_set.size();
             int n_activeperf = 0;
 
             // we need to know for each segment, how many perforation it has and how many segments using it as outlet_segment
             // that is why I think we should use a well model to initialize the WellState here
             std::vector<std::vector<int>> segment_perforations(well_nseg);
-            for (size_t perf = 0; perf < completion_set.size(); ++perf) {
+            for (std::size_t perf = 0; perf < completion_set.size(); ++perf) {
                 const Connection& connection = completion_set.get(perf);
                 if (connection.state() == Connection::State::OPEN) {
                     const int segment_index = segment_set.segmentNumberToIndex(connection.segment());
@@ -703,7 +751,7 @@ void WellState::initWellStateMSWell(const std::vector<Well>& wells_ecl,
             }
 
             const auto& perf_rates = perf_data.phase_rates;
-            std::vector<double> perforation_rates(perf_rates.begin(), perf_rates.end());
+            std::vector<Scalar> perforation_rates(perf_rates.begin(), perf_rates.end());
 
             calculateSegmentRates(segment_inlets, segment_perforations, perforation_rates, np, 0 /* top segment */, ws.segments.rates);
             // for the segment pressure, the segment pressure is the same with the first perforation belongs to the segment
@@ -759,12 +807,13 @@ void WellState::initWellStateMSWell(const std::vector<Well>& wells_ecl,
     }
 }
 
-void
-WellState::calculateSegmentRates(const std::vector<std::vector<int>>& segment_inlets,
-                                 const std::vector<std::vector<int>>& segment_perforations,
-                                 const std::vector<double>& perforation_rates,
-                                 const int np, const int segment,
-                                 std::vector<double>& segment_rates)
+template<class Scalar>
+void WellState<Scalar>::
+calculateSegmentRates(const std::vector<std::vector<int>>& segment_inlets,
+                      const std::vector<std::vector<int>>& segment_perforations,
+                      const std::vector<Scalar>& perforation_rates,
+                      const int np, const int segment,
+                      std::vector<Scalar>& segment_rates)
 {
     // the rate of the segment equals to the sum of the contribution from the perforations and inlet segment rates.
     // the first segment is always the top segment, its rates should be equal to the well rates.
@@ -787,32 +836,36 @@ WellState::calculateSegmentRates(const std::vector<std::vector<int>>& segment_in
     }
 }
 
-void WellState::stopWell(int well_index)
+template<class Scalar>
+void WellState<Scalar>::stopWell(int well_index)
 {
     auto& ws = this->well(well_index);
     ws.stop();
 }
 
-void WellState::openWell(int well_index)
+template<class Scalar>
+void WellState<Scalar>::openWell(int well_index)
 {
     auto& ws = this->well(well_index);
     ws.open();
 }
 
-void WellState::shutWell(int well_index)
+template<class Scalar>
+void WellState<Scalar>::shutWell(int well_index)
 {
     auto& ws = this->well(well_index);
     ws.shut();
 }
 
-void WellState::updateStatus(int well_index, WellStatus status)
+template<class Scalar>
+void WellState<Scalar>::updateStatus(int well_index, WellStatus status)
 {
     auto& ws = this->well(well_index);
     ws.updateStatus(status);
 }
 
-template<class Comm>
-void WellState::communicateGroupRates(const Comm& comm)
+template<class Scalar>
+void WellState<Scalar>::communicateGroupRates(const Parallel::Communication& comm)
 {
     // Compute the size of the data.
     std::size_t sz = 0;
@@ -824,9 +877,8 @@ void WellState::communicateGroupRates(const Comm& comm)
     }
     sz += this->alq_state.pack_size();
 
-
     // Make a vector and collect all data into it.
-    std::vector<double> data(sz);
+    std::vector<Scalar> data(sz);
     std::size_t pos = 0;
     for (const auto& [_, owner_rates] : this->well_rates) {
         (void)_;
@@ -838,7 +890,9 @@ void WellState::communicateGroupRates(const Comm& comm)
                 data[pos++] = 0;
         }
     }
-    pos += this->alq_state.pack_data(&data[pos]);
+    if (!data.empty() && pos != sz) {
+        pos += this->alq_state.pack_data(&data[pos]);
+    }
     assert(pos == sz);
 
     // Communicate it with a single sum() call.
@@ -852,12 +906,14 @@ void WellState::communicateGroupRates(const Comm& comm)
         for (auto& value : rates)
             value = data[pos++];
     }
-    pos += this->alq_state.unpack_data(&data[pos]);
+    if (!data.empty() && pos != sz) {
+        pos += this->alq_state.unpack_data(&data[pos]);
+    }
     assert(pos == sz);
 }
 
-template<class Comm>
-void WellState::updateGlobalIsGrup(const Comm& comm)
+template<class Scalar>
+void WellState<Scalar>::updateGlobalIsGrup(const Parallel::Communication& comm)
 {
     this->global_well_info.value().clear();
     for (std::size_t well_index = 0; well_index < this->size(); well_index++) {
@@ -870,12 +926,14 @@ void WellState::updateGlobalIsGrup(const Comm& comm)
     this->global_well_info.value().communicate(comm);
 }
 
+template<class Scalar>
 data::Segment
-WellState::reportSegmentResults(const int well_id,
-                                const int seg_ix,
-                                const int seg_no) const
+WellState<Scalar>::reportSegmentResults(const int well_id,
+                                        const int seg_ix,
+                                        const int seg_no) const
 {
     using PhaseQuant = data::SegmentPhaseQuantity::Item;
+    using PhaseDensity = data::SegmentPhaseDensity::Item;
 
     const auto& segments = this->well(well_id).segments;
     if (segments.empty()) {
@@ -900,6 +958,7 @@ WellState::reportSegmentResults(const int well_id,
     const auto* velocity = &segments.phase_velocity[seg_ix * pu.num_phases];
     const auto* holdup = &segments.phase_holdup[seg_ix * pu.num_phases];
     const auto* viscosity = &segments.phase_viscosity[seg_ix * pu.num_phases];
+    const auto* density = &segments.phase_density[seg_ix * (pu.num_phases + 2)]; // +2 for mixture densities
 
     if (pu.phase_used[Water]) {
         const auto iw = pu.phase_pos[Water];
@@ -909,6 +968,7 @@ WellState::reportSegmentResults(const int well_id,
         seg_res.velocity.set(PhaseQuant::Water, velocity[iw]);
         seg_res.holdup.set(PhaseQuant::Water, holdup[iw]);
         seg_res.viscosity.set(PhaseQuant::Water, viscosity[iw]);
+        seg_res.density.set(PhaseDensity::Water, density[iw]);
     }
 
     if (pu.phase_used[Oil]) {
@@ -920,6 +980,7 @@ WellState::reportSegmentResults(const int well_id,
         seg_res.velocity.set(PhaseQuant::Oil, velocity[io]);
         seg_res.holdup.set(PhaseQuant::Oil, holdup[io]);
         seg_res.viscosity.set(PhaseQuant::Oil, viscosity[io]);
+        seg_res.density.set(PhaseDensity::Oil, density[io]);
     }
 
     if (pu.phase_used[Gas]) {
@@ -931,15 +992,23 @@ WellState::reportSegmentResults(const int well_id,
         seg_res.velocity.set(PhaseQuant::Gas, velocity[ig]);
         seg_res.holdup.set(PhaseQuant::Gas, holdup[ig]);
         seg_res.viscosity.set(PhaseQuant::Gas, viscosity[ig]);
+        seg_res.density.set(PhaseDensity::Gas, density[ig]);
     }
 
     seg_res.segNumber = seg_no;
 
+    // Recall: The mixture density *without* exponents is stored at offset
+    // 'num_phases' and the mixture density *with* exponents is stored at
+    // offset 'num_phases + 1'.
+    seg_res.density.set(PhaseDensity::Mixture, density[pu.num_phases]);
+    seg_res.density.set(PhaseDensity::MixtureWithExponents, density[pu.num_phases + 1]);
+
     return seg_res;
 }
 
-bool WellState::wellIsOwned(std::size_t well_index,
-                            [[maybe_unused]] const std::string& wellName) const
+template<class Scalar>
+bool WellState<Scalar>::wellIsOwned(std::size_t well_index,
+                                    [[maybe_unused]] const std::string& wellName) const
 {
     const auto& well_info = this->parallelWellInfo(well_index);
     assert(well_info.name() == wellName);
@@ -947,7 +1016,8 @@ bool WellState::wellIsOwned(std::size_t well_index,
     return well_info.isOwner();
 }
 
-bool WellState::wellIsOwned(const std::string& wellName) const
+template<class Scalar>
+bool WellState<Scalar>::wellIsOwned(const std::string& wellName) const
 {
     const auto& well_index = this->index(wellName);
     if (!well_index.has_value()) {
@@ -958,34 +1028,43 @@ bool WellState::wellIsOwned(const std::string& wellName) const
     return wellIsOwned(well_index.value(), wellName);
 }
 
-void WellState::updateWellsDefaultALQ(const std::vector<Well>& wells_ecl)
+template <typename Scalar>
+void WellState<Scalar>::updateWellsDefaultALQ(const Schedule& schedule,
+                                              const int report_step,
+                                              const SummaryState& summary_state)
 {
-    const int nw = wells_ecl.size();
-    for (int i = 0; i<nw; i++) {
-        const Well &well = wells_ecl[i];
-        if (well.isProducer()) {
-            // NOTE: This is the value set in item 12 of WCONPROD, or with WELTARG
-            auto alq = well.alq_value();
-            this->alq_state.update_default(well.name(), alq);
+    const auto wells = schedule.wellNames(report_step);
+    for (const auto& wname : wells) {
+        const auto& well = schedule.getWell(wname, report_step);
+        if (! well.isProducer()) {
+            continue;
         }
+        const auto alq = well.alq_value(summary_state);
+        this->alq_state.update_default(wname, alq);
     }
 }
 
-bool WellState::operator==(const WellState& rhs) const
+template<class Scalar>
+bool WellState<Scalar>::operator==(const WellState& rhs) const
 {
     return this->alq_state == rhs.alq_state &&
            this->well_rates == rhs.well_rates &&
            this->wells_ == rhs.wells_;
 }
 
-const ParallelWellInfo&
-WellState::parallelWellInfo(std::size_t well_index) const
+template<class Scalar>
+const ParallelWellInfo<Scalar>&
+WellState<Scalar>::parallelWellInfo(std::size_t well_index) const
 {
     const auto& ws = this->well(well_index);
     return ws.parallel_info;
 }
 
-template void WellState::updateGlobalIsGrup<Parallel::Communication>(const Parallel::Communication& comm);
-template void WellState::communicateGroupRates<Parallel::Communication>(const Parallel::Communication& comm);
+
+template class WellState<double>;
+
+#if FLOW_INSTANTIATE_FLOAT
+template class WellState<float>;
+#endif
 
 } // namespace Opm
