@@ -123,7 +123,7 @@ testCusparseSolver(Opm::BdaBridge<Matrix<bz>, Vector<bz>, bz>& bridge, Matrix<bz
 {
     Dune::InverseOperatorResult result;
     Vector<bz> x(rhs.size());
-    auto wellContribs = Opm::WellContributions::create("cusparse", false);
+    auto wellContribs = Opm::WellContributions<double>::create("cusparse", false);
     auto mat2 = matrix; // deep copy to make sure nnz values are in contiguous memory
                         // matrix created by readMatrixMarket() did not have contiguous memory
     bridge.solve_system(&mat2, &mat2, /*numJacobiBlocks=*/0, rhs, *wellContribs, result);
@@ -138,7 +138,7 @@ testCusparseSolverJacobi(Opm::BdaBridge<Matrix<bz>, Vector<bz>, bz>& bridge, Mat
 {
     Dune::InverseOperatorResult result;
     Vector<bz> x(rhs.size());
-    auto wellContribs = Opm::WellContributions::create("cusparse", false);
+    auto wellContribs = Opm::WellContributions<double>::create("cusparse", false);
     auto mat2 = matrix; // deep copy to make sure nnz values are in contiguous memory
                         // matrix created by readMatrixMarket() did not have contiguous memory
     auto mat3 = matrix; // another deep copy, to make sure Jacobi matrix memory is different
